/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.awt.event.HierarchyBoundsListener;
import java.awt.event.HierarchyEvent;
import java.nio.ByteBuffer;
/**
 * <code>ThincCMsgEventMouseButton</code> handles resized event from canvas. 
 * it simply implements <code>HierarchyBoundsListener</code> interfaces provided from java, and it is added
 * as a HierarchyBoundsListener on ThincClientCanvas.
 * 
 * @author Young Jin Yoon
 * @see ThincCMsgResize
 */
public final class ThincCMsgResizeEvent extends ThincCMsgResize implements HierarchyBoundsListener{
	public static final int MESSAGE_TYPE = ThincProtoResize.T_RESZ_EVENT;
	public static final int MESSAGE_SIZE = ThincProtoResize.T_RESZ_EVENT_SIZE;
	
	private int width;
	private int height;

	private ThincFrame frame;
	/**
     * creates ThincCMsgResizeEvent object.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param c <code>ThincClientCanvas</code> to get the information
     */
	public ThincCMsgResizeEvent(ThincMsgHandler h, ThincClientCanvas c, ThincFrame f) {
		super(h, c);
		frame = f;
	}


	@Override
	public byte getMsgType() {
		return MESSAGE_TYPE;
	}

	/**
	 * putMessage()
	 * 
	 * simply puts resized width and height.
	 */
	@Override
	public void putMessage(ByteBuffer b) {
		b.putShort((short)width);
		b.putShort((short)height);
	}

	@Override
	public short messageSize() {
		return MESSAGE_SIZE;
	}

	/**
	 * process()
	 * 
	 * Set resized width and height into ThincSMsgFlagResize class and
	 * invokes send method in ThincMsgHandler.
	 * It also creates ThincCMsgResizeUpdate object
	 * and invokes process, also.
	 * (which means send update message for entire screen right after
	 * send itself to the server.)
	 */
	@Override
	public void process() throws ThincException {
		ThincCMsgResizeUpdate update = new ThincCMsgResizeUpdate((ThincMsgHandler)handler,canvas,0,0,canvas.getFBWidth(),canvas.getFBHeight());
		ThincSMsgFlagResize.setResized(canvas.getFBWidth(), canvas.getFBHeight(), width, height);
		canvas.setResized(width, height);
		if(!canvas.getClientResize()){
			try {
				handler.send(this);
				update.process();
			} catch (ThincException e) {
				throw new ThincException("process: send on ResizeEvent",e);
			}
		}
	}

	/* implemented methods from HierarchyBoundsListener interfaces below*/

	/**
	 * ancestorResized()
	 * 
	 * implemented method from HierarchyBoundsListener interface.
	 * it saves resized height and width, then invokes process.
	 */
	public void ancestorResized(HierarchyEvent arg0) {
		width = arg0.getComponent().getWidth();
		height = arg0.getComponent().getHeight();
		if(width > canvas.getFBWidth() || height > canvas.getFBHeight()) {
			frame.pack();
			width = canvas.getFBWidth();
			height = canvas.getFBHeight();
		}
		try {
			process();
		} catch (ThincException e) {
			e.printStackTrace();
			try {
				handler.closeSocketChannel();
			} catch (ThincException e1) {
				e1.printStackTrace();
			}
			//TODO fill it in. make setLogin(String reason) in ThincClientCanvas!
		}
	}
	
	public void ancestorMoved(HierarchyEvent arg0) {}

}
