/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * main class for entire thinc client. it create ThincFrame object to get the  argument from user. Next, it pass the argument into ThincMsgHandler object from ThincFrame object. it creates ThincClientCanvas object and finally,  invokes ThincMsgHandler.init() and ThincMsgHandler.start() to handle the message from the server.
 * @author  Young Jin Yoon
 * @version  1.8
 * @since  2006
 */

public class ThincClient {
	public static final int SEC_ENC = ThincProto2.T_CAPS_SEC_ENC;
	public static final int SEC_AUTH = ThincProto2.T_CAPS_SEC_AUTH;

	public static final int DEFAULT_PORT = 20000;
	public static final int OFFSET_SND_PORT = 10000;
	public static final String DEFAULT_HOST = "192.168.0.1";
	public static final byte[] CLIENT_VERSION             = new byte[] {'T','H','I','N','C','_','0','.','2','\0'};
//	public static final int DEFAULT_SEC_CAP = SEC_AUTH | SEC_ENC; // SSL should be set as default
	public static final int DEFAULT_SEC_CAP = SEC_AUTH; // TODO: making SSL connection extremely slows in Linux
	public static final String DEFAULT_ID = "default_id";
	public static final String DEFAULT_PASSWD = "really_easy_password";

	private ThincClientCanvas canvas;
	private ThincFrame frame;
	private ThincMsgHandler handler;
	private ThincSoundHandler sound;
	
	private String host = DEFAULT_HOST;
	private int port = DEFAULT_PORT;

	private boolean clientResize = false; // true if you want resizing in client, not from server
	private boolean isSound = true; // true if you want to hear the sound :)
	
	private int secCap = DEFAULT_SEC_CAP;
	private String id = null;
	private String passwd = null;

	private boolean isReady = false;
	private int sndPort = DEFAULT_PORT + OFFSET_SND_PORT;

	/**
	 * main()
	 * just makes a ThincClient object. the constructor of ThincClient 
	 * does everything.
	 * @param args
	 */
	public static void main(String[] args) {

		@SuppressWarnings("unused")
		ThincClient t = new ThincClient();
	}

	/**
	 * ThincClient()
	 * makes ThincFrame object and wait the user input by using
	 * Thread.sleep(). then pass the argument into ThincMsgHandler
	 * object and invoke init() and start().
	 */
	public ThincClient ()
	{
		/* creating ThincFrame object */
		try {
		    frame = new ThincFrame(this);
		} catch (ThincException e1) {
			System.err.println("Error: Cannot initialize the frame");
			e1.printStackTrace();
			System.exit(0);
		}
		while(true) {
			/* waiting user input */
			try {
				while(!isReady) // if user input is completed, isReady flag should be true.
					Thread.sleep(100);
			} catch (InterruptedException e1) {
				e1.printStackTrace();
				System.exit(0);
			}
			/* create ThincClientCanvas and ThincMsgHandler object.*/
			try {
				canvas = new ThincClientCanvas(this);
				handler = new ThincMsgHandler(this,canvas);
			} catch (ThincException e) {
				e.printStackTrace();
				frame.getMsgLabel().setText("Cannot connect to "+host+":"+port);
					//Because ThincMsgHandler contructor tries to connect to server,
					//it prints Cannot connect message to user interface. 
				isReady = false;  
				continue; // retry to get the information from user.
			}
			/* open the sound from port 30000 */
			if(isSound){
				try {
					sound = new ThincSoundHandler(this,canvas);
				} catch (ThincException e) {
					System.err.println("Error: Cannot initialize the sound");
					e.printStackTrace();
					sound = null; // if not supported make it as null
				}
			}
			/* set the frame and canvas for each ThincClientCanvas and ThincFrame object,
			 and start ThincSoundHandler and ThincMsgHandler. */
			if(sound != null && isSound) {
				Thread soundThread = new Thread(sound);
				soundThread.start();
			}
			try {
				canvas.setFrame(frame);
				frame.setCanvas(canvas);
				handler.init();
				handler.start();
			} catch (ThincException e) {
				e.printStackTrace();
				frame.setLogin();	//change ThincFrame to login window to get the input again.
				frame.getMsgLabel().setText(e.getMessage()); // show the error to the user
				isReady = false;
				close();
				continue; // retry to get the information from user.
			}
		}
	}

	/**
	 * close()
	 * when the user pushes X on the window, or error occurs, it should be invoked by
	 * try-catch phases.
	 */
	public void close()
	{
		/* close the socket in ThincSound and ThincMsgHandler object */
		try {
			if(handler != null) handler.closeSocketChannel();
			if(sound != null) sound.closeSocketChannel();
		} catch (ThincException e) {
			e.printStackTrace();
		}
	}
	/* encapsulated method used by ThincSound and ThincMsgHandler object */
	public int getSecCap() {return secCap;}
	public String getHost() {return host;}
	public int getPort() {return port;}
	public int getSndPort() {return sndPort;}
	/* encapsulated method used by ThincSMsgInitSecCaps object */
	public void setSecCap(int sec) {secCap = sec;}

	/* encapsulated method used by ThincCMsgInitAuth object */
	public String getPasswd(){return passwd;}
	public String getId(){return id;}

	/**
	 * setLoginProperty()
	 * used to set the initial information from the user. invoked by
	 * ThincFrame object
	 * @param server Server DNS
	 * @param port Server port
	 * @param auth Whether auth is requested by user or not
	 * @param ssl Whether SSL is requested by user or not
	 * @param id User id to connect the server
	 * @param passwd Password to connect the server
	 * @see ThincFrame
	 */
	public void setLoginProperty(String server, int port, boolean auth, boolean ssl, boolean sound,int sndPort, String id, String passwd) {
		if(host != null) host = server;
		if(this.port != 0) this.port = port;
		if(auth) secCap |= SEC_AUTH;
		else secCap = secCap & (~SEC_AUTH); // authentication option should be disabled.
		if(ssl) secCap |= SEC_ENC;
		else secCap = secCap & (~SEC_ENC);
		this.sndPort  = sndPort;
		isSound = sound;
		this.id = id;
		this.passwd = passwd;
		isReady = true;
	}

	public boolean getClientResize() {
		return clientResize;
	}
}
