/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.SocketChannel;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import javax.net.ssl.HandshakeCompletedEvent;
import javax.net.ssl.HandshakeCompletedListener;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLEngine;
import javax.net.ssl.SSLEngineResult;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.net.ssl.SSLEngineResult.HandshakeStatus;

/**
 * abstract class for handling network messages. it provides various methods to implement any network handler derived from this.
 * @author  Young Jin Yoon
 */
public abstract class ThincHandler {
	protected static final int THINC_HEADERSIZE = ThincProto.THINC_HEADERSIZE; //4
	protected static final int MAX_DIRECT_ALLOCATE = 16384;
//	private static final int SLEEPING_DELAY = 30;
	protected SocketChannel sc;
	protected ByteBuffer recv;
	protected ByteBuffer send;
	protected ThincClient client;
	protected ThincClientCanvas canvas;
	protected byte lastSent;

	protected boolean isSSL = false;
	protected SSLEngine sslEngine = null;
	protected SSLSocket sslSocket = null;
	
	protected ByteBuffer clientAppData;
	protected ByteBuffer clientPacData;
	protected ByteBuffer serverAppData;
	protected ByteBuffer serverPacData;	
	protected SSLEngineResult rcvResult = null;
	
	public ThincHandler(ThincClient tc, ThincClientCanvas ca) {
		client = tc;
		canvas = ca;
		recv = null;
		send = null;
		sc = null;
	}
	
	/**
	 * createSocketChannel()
	 * 
	 * creates network connection to server
	 * @param host domain name for server
	 * @param port port number for server
	 * @return <code>SocketChannel</code> object to be created
	 * @throws ThincException
	 */
	protected SocketChannel createSocketChannel(String host,int port) throws ThincException{
		System.out.println("Attempting to connect to host " + host + ", port "
            + port + "...");
		InetSocketAddress i;
		try {		        
            sc = SocketChannel.open();
            // set to block until initial handshake is completed.
            sc.configureBlocking(true);
			i = new InetSocketAddress(host,port);
			System.out.println(i.getAddress());
            sc.socket().connect(i, 100); // set timeout for 100 milli seconds. Just for Linux Environment
            sc.finishConnect();
        } catch (IOException e) {
            throw new ThincException("An error occurred while attempting to "
                + "connect to host " + host + ", port " + port + ".", e);
        }
        if (!sc.isConnected()) {
            throw new ThincException("Could not connect to server with host "
                + host + ", port " + port + ". Handshaking was not "
                + "successful.");
        }
        return sc;
	}
	/**
	 * closeSocketChannel()
	 * 
	 * closes the connection from the server
	 * @throws ThincException
	 */
	protected void closeSocketChannel() throws ThincException
	{
		try {
			if(isSSL){
				sslEngine.closeOutbound();
			}
			sc.close();
		} catch (IOException e) {
			throw new ThincException ("clostSocketChannel: Error on close",e);
		}
	}
	
	/**
	 * send()
	 * 
	 * send messages to the server. 
	 * @param c <code>ThincCMsg</code> object to send
	 * @return number of bytes sent
	 * @throws ThincException
	 */
	public int send(ThincCMsg c) throws ThincException {
		int retVal;
		
		/* allocate & initialize ByteBuffer. if previous one is suitable to send message, 
		 * just use it. otherwise it should be created by ByteBuffer.allocateDirect() 
		 * or ByteBuffer.allocate() */
		if(send == null || (((int)c.messageSize()+THINC_HEADERSIZE) > send.capacity())) {
			if(((int)c.messageSize()+THINC_HEADERSIZE)< MAX_DIRECT_ALLOCATE) 
				send = ByteBuffer.allocateDirect((int)c.messageSize()+THINC_HEADERSIZE);
			else
				send = ByteBuffer.allocate((int)c.messageSize()+THINC_HEADERSIZE);
		}			
		else 
			send.clear();
		send.limit((int)c.messageSize()+THINC_HEADERSIZE);
		send.order(ByteOrder.BIG_ENDIAN);

		/* put Thinc header by using putThincHeader method */
		putThincHeader(send, c.getMsgType(),c.getFlag(),c.headerSize());
		
		/* invokes putMessage() for ThincCMsg object */
		c.putMessage(send);
	
		send.flip();	//filps the buffer to send it 
		/* send the messages using primitive sendBytes() method */
		try {
		retVal = sendBytes(send);
		} catch (ThincException e) {
			throw new ThincException("send",e);
		}

		send.clear(); // clears the buffer for next use
		return retVal;
	}
	
	/**
	 * putThincHeader()
	 * 
	 * puts the general thinc header into <code>ByteBuffer</code>
	 * @param b target <code>ByteBuffer</code> to put
	 * @param type type of the messages
	 * @param flag flag of the messages
	 * @param length length of the messages
	 */
	private void putThincHeader(ByteBuffer b, byte type, byte flag, short length) {
		b.put(type);
		b.put(flag);
		b.putShort(length);
	}

	/**
	 * sendBytes()
	 * 
	 * is the primitive method used by send method. it also
	 * used when version exchange required (which does not have
	 * general thinc header 
	 * @param bb <code>ByteBuffer</code> to send
	 * @return number of bytes sent
	 * @throws ThincException
	 */
	protected int sendBytes(ByteBuffer bb) throws ThincException {
		int numBytes = 0;
		try {
			if(isSSL)
				numBytes = wrapSSL(bb);
			else {
				while(bb.hasRemaining()){ // try to send it until ByteBuffer is empty
					numBytes += sc.write(bb);
				}
			}
		} catch (IOException e) {
			throw new ThincException("sendBytes",e);
		} catch (ThincException e) {
			throw new ThincException("sendBytes: SSLException",e);
		}
		return numBytes;
	}

	/**
	 * wrapSSL()
	 * wrap into ssl messages & send it
	 * @param bb original <code>ByteBuffer</code> to be encoded
	 * @return number of bytes transmitted count from original buffer 
	 * @throws ThincException
	 */
	private int wrapSSL(ByteBuffer bb) throws ThincException {
		SSLEngineResult result;
		int numBytes = bb.limit();
//		System.out.println("bb size:"+bb);
		while(bb.hasRemaining()) {
			try {
//				canvas.repaint(30);
				canvas.drawThincCursor(); // update cursor position
				//Thread.yield();
				while(true) {
					if(bb.hasRemaining() && clientAppData.hasRemaining()) clientAppData.put(bb.get());
					else break;
//					Thread.sleep(SLEEPING_DELAY);
				}
//				System.out.println("Phase 1: plainText filled up - "+clientAppData.position());
				clientAppData.flip();
				while(clientAppData.hasRemaining()) {
					result = sslEngine.wrap(clientAppData, clientPacData);
//					System.out.println("Phase 2: cipherText filled up");
//					System.out.println(result);
//					System.out.println(clientPacData);
					if(result.getStatus() == SSLEngineResult.Status.OK) {
						clientPacData.flip();
						while(clientPacData.hasRemaining()) {
							sc.write(clientPacData);
						}
					
//						System.out.println("Phase 3: cipherText sent");
						clientPacData.clear();
					}
				}
				clientAppData.clear();
			} catch (SSLException e) {
				throw new ThincException("wrapSSL: cannot wrap bytes",e);
			} catch (IOException e) {
				throw new ThincException("wrapSSL: cannot send ciphertext into socketChannel",e);
			} /*catch (InterruptedException e) {
				throw new ThincException("wrapSSL: Thread interrupted",e);
			}*/
		}
//		System.out.println("Done?");
		return numBytes;
	}
	
	
	/**
	 * getMessage()
	 * 
	 * get messages from the server. 
	 * @param size <code>ByteBuffer</code> size to be received
	 * @return <code>ByteBuffer</code> received
	 */
	public ByteBuffer getMessage(int size) throws ThincException
	{
		/* allocate & initialize ByteBuffer. if previous one is suitable to receive message, 
		 * just use it. otherwise it should be created by ByteBuffer.allocate() */
		if (recv == null || size > recv.capacity()){
			if (size < MAX_DIRECT_ALLOCATE) recv = ByteBuffer.allocateDirect(size);
			else recv = ByteBuffer.allocate(size);
		} else 
		recv.clear();
		recv.limit(size);
		
		/* receive the messages using primitive recvBytes() method */
		try {
			recvBytes(recv);
		} catch (ThincException e) {
			throw new ThincException ("getMessage: Error on recvBytes",e);
		}
		recv.flip(); // flips the received message to use it later by ThincCMsg
		return recv;
	}
	
	/**
	 * recvBytes()
	 * 
	 * is the primitive method used by getMessage method. it also
	 * used when version exchange required (which does not have
	 * general thinc header 
	 * @param bb <code>ByteBuffer</code> to receive
	 * @return number of bytes received
	 * @throws ThincException
	 */
	protected int recvBytes(ByteBuffer bb) throws ThincException
	{
		int numBytes = 0;

		try {
			if(!isSSL) {
				while(bb.hasRemaining()) {// receive until ByteBuffer is full
						numBytes += sc.read(bb);
					}
			} else
				numBytes = unwrapSSL(bb);
			
		} catch (IOException e) {
			throw new ThincException("recvBytes: IOException",e);
		} catch (ThincException e) {
			throw new ThincException("recvBytes: SSLException",e);
		}
		return numBytes;
	}
	
	/**
	 * unwrapSSL()
	 * unwrap from ssl messages & send it
	 * @param bb original <code>ByteBuffer</code> to receive
	 * @return number of bytes received count from original buffer 
	 * @throws ThincException
	 */
	private int unwrapSSL(ByteBuffer bb) throws ThincException {

		//Phase 3: receive data if serverPacData is empty and decipher that and fill bb
		//Phase 3-1: make sure every serverPacData deciphered into serverAppData before finishing
		//Phase 4: left serverAppData and  after you finish sending 
		int numBytes = bb.limit();
//		System.out.println();
//		System.out.println("Phase 0: before doing something");
//		System.out.println("serverAppData: "+serverAppData); //ready to use
//		System.out.println("serverPacData: "+serverPacData); //ready to use
//		System.out.println("bb: "+bb);
//		System.out.println();
		
		//Phase 1: check if serverAppData is not empty, fill this into bb
		while(serverAppData.hasRemaining()) {
			if(bb.hasRemaining()) bb.put(serverAppData.get());
			else {
				serverAppData.compact();
				serverAppData.flip();
//				System.out.println("Exit 1: serverAppData hasRemaining");
//				System.out.println("serverAppData: "+serverAppData);
//				System.out.println();
				return numBytes;
			}
		}
		

		// ready to receive
		serverAppData.compact();
		serverPacData.compact();
		
		while(bb.hasRemaining()) {
			try {
				int num = 0;
//				canvas.drawThincCursor(); // update cursor position
//				canvas.repaint(30);
				//Phase 3-1: receive packet from the server
				//Thread.yield();
				
				//		if overflow, don't receive and try to unwrap it
				//		if ok, unwrap it in put it into bb until serverPacData is empty
				//		if first time, just receive and try to unwrap it
				//		if underflow, plz receive and try to unwrap it
				//		if closed, close it and throw exception

				// try if it is first time or serverPacData doesn't have Remaining
//				System.out.println("Phase 2: receive or not?");
//				System.out.println("serverAppData: "+serverAppData);
//				System.out.println("serverPacData: "+serverPacData);
					while(num == 0){
						if(rcvResult == null || rcvResult.getStatus() != SSLEngineResult.Status.BUFFER_OVERFLOW) 
							num = sc.read(serverPacData);
						else break;
					}

					serverPacData.flip(); //ready to use					
//					if(num != 0) {
//						System.out.println("Phase 2-1: received in SSL: "+num);
//						System.out.println("serverPacData: "+serverPacData);						
//					} else {
//						System.out.println("Phase 2-2: serverPacData has remaining.");						
//					}

//				System.out.println();

//				System.out.println("Phase 3: unwrap it");
				do {
					rcvResult = sslEngine.unwrap(serverPacData, serverAppData);
//					System.out.println("Phase 3-1: in Loop: "+rcvResult);
//					System.out.println("serverPacData: "+serverPacData);						
//					System.out.println("serverAppData: "+serverAppData);						
//					System.out.println();
				} while (rcvResult.getStatus() == SSLEngineResult.Status.OK &&
						serverPacData.hasRemaining());

//					System.out.println("Phase 3-2: PacData unwraped");
//					System.out.println("serverPacData: "+serverPacData);						
//					System.out.println("serverAppData: "+serverAppData);						
//					System.out.println("rcvResult:\n"+rcvResult);
//					System.out.println();

					serverPacData.compact(); // ready to receive						

//					System.out.println("Phase 4: processing unwrap");
					switch(rcvResult.getStatus()) {
					case OK:
//						System.out.println("Phase 4-1: OK: move serverAppData into bb");
						if(bb.hasRemaining()) serverAppData.flip(); // ready to use
						while(true) {
							if(serverAppData.hasRemaining() && bb.hasRemaining()) bb.put(serverAppData.get());
							else {
//								System.out.println("Phase 4-1-1: bb or serverAppData is full");								
//								System.out.println("serverAppData: "+serverAppData);						
//								System.out.println("bb: "+serverPacData);						
//								System.out.println();						
								break;
							}
//							Thread.sleep(SLEEPING_DELAY);
						}
//						System.out.println("Phase 4-1-2: before compact");
//						System.out.println("serverAppData: "+serverAppData);						
						serverAppData.compact(); // ready to receive
//						System.out.println("Phase 4-1-3: after compact");
//						System.out.println("serverAppData: "+serverAppData);						
//						System.out.println();
						break;
					case BUFFER_UNDERFLOW: // underflow -> try to receive data again
//						System.out.println("Phase 4-2: BUFFER_UNDERFLOW: fall through to recv it again");
//						System.out.println("serverPacData: "+serverPacData);						
//						System.out.println("serverAppData: "+serverAppData);						
//						System.out.println();
						canvas.drawThincCursor();
//						Thread.sleep(SLEEPING_DELAY);
						break;
					case BUFFER_OVERFLOW:
//						System.out.println("Phase 4-3: BUFFER_OVERFLOW: try to put serverAppData into buffer");
//						System.out.println("serverPacData: "+serverPacData);						
//						System.out.println("serverAppData: "+serverAppData);						
//						System.out.println("bb: "+bb);						
//						System.out.println();

//						System.out.println("Phase 4-3-1: Spend serverAppData as much as possible");
						if(bb.hasRemaining()) serverAppData.flip(); // ready to use
						while(true) {
							if(bb.hasRemaining() && serverAppData.hasRemaining()) bb.put(serverAppData.get());
							else //{
//								System.out.println("Phase 4-3-2: serverAppData or bb is full");								
//								System.out.println("serverAppData: "+serverAppData);						
//								System.out.println("bb: "+bb);						
//								System.out.println();
								break;
//							}
						}
						serverAppData.compact(); //ready to receive no matter it is filled or not
						break;
					default:
//						System.out.println("Phase 4-4: unknown including CLOSE: do not handle it for a while");
						break;
					}
			} catch (SSLException e1) {
				System.err.println(rcvResult);
				throw new ThincException("unwrapSSL: cannot convert cipherText into plainText",e1);				
			} catch (IOException e2) {
				throw new ThincException("unwrapSSL: cannot receive cipherText from the server",e2);
			} /*catch (InterruptedException e) {
				throw new ThincException("unwrapSSL: thread Interrupted",e);
			}*/
		}
		serverAppData.flip(); // ready to use
		serverPacData.flip(); // ready to use
//		System.out.println("Exit 2: bb is good now");
//		System.out.println("serverPacData: "+serverPacData);						
//		System.out.println("serverAppData: "+serverAppData);						
//		System.out.println("bb: "+bb);						
//		System.out.println("----------------------------------");
		return numBytes;
	}
	
	
	
	/**
	 * getLastSent() get the thinc type of last sent messages
	 * @return  a thinc type of last sent messages
	 */
	public byte getLastSent () { return lastSent; }

	/**
	 * setSSLSocketChannel()
	 * 
	 * uses <code>SSLEngine</code> class in Java to establish SSL connection.
	 * <code>SSLEngine</code> supports only non-blocking mode.
	 * We need to find another way to use <code>SSLEngine</code> because 
	 * it is currently implemented with polling
	 * @throws ThincException
	 */
	public void setSSLSocketChannel(String host, int port) throws ThincException{
		try {
			/* create TrustManage for certification*/
			TrustManager[] trustAllCerts = new TrustManager[] {
				new X509TrustManager() {
					public java.security.cert.X509Certificate[] getAcceptedIssuers() {
						return null;
					}
					public void checkClientTrusted(java.security.cert.X509Certificate[] certs, String authType) {}
					public void checkServerTrusted(java.security.cert.X509Certificate[] certs, String authType) {}
				}
			};
			/* create SSLContext, initialize, and get the SSLSocket from there */
			SSLContext sslContext = SSLContext.getInstance("SSL");
			sslContext.init(null, trustAllCerts, new SecureRandom());
			sslEngine = sslContext.createSSLEngine(host, port);
//			sslSocket = (SSLSocket) sslContext.getSocketFactory().createSocket(sc.socket(),client.getHost(), client.getPort(), true);
			/* register HandshakeCompletedListener for this */
//			sslSocket.addHandshakeCompletedListener(this);
			System.out.println("SSL Created:"+sslEngine.getPeerHost()+"-"+sslEngine.getPeerPort());
			/* do handshake */
//			sslSocket.startHandshake();
//			sslEngine.beginHandshake();
			sslEngine.setUseClientMode(true);
			SSLHandshake();
			
			clientAppData = ByteBuffer.allocateDirect(sslEngine.getSession().getApplicationBufferSize());
			clientPacData = ByteBuffer.allocateDirect(sslEngine.getSession().getPacketBufferSize());
			serverAppData = ByteBuffer.allocateDirect(sslEngine.getSession().getApplicationBufferSize());
			serverPacData = ByteBuffer.allocateDirect(sslEngine.getSession().getPacketBufferSize());

			serverAppData.limit(serverAppData.position());
			serverPacData.limit(serverPacData.position());
			/* wait until completed*/
//			while (!sslHandshakeCompleted) {
//				Thread.sleep(100);
//			}
			
			/*set itself as SSL secured channel */
			isSSL = true;
		} catch (NoSuchAlgorithmException e) {
			throw new ThincException("cannot establish SSL connection",e);
		} catch (KeyManagementException e) {
			throw new ThincException("cannot establish SSL connection",e);
//		} catch (SSLException e) {
//			throw new ThincException("cannot establish SSL connection",e);
//		} catch (InterruptedException e) {
//			throw new ThincException("cannot establish SSL connection",e);
//		} catch (IOException e) {
//			throw new ThincException("cannot establish SSL connection",e);
		}
		
	}
	/**
	 * SSLHandshake()
	 * 
	 * do handshake for <code> SSLEngine</code>.
	 * @throws ThincException 
	 */
	 public void SSLHandshake() throws ThincException {
		try {
			sc.configureBlocking(false);
			sslEngine.beginHandshake();
		} catch (SSLException e1) {
			throw new ThincException("SSLHandshake: beginHandshake error",e1);
		} catch (IOException e) {
			throw new ThincException("SSLHandshake: making non-blocking I/O",e);
		}
		HandshakeStatus status = sslEngine.getHandshakeStatus();
		ByteBuffer dummy = ByteBuffer.allocate(sslEngine.getSession().getApplicationBufferSize());
		ByteBuffer netData = ByteBuffer.allocateDirect(sslEngine.getSession().getPacketBufferSize());
		SSLEngineResult result = null;
		while(true) {
			switch(status) {
			case NEED_TASK:
				Runnable task;
				while ((task = sslEngine.getDelegatedTask()) != null) {
					task.run();
				}
				break;
			case NEED_UNWRAP:
				try {
						sc.read(netData);
				} catch (IOException e) {
					throw new ThincException("SSLHandshake: cannot receive wrapped initial message",e);
				}
				netData.flip();
				try {
					do {
						result = sslEngine.unwrap(netData, dummy);
					} while(result.getStatus() == SSLEngineResult.Status.OK &&
							result.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.NEED_UNWRAP &&
							result.bytesProduced() == 0);
					
					if(dummy.position() == 0 && 
							result.getStatus() == SSLEngineResult.Status.OK && 
							netData.hasRemaining()) {
						result = sslEngine.unwrap(netData, dummy);
					}
					assert(result.bytesProduced() == 0);					
					if(result.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.FINISHED) {
						System.out.println("SSLHandshake: handshake complete!");
						return;
					}
				} catch (SSLException e) {
					throw new ThincException("SSLHandshake: cannot unwrap initial message",e);
				}
				netData.compact();
				break;
			case NEED_WRAP:
				try {
						result = sslEngine.wrap(dummy,netData);
					} catch (SSLException e) {
						throw new ThincException("SSLHandshake: cannot wrap initial message",e);
					}
				assert(result.bytesConsumed() == 0);
				assert(result.bytesProduced() != 0);
				netData.flip();
				try {
					while(netData.hasRemaining()) {
							sc.write(netData);
					}
				} catch (IOException e) {
					throw new ThincException("SSLHandshake: cannot send wrapped initial message",e);
				}
				netData.clear();
				break;
			case FINISHED:
				System.out.println("SSLHandshake: handshake complete!");
				return;
			case NOT_HANDSHAKING:
				throw new ThincException("SSLHandshake: In handshake, application should not be reached in this statement");
				//System.err.println("SSLHandshake: In handshake, application should not be reached in this statement");
				//throw new ThincException(status.toString());
				//status = sslEngine.getHandshakeStatus();
				//break;
			default:
				assert(true);
				break;
			}
//			if(status != null)System.out.println("before:"+status);
			status = sslEngine.getHandshakeStatus();
//			if(status != null)System.out.println("after:"+status);
		}
	}
}
