/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.SocketChannel;

/**
 * handles the general frame buffer / video / cursor / init messages from/to the server.  it recoginizes the message's type, flag and length then pass them into  the proper message object. Moreover, it also takes the control from client messages to send a message. in case of init_reply messages, it pass them into <Code>ThincMsgInitReplyHandler</code> object to process it.
 * @author  Young Jin Yoon
 * @see  ThincHandler
 */
public final class ThincMsgHandler extends ThincHandler{
	
	public static final int MAX_MESSAGE_TYPE = 150;
	
	protected ByteBuffer hd = null; // ByteBuffer object to receive header
//	protected ThincClientCanvas canvas; // pointer to canvas to draw

	private ThincMsgInitReplyHandler initHandler; // reply handler
	private ThincSMsg serverMessage[] = new ThincSMsg[MAX_MESSAGE_TYPE]; // array to contain ThincSMsg objects
	
	private byte type; //message type
	private byte flag; // message flag
	private short length; // message length

	/**
	 * create <code>ThincMsgHandler</code> object
	 * @param tc <code>ThincClient</code> object which invokes this
	 * @param c <code>ThincClientCanvas</code> object to draw
	 * @throws ThincException
	 */
	public ThincMsgHandler(ThincClient tc, ThincClientCanvas c) throws ThincException {
		super(tc,c);
		/* set the pointer for ThincClientCanvas and ThincMsgHandler each other */
//		canvas = c;
		canvas.setMsgHandler(this);
		
		/* allocating header ByteBuffer object */
		hd = ByteBuffer.allocateDirect(THINC_HEADERSIZE);
		hd.order(ByteOrder.BIG_ENDIAN);
		

		/* try to make the connection to the server */
		try {
			createSocketChannel(client.getHost(),client.getPort());
		} catch (ThincException e) {
			throw new ThincException("ThincMsgHandler: cannot connect to server",e);
		}
		/* creating ThincMsgInitReplyHandler object */
		initHandler = new ThincMsgInitReplyHandler(tc,c,this); 

		/* initializing serverMessage array by using initServerMessage() */
		initServerMessage();
	}
	/**
	 * initServerMessage()
	 *
	 * registers server message into array named serverMessage. when reply from server
	 * comes, the handler automatically choose the proper message and pass the control
	 * to it. if you want to add another server message, you can register the message 
	 * by adding the line of here.
	 */
	private void initServerMessage() {
		/* initialize all serverMessage element to null */
		for(int i=0;i < serverMessage.length;i++)
			serverMessage[i] = null;
		
		/* register proper serverMessage element with proper classes */
		serverMessage[ThincSMsgAck.MESSAGE_TYPE] = 					new ThincSMsgAck(this,canvas);
		serverMessage[ThincSMsgCursorChange.MESSAGE_TYPE] = 		new ThincSMsgCursorChange(this,canvas);
		serverMessage[ThincSMsgCursorColor.MESSAGE_TYPE] = 			new ThincSMsgCursorColor(this,canvas);
		serverMessage[ThincSMsgCursorMove.MESSAGE_TYPE] =			new ThincSMsgCursorMove(this,canvas);
		serverMessage[ThincSMsgCursorShowHide.MESSAGE_TYPE] = 		new ThincSMsgCursorShowHide(this,canvas);
		serverMessage[ThincSMsgFBCopy.MESSAGE_TYPE] = 				new ThincSMsgFBCopy(this,canvas);
		serverMessage[ThincSMsgFBFillBilevel.MESSAGE_TYPE] = 		new ThincSMsgFBFillBilevel(this,canvas);
		serverMessage[ThincSMsgFBFillGlyph.MESSAGE_TYPE] = 			new ThincSMsgFBFillGlyph(this,canvas);
		serverMessage[ThincSMsgFBFillPixMap.MESSAGE_TYPE] = 		new ThincSMsgFBFillPixMap(this,canvas);
		serverMessage[ThincSMsgFBFillSolid.MESSAGE_TYPE] = 			new ThincSMsgFBFillSolid(this,canvas);
		serverMessage[ThincSMsgFBRawUpdate.MESSAGE_TYPE] =  		new ThincSMsgFBRawUpdate(this,canvas);
		serverMessage[ThincSMsgInitSecCaps.MESSAGE_TYPE] =  		new ThincSMsgInitSecCaps(client,this,canvas);
		serverMessage[ThincSMsgInitSessCaps.MESSAGE_TYPE] =  		new ThincSMsgInitSessCaps(client,this,canvas,initHandler);
		serverMessage[ThincSMsgInitOK.MESSAGE_TYPE] =  				new ThincSMsgInitOK(client,this,canvas);
		serverMessage[ThincSMsgInitNotOK.MESSAGE_TYPE] =  			new ThincSMsgInitNotOK(client,this,canvas);
		serverMessage[ThincSMsgVideoEnd.MESSAGE_TYPE] = 			new ThincSMsgVideoEnd(this,canvas);
		serverMessage[ThincSMsgVideoMove.MESSAGE_TYPE] = 			new ThincSMsgVideoMove(this,canvas);
		serverMessage[ThincSMsgVideoNext.MESSAGE_TYPE] = 			new ThincSMsgVideoNext(this,canvas);
		serverMessage[ThincSMsgVideoResize.MESSAGE_TYPE] = 			new ThincSMsgVideoResize(this,canvas);
		serverMessage[ThincSMsgVideoScale.MESSAGE_TYPE] = 			new ThincSMsgVideoScale(this,canvas);
		serverMessage[ThincSMsgVideoStart.MESSAGE_TYPE] = 			new ThincSMsgVideoStart(this,canvas);
	}
	
	/**
	 * init()
	 * 
	 * receives version information from the server. try to compare the version
	 * then send client version information to the server it uses recvBytes() and
	 * sendBytes() methods because version information exchange should not have
	 * any general thinc header.
	 * @throws ThincException
	 */
	public void init() throws ThincException {
		byte target;
		/*recv version from server */
		recv = ByteBuffer.allocateDirect(ThincClient.CLIENT_VERSION.length);
		try {
			recvBytes(recv);
		} catch (ThincException e) {
			throw new ThincException("init: error on recvByte",e);
		}
		recv.flip();
		
		/* compare received version with client version */
		for(int i=0; i<ThincClient.CLIENT_VERSION.length; i++) {
			target = recv.get();
			if(target != ThincClient.CLIENT_VERSION[i]) throw new ThincException("init: error on version exchanges");
		}
		recv.clear();
		
		/* send client version to the server */
		recv.put(ThincClient.CLIENT_VERSION);
		recv.flip();
		try {
			sendBytes(recv);
		} catch (ThincException e) {
			throw new ThincException("init: error on recvByte",e);
		}
	}
	
	
	/**
	 * start()
	 * 
	 * actually starts to receive message from server. if it comes,
	 * this parses general thinc header and pass the control to the proper
	 * server message class.
	 * @throws ThincException
	 */
	public void start() throws ThincException {

		while(true) {// keep trying to receive message 
			try {
				recvBytes(hd);
			} catch (ThincException e) {
				throw new ThincException ("receiver: Error on recvBytes",e);
			}
			hd.flip();
			
			/* get the type, flag and length of the message */
			type = hd.get();
			flag = hd.get();
			length = hd.getShort();
			hd.clear();
			
			/* pass the control to the proper message*/
			if(type == ThincSMsgInitReply.MESSAGE_TYPE)
				initHandler.start();
			else if(type >= 0 && type < MAX_MESSAGE_TYPE && serverMessage[type] != null)
				serverMessage[type].process();
			else System.err.println("receiver: Unknown "+type+", ignored");
		}
	}
	
	/**
	 * send()
	 * 
	 * overriden to save last sent messages.
	 * @see ThincHandler
	 * @param c <code>ThincCMsg</code> classes to send
	 */
	@Override
	public int send(ThincCMsg c) throws ThincException{
		int retVal;
		retVal = super.send(c);
		lastSent = c.getMsgType(); // save last sent message in lastSent ( for ACK or NAK )
		return retVal;
	}
	/**
	 * setSSL()
	 *
	 * sets <code>ThincMsgInitReplyHandler</code> object as
	 * SSL secured handler.
	 */
	public void setSSL() {
		if (isSSL) {
			initHandler.isSSL = true;
			initHandler.sslEngine = sslEngine;
			initHandler.clientAppData = clientAppData;
			initHandler.clientPacData = clientPacData;
			initHandler.serverAppData = serverAppData;
			initHandler.serverPacData = serverPacData;
			initHandler.rcvResult = rcvResult;
		}
	}	
	/* general encapsulated methods */
	public ThincMsgInitReplyHandler getInitHandler() {return initHandler;}
	protected SocketChannel getSocketChannel()	{return sc;}
	public int getFlag () {return flag;}
	public int getLength (){return length;}

}
