/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;

/**
 * handles changing the cursor. it comes with new hot x and y coordinates,
 * and images/bitmaps for brand new cursor.
 * @author Young Jin Yoon
 * @see ThincSMsgCursor
 */
public final class ThincSMsgCursorChange extends ThincSMsgCursor {
	public static final int MESSAGE_TYPE = ThincProto.T_CURSOR_CHANGE;
	public static final int MESSAGE_SIZE = ThincProto.T_CURSOR_CHANGE_SIZE;
	public static final int FLAG_ARGB = ThincProto.T_CURSOR_ARGB; // flag for 'colorful' cursor :)
	
    /**
     * creates ThincSMsgCursorChange object.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param c <code>ThincClientCanvas</code> to get the information
	 */
	public ThincSMsgCursorChange(ThincMsgHandler h, ThincClientCanvas c) {
		super(h, c);
	}

	/**
	 * process()
	 * 
	 * get the new information about cursor such as hot x, hot y, size and image respectively,
	 * and set it into ThincClientCanvas object.
	 * @throws ThincException
	 */
	@Override
	public void process() throws ThincException {
		ByteBuffer bb;
		byte xhot, yhot;
		short size;
		int flag;
		int[] imageArray = null;
		
		/* get header message and parse it */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException ("process: error on Cursor Change",e);
		}
		xhot = bb.get(); //get hot x
		yhot = bb.get(); //get hot y
		size = bb.getShort(); // get size of image followed
	
		/* get image/bitmap */
		try {
			bb = handler.getMessage(size);
		} catch (ThincException e) {
			throw new ThincException ("process: error on Cursor Change",e);
		}
		
		flag = ((ThincMsgHandler)handler).getFlag();
		
		/* create cursor image by using createCursor method in ThincClientCanvas class */
		if((flag & FLAG_ARGB) > 0) { // if it is a 'colorful' cursor
			imageArray = new int[size/4];
			bb.asIntBuffer().get(imageArray);
			canvas.createCursor(imageArray, bb); 
				// actually it simply copies the contents of IntBuffer
		} else { //if it is bitmap cursor
			byte[] fgb, bgb;
			fgb = new byte[size/2]; // assumed bitmap for fg and bg should be the half of received message
			bgb = new byte[size/2]; // and, as you know, it should be!
			imageArray = new int[canvas.getCursorHeight() * canvas.getCursorWidth()];
			bb.get(fgb);
			bb.get(bgb);
			canvas.createCursor(imageArray,fgb,bgb); 
				// converts bitmap image returns IntBuffer filled with ABGR image
		}
		/* set created cursor onto canvas */
		canvas.setThincCursor(imageArray, (int)(xhot & 0xFF), (int)(yhot&0xFF));
	}

	
}
