/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;
import java.nio.IntBuffer;

/**
 * class for filling the regions with bitmap images to the frame buffer.
 * @author  Young Jin Yoon
 * @see  ThincSMsgFB
 */
public final class ThincSMsgFBFillBilevel extends ThincSMsgFB {
	public static final int MESSAGE_TYPE = ThincProto.T_FB_BILEVEL;
	public static final int MESSAGE_SIZE = ThincProto.T_FB_BILEVEL_SIZE;
	public static final int RECT_MESSAGE_SIZE = ThincProto.T_RECTREGION_SIZE;
	
	/* flag from thinc header */
	public static final int FLAG_CACHED = ThincProto.T_FB_BITMAP_CACHED;
	public static final int FLAG_ADDCACHE = ThincProto.T_FB_BITMAP_ADDCACHE;
	public static final int FLAG_RESIZED = ThincProto.T_FB_BITMAP_RESIZED;
	
	/* pointer for cache / resize operation */
	private ThincSMsgFlagCache cache;
	private ThincSMsgFlagResize resize;

	/**
     * creates ThincSMsgFBFillBilevel object.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param c <code>ThincClientCanvas</code> to get the information
	 */
	public ThincSMsgFBFillBilevel(ThincMsgHandler h, ThincClientCanvas c) {
		super(h, c);
		cache = new ThincSMsgFlagCache(h,c); // create object corresponding with cache flag 
		resize = new ThincSMsgFlagResize(h,c); // create object corresponding with resize flag
	}

	/**
	 * process()
	 * 
	 * get the source bitmap image from cache or network and try to generate
	 * RGB image with those. finally draw it into frame buffer using
	 * drawTile(), drawImage(), drawImageScaled() methods from 
	 * <code>ThincClientCanvas</code> class.
	 * @throws ThincException
	 */
	@Override
	public void process() throws ThincException {
		ByteBuffer bb;
		short numRects;
		short x, y, width, height;
		short dx,dy,dw,dh;
		int fg, bg;
		ByteBuffer contentWrap;
		byte[] content = null;
		int[] imageArray;

		/* get header message and parse it */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException ("process: error on FBPixmap",e);
		}
		x = bb.getShort(); // x to draw
		y = bb.getShort(); // y to draw
		width = bb.getShort(); // width to draw
		height = bb.getShort(); // height to draw
		fg = bb.getInt(); // foreground color to draw
		bg = bb.getInt(); // background color to draw
		numRects = bb.getShort(); // number of regions represented by thinc_rectRegion.
								  // if it is 0, just draw image from (x,y) to (x+w,y+h).

		int flag;
        /*calculating bitmap size in bytes */		
        int bitMapSizeInBytes = (int)(Math.ceil((double)width/8)) * height;
        /* get flag from ThincMsgHandler */
        flag = ((ThincMsgHandler)handler).getFlag();

        /* flag handling */
        /*FIXME: different from protocol design, resize comes first */
		if((flag & FLAG_RESIZED) > 0) resize.process(null); 
		//take out the useless resized flag message out of messages. different from protocol specificaion,
		// it should have the garbage value.
        if((flag & FLAG_CACHED) > 0) { // if cached, get the bitmap image from cache.
			content = (byte[])cache.getCache(ThincClientCanvas.BIT_CACHE);
		}
		if((flag & FLAG_ADDCACHE) > 0){ // if it is required to save, get the cache id information
			cache.setCacheId(ThincClientCanvas.BIT_CACHE);
		}
		if((flag & FLAG_CACHED) == 0) {
			//if it is not cached already, try to receive bitmap image from the network
			content = new byte[bitMapSizeInBytes];
			try {
				contentWrap = handler.getMessage(bitMapSizeInBytes);
			} catch (ThincException e) {
				throw new ThincException ("process: error on FBPixmap",e);
			}
			contentWrap.get(content);
			if((flag & FLAG_ADDCACHE) > 0) cache.addCache(content.clone());
		}
		/* convert ARGB image into ABGR which is suitable for JAVA BufferedImage object */
		fg = canvas.argbToAbgr(fg);
		bg = canvas.argbToAbgr(bg);

		/* create ABGR image using createBitMapImage in ThincClientCanvas */
		imageArray = new int[width *height];
		canvas.createBitMapImage(imageArray, content, null, fg, bg, width);
		
		/* if numRect is 0, draw directly from the information such as x,y,width,height */
		if(numRects == 0){
			IntBuffer imgWrap = IntBuffer.wrap(imageArray);
				canvas.drawImage(imgWrap,x,y,width,height, ThincClientCanvas.TYPE_BITMAP);
			return;
		}
		/* if numRect is not 0, set tile using width and height */
		canvas.setTile(0, 0, width, height, imageArray);
		/* get thinc_rectRegion and draw it using for loop */
		for(int i=0; i<numRects; i++) {
			try {
				bb = handler.getMessage(RECT_MESSAGE_SIZE);
			} catch (ThincException e) {
				throw new ThincException ("process: error on FBPixmap",e);
			}
			dx = bb.getShort(); // get x
			dy = bb.getShort(); // get y
			dw = bb.getShort(); // get width
			dh = bb.getShort(); // get height
			canvas.drawTile(dx, dy, dw, dh); //draw it!
		}
	}
}
