/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.awt.Rectangle;
import java.nio.ByteBuffer;

/**
 * class for resize flag message.
 * @author  Young Jin Yoon
 * @see ThincSMsgFlag
 * @see  ThincSMsgFBRawUpdate
 */
public final class ThincSMsgFlagResize extends ThincSMsgFlag {

	public static final int DATA_MESSAGE_SIZE = ThincProtoResize.T_RESZ_DATA_SIZE;
	
	private ThincMsgHandler handler;
	private ThincClientCanvas canvas;
	
	/* various resized information */
	private static short width;
	/* various resized information */
	private static short nWidth;
	private static short height;
	private static short nHeight;
	private static float 	fb_rateWidth = 1; // resized rate for width
	private static float 	fb_rateHeight = 1; // resized rate for height
	private static float	fb_inverseRateWidth = 1; // inverse value of resized rate for width
	private static float	fb_inverseRateHeight = 1; // inverse value of resized rate for height
	
	private static boolean isResized = false;
	private static boolean 	isProceed = true;	

	/**
     * creates ThincSMsgFlagResize object.
     * @param h <code>ThincMsgHandler</code> to handle the message
	 */
	public ThincSMsgFlagResize(ThincMsgHandler h, ThincClientCanvas c) {
		handler = h;
		canvas = c;
	}
	/**
	 * setResized()
	 * set resized width and height with original width and height.
	 * @param oldWidth original width
	 * @param oldHeight original height
	 * @param newWidth resized width
	 * @param newHeight resized height
	 */
	public static void setResized(int oldWidth, int oldHeight, int newWidth, int newHeight) {
		/* calculating various information about resizing */
		width = (short) oldWidth;
		height = (short) oldHeight;
		nWidth = (short) newWidth;
		nHeight = (short) newHeight;
		fb_rateWidth = (float)newWidth / (float)oldWidth;
		fb_rateHeight = (float)newHeight / (float)oldHeight;
		fb_inverseRateWidth = (float)oldWidth / (float)newWidth;
		fb_inverseRateHeight = (float)oldHeight / (float)newHeight;

		/* if it is resized, turn on the resized flag. if not, turn it off */
		if(width == newWidth && height == newHeight) isResized = false;
		else isResized = true;
	}

	/**
	 * process()
	 * receives resized flag message from the network. if hasInformation is true,
	 * it compares the resized flag message with resized height and width of frame
	 * buffer. if it is false, just discard resized flag message
	 * @param r Rectangle which will be updated (in <code> ThincSMsgRawUpdate </code> only). null if not required.
	 * @throws ThincException
	 */
	public void process(Rectangle r) throws ThincException
	{
		ByteBuffer hd;
		short width, height;
		int size = 0;
		/* parse the header of resized flag message */
		try {
			hd = handler.getMessage(DATA_MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException("resize: error on process",e);
		}
		width = hd.getShort(); // width of resized frame buffer
		height = hd.getShort(); // height of resized frame buffer
		size = hd.getInt(); // size of resized frame buffer (useless)

		/* if received message has resized frame buffer information,
		   compare it with resized frame buffer info that client has. */
		if(r != null) {
			if(nWidth != width || nHeight != height) {
				isProceed = false;
			} else {
				isProceed = true;
			}
		}
	}
	
	/* encapsulated method to get isProceed or isResized flag */
	public boolean isProceed() { return isProceed; }
	public void isProceed(boolean value) { isProceed = value; }
	public static boolean isResized() { return isResized; }
	
	/* encapsulated or simple method to get the resized width, height, and coordination */
	public static short getResizedX(short x) { return (short)((double) (nWidth * x) / (double)width); } 
	public static short getResizedY(short y) { return (short)((double) (nHeight * y) / (double)height); } 
	public static short getResizedWidth(short x) { 
		short tempX;
		tempX = (short)((double) (nWidth * x) / (double)width); 
		if(tempX <= 0) return 1;
		else return tempX;
	} 
	public static short getResizedHeight(short y) { 
		short tempY;
		tempY = (short)((double) (nHeight * y) / (double)height); 
		if(tempY <= 0) return 1;
		else return tempY;
	} 
	public static short getOriginalX(short x) { return (short)(fb_inverseRateWidth * (float)x); } 
	public static short getOriginalY(short y) { return (short)(fb_inverseRateHeight * (float)y); } 
	public static double getRateHeight() { return fb_rateHeight; }
	public static double getRateWidth() { return fb_rateWidth; }
}
