/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;


/**
 * class for initializing cursor's image, hot x and hot y.
 * @author Young Jin Yoon
 * @see ThincSMsgInitReply
 */
public final class ThincSMsgInitReplyCursor extends ThincSMsgInitReply {

	public static final int INIT_MESSAGE_TYPE = ThincProto2.T_REQ_CURSOR; 
	public static final int MESSAGE_SIZE = ThincProto2.T_REPLY_CURSOR_SIZE;
	
	/* flag for ARGB cursor */
	public static final int CURSOR_ARGB = ThincProto.T_CURSOR_ARGB;

	/**
     * creates ThincSMsgInitReplyCursor object.
     * @param tc <code>ThincClient</code> to get the information for handshaking.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param cs <code>ThincClientCanvas</code> to get the information
	 */
	public ThincSMsgInitReplyCursor(ThincClient tc, ThincMsgHandler h, ThincClientCanvas cs) {
		super(tc, h, cs);
	}

	/**
	 * process()
	 * gets the information for cursor and set it using createCursor() and setThincCursor()
	 * method from <code>ThincClientCanvas</code> class.
	 */
	@Override
	public void process() throws ThincException {

		ByteBuffer bb;
		int[] imageArray;
		short flags, size;
		byte width, height, hotx, hoty;
		short x, y;
		int fg,bg;
		/* get and parse the cursor information from the network */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		}catch (ThincException e) {
				throw new ThincException("process: Error on getMessage in Cursor",e);
		}
		flags = bb.getShort(); // is ARGB cursor or bitmap cursor?
		size = bb.getShort(); // size of followed image
		width = bb.get(); // width of cursor
		height = bb.get(); // height of cursor
		hotx = bb.get(); // x coordinate of hot spot
		hoty = bb.get(); // y coordinate of hot spot
		x = bb.getShort(); // x location of initial cursor
		y = bb.getShort(); // y location of initial cursor
		fg = bb.getInt(); // foreground color for bitmap image
		bg = bb.getInt(); // background color for bitmap image

		/* convert ARGB color into ABGR color which is suitable to BufferedImage class */
		fg = canvas.argbToAbgr(fg) | 0xff000000;
		bg = canvas.argbToAbgr(bg) | 0xff000000;
					
		/* try to get the image */
		
		try {
			bb = handler.getMessage(size);
		}catch (ThincException e) {
			throw new ThincException("process: Error on CursorData in Cursor",e);
		}
		imageArray = new int[width * height];
		/* create cursor image by using createCursor method in ThincClientCanvas class */
		if((flags &CURSOR_ARGB)>0){ // if it is a 'colorful' cursor
			canvas.createCursor(imageArray,bb);
				// actually it simply copies the contents of IntBuffer
		} else { //if it is bitmap cursor
			byte[] fgb, bgb;
			fgb = new byte[size/2];// assumed bitmap for fg and bg should be the half of received message
			bgb = new byte[size/2];// and, as you know, it should be!
			bb.get(fgb);
			bb.get(bgb);
			canvas.createCursor(imageArray,fgb,bgb,fg,bg,width);
			// converts bitmap image returns IntBuffer filled with ABGR image
		}
		/* set created cursor onto canvas: 
		 *  Note that convert byte to unsigned int using bitwise operation */
		canvas.setThincCursor(imageArray, (int)(hotx&0xFF), (int)(hoty&0xFF), width, height);

		/* move the cursor onto initial position:
		 *  Note that convert short to unsigned int using bitwise operation */
		try {
			canvas.moveThincCursor((int)(x&0xFFFF), (int) (y&0xFFFF));
		} catch (ThincException e) {
			throw new ThincException("process: ThincSMsgInitReplyCursor - cannot set cursor into position",e);
		}
	}

}
