/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.io.IOException;
import java.nio.ByteBuffer;
import javax.net.ssl.SSLSocketFactory;
/**
 * class for init reply session secirity capability message from the server. it is actually ACK for <code>ThincCMsgInitSecCaps</code> from the server.
 * @author  Young Jin Yoon
 * @see  ThincSMsgInit
 */
public final class ThincSMsgInitSessCaps extends ThincSMsgInit {
	public static final int MESSAGE_TYPE = ThincProto2.T_SESS_SEC_CAPS;
	public static final int MESSAGE_SIZE = ThincProto2.T_SESS_SEC_CAPS_SIZE;

	ThincMsgInitReplyHandler initHandler;

	/**
     * creates ThincSMsgInitSessCaps object.
     * @param tc <code>ThincClient</code> to get the information for handshaking.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param cs <code>ThincClientCanvas</code> to get the information
	 */	
	public ThincSMsgInitSessCaps(ThincClient tc, ThincMsgHandler h, ThincClientCanvas cs,	ThincMsgInitReplyHandler in) {
		super(tc, h, cs);
		initHandler = in;
	}

	/**
	 * process()
	 * 
	 * gets the session security capability from the server and 
	 * trys to send next message by creating proper <code>ThincCMsg</code>
	 * object or using sendNextMessage() method from <code>ThincMsgInitReplyHandler</code>
	 * class. because SSL seems not to be implemented on the server side, it is simply ignored
	 * just for now.
	 * @throws ThincException
	 */
	@Override
	public void process() throws ThincException {
		int sess;
		ByteBuffer bb = null;
		/* receiving session security capabilities */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException("process: in SessCap",e);
		}
		sess = bb.getInt();
		/* if encrypted connection is required, conduct SSL handshake here.*/
		if ((sess & ThincClient.SEC_ENC) > 0) {
			System.out.print("Handshaking SSL...");
			try {
				handler.setSSLSocketChannel(client.getHost(),client.getPort());
			} catch(ThincException e) {
				throw new ThincException("process: in SessCaps, cannot establish SSL connection",e);
			}
			((ThincMsgHandler)handler).setSSL();
			System.out.println("Complete!");
		}
		/* if authentication is required, conduct authentication here.*/
		if ((sess & ThincClient.SEC_AUTH) > 0) {
			/* creates and invokes process() method in ThincCMsgInitAuth object */
			ThincCMsgInitAuth auth = new ThincCMsgInitAuth(client,(ThincMsgHandler)handler,canvas);
			try {
				auth.process();
			} catch (ThincException e) {
				throw new ThincException("process: while processing ThincCMsgInitAuth",e);
			}
		}
		/* if nothing requires for security, send first init request message by using
		 * sendNextMessage() method in ThincMsgInitReplyHandler class.*/
		if((sess &(ThincClient.SEC_AUTH | ThincClient.SEC_ENC)) == 0) {
			initHandler.sendNextMessage();
		}
	}


}
