/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

/**
 * class for creating and getting sound stream.
 * @author Young Jin Yoon
 * @see ThincSMsgSound
 */
public final class ThincSMsgSoundPlay extends ThincSMsgSound {
	public static final int MESSAGE_TYPE = ThincProtoSound.T_SND_PLAY;
	public static final int MESSAGE_SIZE = ThincProtoSound.T_SND_PLAY_SIZE;
	
	/**
     * creates ThincSMsgSoundPlay object.
     * @param h <code>ThincSoundHandler</code> to handle the message
 	 */
	public ThincSMsgSoundPlay(ThincSoundHandler h) {
		super(h);
	}

	/**
	 * process()
	 * 
	 * get the stream from the server and try to play it. if it is the
	 * first stream from the server, save the information for stream
	 * in the hash table.
	 * @throws ThincException
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void process() throws ThincException{
		ByteBuffer bb;
		SourceDataLine src;
		int size, timestamp;
		byte id;
		/* get the stream information from the network */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException("process: in ThincSMsgSoundPlay");
		}
		size = bb.getInt(); // size of sound message
		timestamp = bb.getInt(); // timestamp of message
		id = bb.get(); // stream identifier

		/* try to get the stream information */
		src = (SourceDataLine) streamId.get(id);
		/* if getting stream information failed, 
		 * create the stream information and set it into hash table */
		if(src == null) { 
            try {
				if (playInfo != null) 
					src = (SourceDataLine) AudioSystem.getLine(playInfo);
				else {
					format = new AudioFormat(FRAME_RATE,SAMPLE_SIZE,2,true,false);
			        playInfo = new DataLine.Info(SourceDataLine.class, format);	
					src = (SourceDataLine) AudioSystem.getLine(playInfo);
				}
            	src.open(format);
				src.start();
			} catch (LineUnavailableException e) {
				throw new ThincException("process: in ThincSMsgSoundPlay, Line is not available",e);
			}
			streamId.put((Object)id,(Object)src);
		}
		
		/* play sound by using playSound method */
		try {
			playSound(src, size, timestamp);
		} catch (ThincException e) {
			throw new ThincException("process: in ThincSMsgSoundPlay, play is not available",e);
		}
	}
	
	/**
	 * playSound()
	 * 
	 * play the sound by using write method from <code>SourceDataLine</code> object.
	 * @throws ThincException
	 */
	private void playSound(SourceDataLine src, int size, int timestamp) throws ThincException {
		ByteBuffer bb;
		byte[] sound = new byte[size];
		int offset = 0;
		/* get sound contents from the network */
		try {
			bb = handler.getMessage(size);
		} catch (ThincException e) {
			throw new ThincException("playSound: in ThincSMsgSoundPlay",e);
		}
		bb.get(sound);
		/* try to write it into SourceDataLine object to play the sound */
		try {
			while(offset < sound.length) {
				offset += src.write(sound, offset, sound.length-offset);
			}		
	    } catch (IllegalArgumentException e) {
	        try { // reopen the stream if error occurs.
	            System.err.println("Error occured LINE RESTARTED!!");
	            src = (SourceDataLine) AudioSystem.getLine(playInfo);
	            src.open(format);
	            src.start();
	        } catch (Exception f) { // throw ThincException if error occurs again.
	            throw new ThincException("playSound: in ThincSMsgSoundPlay",e);
	        }
	    }
	}
}
