/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;

/**
 * class for starting the video stream.
 * @author Young Jin Yoon
 * @see ThincSMsgVideo
 */
public final class ThincSMsgVideoStart extends ThincSMsgVideo {
	public static final int MESSAGE_TYPE = ThincProtoVideo.T_VIDEO_START;
	public static final int MESSAGE_SIZE = ThincProtoVideo.T_VIDEO_START_SIZE;
	public boolean accept;
	
	/**
     * creates ThincSMsgVideoStart object.
     * @param h <code>ThincSoundHandler</code> to handle the message
     * @param c <code>ThincClientCanvas</code> to get the information
 	 */
	public ThincSMsgVideoStart(ThincMsgHandler h, ThincClientCanvas c) {
		super(h, c);
	}

	/**
	 * process()
	 * 
	 * gets the initialized information for stream id from the server, 
	 * then creates and sets them onto the stream information vector. 
	 * 
	 * @throws ThincException
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void process() throws ThincException {
		ByteBuffer bb;
		int id, fmt_id;
		short x, y, width, height, dst_width, dst_height;
		boolean fmtInHS = false;
		ThincVideoStream target;
		/* gets the initialized information for stream id from the network */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException("process: in ThincSMsgVideoStart",e);
		}
		id = bb.getInt(); // stream id to create
		fmt_id = bb.getInt(); // format identifier
		x = bb.getShort(); // x for stream
		y = bb.getShort(); // y for stream
		width = bb.getShort(); // original width for stream
		height = bb.getShort(); // original height for stream
		dst_width = bb.getShort(); // resized width for stream
		dst_height = bb.getShort(); // resized height for stream
		
		/* make and set a new video stream information */
		target = new ThincVideoStream();
		target.x = x;
		target.y = y;
		target.width = width; 
		target.height = height;
		target.dst_width = dst_width;
		target.dst_height = dst_height;

		/* double check server format is supported in this client */
		if(ThincVideoFormat.finalFormats != null) {
			for(int i=0; i<ThincVideoFormat.finalFormats.length; i++) {
				if(ThincVideoFormat.finalFormats[i] == fmt_id) fmtInHS = true;
			}
		}
		/* if supported, create a proper ThincVideoFormat derived object */
		try {
			switch(fmt_id) { 
			case YV12:
				target.format = new ThincVideoFormatYV12((ThincMsgHandler)handler,width,height);// message received in constructor
				accept = true;
				break;
			case UYVY:
				target.format = new ThincVideoFormatUYVY((ThincMsgHandler)handler,width,height);// message received in constructor
				accept = true;
				break;
			case YUY2:
				target.format = new ThincVideoFormatYUY2((ThincMsgHandler)handler,width,height);// message received in constructor
				accept = true;
				break;
			default:
				accept = false;
				break;
		
			}
		} catch(ThincException e) {
			throw new ThincException("process: in ThincSMsgVideoStart",e);
		}
		if(!fmtInHS) accept = false;
		
		if(!accept) throw new ThincException("ThincSMsgVideoStart: unsupported video format"+Integer.toString(fmt_id));
		/* add stream information into vector */
		streamId.add(id, (Object)target);
		/**
		 * NOTE: startOK message for Thinc video is not used anymore.
		 */
	}
}
