/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.IntBuffer;
/**
 * abstract class for video format used in <code>ThinSMsgVideo</code>. 
 * 
 * @author Young Jin Yoon
 */
public abstract class ThincVideoFormat {
	public static final int YV12 = ThincProtoVideo.T_FOURCC_YV12;
	public static final int YUY2 = ThincProtoVideo.T_FOURCC_YUY2;
	public static final int UYVY = ThincProtoVideo.T_FOURCC_UYVY;
	public static final int MAX_FORMAT = 3;
	
	private static final int[] clientFormats = {YV12,YUY2,UYVY}; // client supported formats
	private static int[] serverFormats = null; // server supported formats, will be filled
	public static int[] finalFormats = null; // final formats both client and server supported

	private static int current = 0; // pointer for top of the filled server formats

	/**
	 * setFormat()
	 * 
	 * sets the format from the server into <code>ThincVideoFormat</code> class.
	 * @param id format id to set
	 */
	public static void setFormat(int id) {
		if(serverFormats == null) serverFormats = new int[MAX_FORMAT];
		for(int i=0; i<clientFormats.length; i++) {
			if(clientFormats[i] == id) {
				serverFormats[current] = id;
				current++;
			}
		}
	}

	/**
	 * setFormat()
	 * 
	 * gets an array of final format which both server and client accepted from 
	 * <code>ThincVideoFormat</code> class.
	 * @return int array for final format.
	 */
	public static int[] getFormat()	{
		int counter = 0;
		
		finalFormats = new int[current];
		for(int i=0; i< current; i++) {
				finalFormats[counter] = serverFormats[i];
				counter++;
		}
		return finalFormats;
	}
	
	/**
	 * process()
	 * 
	 * receives YUV image and converts it into ARGB images. it is used in
	 * <code>ThinSMsgVideo</code> to draw ARGB image into frame buffer.
	 * @param width width of YUV image
	 * @param height height of YUV image
	 * @param size size of YUV image
	 * @return ARGB image represented as IntBuffer object
	 * @throws ThincException
	 */
	public abstract IntBuffer process(int width, int height,int size) throws ThincException;
	
	/**
	 * yuvToRgb
	 * primitive method which converts YUV pixel into ARGB pixel 
	 * @param y y value for source pixel
	 * @param u u value for source pixel
	 * @param v v value for source pixel
	 * @return ARGB pixel
	 */
	protected int yuvToRgb(byte y, byte u, byte v) {
		short r,g,b;
		/* make y,u,v as unsigned value */
		short uy = unsigned(y); 
		short uu = unsigned(u);
		short uv = unsigned(v);
		int retVal = 0;
		
		/* get the r,g,b value by using formulas below */
		r = (short)((298*(uy- 16) + 409*(uv-128) + 128) >> 8);
		g = (short)((298*(uy- 16) - 100*(uu-128) - 208*(uv-128) + 128) >> 8);
		b = (short)((298*(uy- 16) + 516*(uu-128) + 128) >> 8);

		/* if RGB is larger then 255 or smaller than 0, cut it out */ 
		if(r > 255) r = (short)255;
		if(r < 0) r = (short)0;
		if(g > 255) g = (short)255;
		if(g < 0) g = (short)0;
		if(b > 255) b = (short)255;
		if(b < 0) b = (short)0;
		
		/* set the r,g,b into one int type */
		retVal = 0xff;
		retVal = (retVal << 8) | b;
		retVal = (retVal << 8) | g;
		retVal = (retVal << 8) | r;
		return retVal;
	}

	private short unsigned(byte src) {
		return (short) (src & 0xFF);
	}
}
