/* Copyright 2005 Ricardo Baratto */
/* Copyright 2005 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Instrumentation code.
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

#include "thinc_client.h"
#include <string.h>
#include <errno.h>
#include <stdio.h>

#ifdef TC_INSTRUMENT
/* useful when your cpu speed is not what cpuinfo says. this may happen if
 * for example your cpu has speedstep (although if your cpu speed is 
 * changing all the time then you're kinda screwed) */
long double get_cpu_speed_measure()
{
	instr_vars(first, second);
	const int howlong = 2;
	long double speed;

	tclient_msg("\n");
	tclient_msg("measuring cpu speed empirically.\nthis should take "
		"no longer than %d seconds\n", howlong);

	gettime_tsc(first);
	sleep(howlong);
	gettime_tsc(second);

	speed = second - first;
	speed /= howlong;

	/* do it in mhz */
	speed /= (1000*1000);

	tclient_msg("got it, it's %.3LfMHz. thank you for your patience\n",
		speed);

	return speed;
}

long double get_cpu_speed_proc()
{
	const char *func = __func__;
	const char *file = "/proc/cpuinfo";
	long double speed = 0;
	ssize_t l;
	size_t n = 0;
	char *line=NULL, *p;
	FILE *fp;

	fp = fopen(file, "r");
	if(!fp) {
		tclient_msg("%s: cannot read %s: %s\n", func, file,
			strerror(errno));
		return -1;
	}

	do {
		l = getline(&line, &n, fp);
		if(l == -1) {
			return -1;
		}

		if(l <= 7) goto next;

		if(!strncmp("cpu MHz", line, 7)) {
			p = strchr(line, ':');
			++p;
			speed = strtod(p, NULL);
			free(line);
			break;
		}
next:
		free(line);
		line = NULL;
	} while(!feof(fp));

	fclose(fp);

	return speed;
}

inline long double ticks_to_sec(long double speed, 
	unsigned long long ticks)
{
	return ticks/speed;
}
#endif	/* TC_INSTRUMENT */

#ifdef TC_TIMELINE
struct t_irecord* t_ievents = NULL;
unsigned long t_icount = 0;
unsigned long t_iwrap = 0;

/* where we print the timeline */
static const char *fn = "/tmp/client.timeline";

int tc_time_init()
{
	t_ievents = malloc(TC_INSTR_HOWMANY*sizeof(*t_ievents));
	if(!t_ievents)
		return TC_ERROR;

	return TC_OK;
}

void tc_time_end()
{
	free(t_ievents);
	t_icount = t_iwrap = 0;
}

void tc_time_print()
{
	register unsigned long i = 0;
	FILE *f = fopen(fn, "w");
	if(!f) {
		tclient_msg("Can't open timeline file %s: %s\n", fn,
			strerror(errno));
		return;
	}

	tclient_msg("\n===== dumping timeline to %s...", fn);

	time_print(f, "EVENTS\n");
	time_print_ev(f, TI_START);
	time_print_ev(f, TI_END);
	time_print_ev(f, TI_DRAW_START);
	time_print_ev(f, TI_DRAW_END);
	time_print_ev(f, TI_READ_START);
	time_print_ev(f, TI_READ_END);
	time_print_ev(f, TI_NET_START);
	time_print_ev(f, TI_NET_END);
	time_print(f, "EVENTS END\n");

	time_print(f, "TIMELINE %lu %lu\n", t_icount, t_iwrap);
	for( ; i<t_icount; ++i) {
		time_print_rec(f, i);
	}
	time_print(f, "TIMELINE END\n");
	fclose(f);
	tclient_msg(" done =====\n");
}

void tc_time_reset()
{
	tc_dbg_name(func)
	time_dbg(("%s: resetting instrumentation. had %lu records\n", 
		func, t_icount));
	t_icount = t_iwrap = 0;
}
#else						/* TC_TIMELINE */
int tc_time_init()
{
	return TC_OK;	/* noop */
}

void tc_time_end()
{
	/* noop */
}

void tc_time_print()
{
	/* noop */
}

void tc_time_reset()
{
	/* noop */
}

#endif 						/* TC_TIMELINE */
