/* Copyright 2005, 2007 Ricardo Baratto */
/* Copyright 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Event recording functions
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * Copyright 2004
 */

/* set tabstop=8 */

#ifdef TC_RECORD

#include "thinc_client.h"
#include <stdio.h>
#include <string.h>

int tc_record_init(struct tc_record *record, const char *file, int flags)
{
	tc_dbg_name(func)

	tclient_dbg(1, ("%s: using file %s to record events\n", func, 
		file));

	record->fp = fopen(file, "w");
	if(!record->fp) {
		tclient_msg("Error: Couldn't open %s to record events\n",
			file);
		return TC_ERROR;
	}

	record->flags = flags;

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

void tc_record_end(struct tc_record *record)
{
	if(!record->fp)
		return;
	fclose(record->fp);
}

void tc_record_mouse_click(struct tc_record *record, int which, int updown)
{
	unsigned long now;

	if(!(record->flags & TC_RECORD_MOUSE_CLICK))
		return;
	gettime_ms(now);
	fprintf(record->fp, "%c %lu %d %d\n", TC_REC_MCLICK, now, which, 
		updown);
}

void tc_record_mouse_move(struct tc_record *record, int x, int y)
{
	unsigned long now;
	if(!(record->flags & TC_RECORD_MOUSE_MOVE))
		return;

	gettime_ms(now);
	fprintf(record->fp, "%c %lu %d %d\n", TC_REC_MMOVE, now, x, y);
}

void tc_record_keyb(struct tc_record *record, int key, int updown)
{
	unsigned long now;
	if(!(record->flags & TC_RECORD_KEYB))
		return;
	gettime_ms(now);
	fprintf(record->fp, "%c %lu %d %d\n", TC_REC_KEYB, now, key, updown);
}

int tc_playback_init(struct tc_fb *fb, const char *file, int which)
{
	tc_dbg_name(func)

	tclient_dbg(1, ("%s: Playing back events from %s\n", func, file));

	fb->record.fp = fopen(file, "r");
	if(!fb->record.fp) {
		tclient_msg("Error: couldn't open events file %s\n", file);
		return TC_ERROR;
	}

	fb->record.flags = which;
	fb->record.ev = 0;
	fb->record.now = 0;

	return TC_OK;
}

int tc_playback_start(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int ret;

	tclient_dbg(1, ("%s: starting playback in 1 second\n", func));

	ret = fscanf(fb->record.fp, "%c %lu ", &fb->record.ev, &fb->record.now);
	if(ret == EOF || ret == 0) {
		tclient_msg("Error: empty or corrupted events file\n");
		fclose(fb->record.fp);
		return TC_ERROR;
	}

	/* set alarm to start playback */
	set_alarm(ALARM_RECORD, 1);

	return TC_OK;
}

void tc_playback_next(struct tc_fb *fb) 
{
	int ret;
	unsigned long next, diff;

again:
	switch(fb->record.ev) {
		case TC_REC_MCLICK:
			tc_playback_mouse_click(fb);
			break;
		case TC_REC_MMOVE:
			tc_playback_mouse_move(fb);
			break;
		case TC_REC_KEYB:
			tc_playback_keyb(fb);
			break;
		default:
			tclient_msg("Error: corrupted events file "
				"bailing out\n");
			fclose(fb->record.fp);
			return;
	}

	/* read next one */
	ret = fscanf(fb->record.fp, "%c %lu ", &fb->record.ev, &next);
	if(ret == EOF) {
		tclient_msg("Finished playing back events\n");
		return;
	}

	diff = next-fb->record.now;
	if(diff) {
		set_timer(ALARM_RECORD, 0, ms_to_us(diff), ret);
		if(ret == -1) {
			tclient_msg("Error: cannot set timer for next "
				"event: %s\n", strerror(errno));
			return;
		}
	} else
		goto again;

	/* ready for next one */
	fb->record.now = next;

	return;
}

void tc_playback_end(struct tc_fb *fb)
{
	fclose(fb->record.fp);
}

void tc_playback_mouse_click(struct tc_fb *fb)
{
	int which, updown;
	fscanf(fb->record.fp, "%d %d\n", &which, &updown);
	if(fb->record.flags & TC_RECORD_MOUSE_CLICK) {
		tclient_msg("button %d %s\n", which, 
			(updown == T_DOWN ? "down": "up"));
		tc_sendButtonEv(fb, which, updown);
	}
}

void tc_playback_mouse_move(struct tc_fb *fb)
{
	int x, y;
	fscanf(fb->record.fp, "%d %d\n", &x, &y);
	if(fb->record.flags & TC_RECORD_MOUSE_MOVE) {
		tclient_msg("mouse move to %d,%d\n", x, y);
#ifndef NO_X
		tc_x11cursorMove(fb, x, y);
#endif
		tc_sendMotionEv(fb, x, y);
	}
}

void tc_playback_keyb(struct tc_fb *fb)
{
	int key, updown;
	fscanf(fb->record.fp, "%d %d\n", &key, &updown);
	if(fb->record.flags & TC_RECORD_KEYB) {
		tclient_msg("key %d %s\n", key,
			(updown == T_DOWN ? "down": "up"));
		tc_sendKeyEv(fb, key, updown);
	}
}

#endif	/* TC_RECORD */
