/* Copyright 2005, 2007 Ricardo Baratto */
/* Copyright 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * tc_sync.c
 *
 * Functions for handling audio/video synchronization.
 *
 * Author: Leo Kim
 * Network Computing Lab - Columbia University
 */

#ifdef TC_USESYNC

#include "thinc_client.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* for strerror */
#include <errno.h>
#include <fcntl.h>	
#include <netinet/tcp.h>
#include <unistd.h>
#include <sys/poll.h>

/* for synchronization. represents delta for number of microseconds
 * difference between audio and video streams.
 * 1 millisecond = 1000 microseconds */
#define TC_TIMESTAMP_DELTA				90000
/* time to poll for additional data if in the process of sync'ing. 
 * this number is completely arbitrary. */
#define TC_TIMESTAMP_POLL_MAX_MS			25

/* For the synchronization functions, we take advantage of the timestamp 
 * protocol information for both sound and video to calculate the delta 
 * arrival difference between the sound and video data. We calculate this
 * delta as follows:
 *
 * delta = (sound-timestamp) - (video-timestamp)
 *
 * We use this delta handle the following general cases:
 *
 * - If delta > THRESHOLD, then sound timestamps are newer than video
 * timestamps. We must WAITSYNC SOUND and DROP VIDEO.
 * - If delta < THRESHOLD, then video timestamps are newer than sound
 * timestamps. We must DROP SOUND and WAITSYNC VIDEO.
 *
 * The behavior of these two cases are encapsulated in two functions, 
 * doSoundSync and doVideoSync.
 *
 * Dropping either sound or video is easy enough. We must still read this
 * data from the network, but we simply discard the data before we play it.
 * Pausing requires that the current media path to either buffer data or 
 * poll until additional data of the concomitant stream is received.
 *
 * A serious complication is that we do not necessarily know when a video 
 * stream starts or ends. This requires additional assumptions on sound
 * timestamps on the behavior of video timestamps. These assumptions are
 * noted throughout comments in this file.
 */

/* This function is used by tc_sound.c during playback to determine whether 
 * or not it should DROP or WAITSYNC. */
int tc_doSoundSync(struct tc_fb *fb, struct tc_snd *snd)
{
	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;
	int delta;

	/* :HACK: Assume there is only one audio stream and one video stream */
	stream = vid->streams;

	if (!stream || stream->timestamp == 0 || snd->timestamp == 0) {
		return TC_OK;
	}

	delta = snd->timestamp - stream->timestamp;	
#if 0
	tc_timestamp((stdout,"SOUND: video stamp: %u, sound stamp %u, "
		"delta %d\n", stream->timestamp, snd->timestamp, delta));
#endif

	if (delta > TC_TIMESTAMP_DELTA || delta < -TC_TIMESTAMP_DELTA) {
		/* sound timestamps are newer than video. WAITSYNC SOUND. */
		tclient_dbg(1, ("waitsync sound\n"));
		return delta;
	}

	return TC_OK;
}

int tc_delta_timestamp(struct tc_fb *fb)
{
	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;
	struct tc_snd *snd = &fb->snd;

	/* :HACK: Assume there is only one audio stream and one video stream */
	stream = vid->streams;

	return snd->timestamp - stream->timestamp;	
}

int tc_delta_recvstamp(struct tc_fb *fb)
{
	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;
	struct tc_snd *snd = &fb->snd;

	/* :HACK: Assume there is only one audio stream and one video stream */
	stream = vid->streams;

	return snd->instr_stamps.recvstamp - stream->instr_stamps.recvstamp;	
}

/* This function is used by xvideo.c during video playback to determine whether 
 * or not it should DROP or WAITSYNC. */
int tc_doVideoSync(struct tc_fb *fb, struct tc_video_stream *stream)
{
	struct tc_snd *snd = &fb->snd;
	int delta;

	/* If sound is not configured or disabled, then we don't bother */
	if (!snd->pcm || !fb->use_sound || snd->timestamp == 0 
		|| stream->timestamp == 0) 
	{
		return TC_OK;
	}

	delta = snd->timestamp - stream->timestamp;	
#if 0
	tc_timestamp(("VIDEO: video stamp: %u, sound stamp %u, delta %d\n", 
		stream->timestamp, snd->timestamp, delta));
#endif
	if (delta < -TC_TIMESTAMP_DELTA) {
		/* video timestamps are newer than sound. WAITSYNC VIDEO. */
		tclient_dbg(1, ("waitsync video\n"));
		return delta;
	}

	return TC_OK;
}

int tc_waitSyncSound(struct tc_fb *fb, struct tc_snd *snd, int delta)
{
	const char *func = __func__;
	struct pollfd fd[1];
	int res;

	while (1) {
		/* check if video is playing. res should be a negative number,
		 * and the sum of TC_TIMESTAMP_DELTA + delta should never be
		 * less than zero. */
		delta /= 1000;
		/* :HACK: we don't want to poll for video for more than the 
		 * timestamp delta itself. */
		if (delta > TC_TIMESTAMP_POLL_MAX_MS) {
			delta = TC_TIMESTAMP_POLL_MAX_MS;
		} else if (delta < 0) {
			delta = 0;
		}
		tclient_dbg(3, ("%s: polling for %u ms.\n", func, delta));
		/* waitsync video */
		/* poll on video data */
		fd[0].fd = fb->sock;
		fd[0].events = POLLIN;
		fd[0].revents = 0;
		res = poll(fd, 1, delta);
		if (fd[0].revents & (POLLHUP | POLLNVAL | POLLERR)) {
			tclient_msg("Error on server connection while "
				"synchronizing sound\n");
			return TC_ERROR;
		}
		if ( !(fd[0].revents & POLLIN) ) {
			break;
		}
		/* process the message */
		tclient_dbg(3, ("%s: processing handleMsg\n", func));
		res = tc_handleMsg(fb);
		tclient_dbg(3, ("%s: done processing handleMsg\n", func));
		if(res == TC_ERROR) {
			tclient_msg("Error handling server message\n");
			return TC_ERROR;
		} else if(res == TC_CLOSED) {
			tclient_msg("Server has closed the connection\n");
			return TC_ERROR;
		}
		res = tc_doSoundSync(fb, snd);

		if (res == TC_OK) {
			tclient_msg("%s: finished\n", func);
			break;
		}

		delta = res;
		tclient_dbg(1, ("%s: polling again based on "
			"delta %d.\n", func, delta));
	}

	return TC_OK;
}

int tc_waitSyncVideo(struct tc_fb *fb, struct tc_video_stream *stream, 
	int delta)
{
	tc_dbg_name(func)
	struct pollfd fd[1];
	int res;

	while (1) {
		/* check if sound is playing. res should be a negative number,
		 * and the sum of TC_TIMESTAMP_DELTA + res should never be less
		 * than zero. */
		delta /= 1000;
		/* :HACK: we don't want to poll for sound for more than the 
		 * timestamp delta itself. */
		if (delta > TC_TIMESTAMP_POLL_MAX_MS) {
			delta = TC_TIMESTAMP_POLL_MAX_MS;
		} else if (delta < 0) {
			delta = 0;
		}
		tclient_dbg(3, ("%s: Doing waitSyncVideo, polling for "
			"%u ms.\n", func, delta));
		/* waitsync video */
		/* poll on sound data */
		fd[0].fd = fb->snd.sock;
		fd[0].events = POLLIN;
		fd[0].revents = 0;
		res = poll(fd, 1, delta);
		if (fd[0].revents & (POLLHUP | POLLNVAL | POLLERR)) {
			tclient_msg("Error on tsd connection while "
				"synchronizing video\n");
			return TC_ERROR;
		}

		if ( !(fd[0].revents & POLLIN)) {
			break;
		}

		res = tc_handleSound(fb);
		if (res == TC_ERROR) {
			tclient_msg("Error handling tsd data while "
				"synchronizing video.\n");
			return TC_ERROR;
		} else if (res == TC_CLOSED) {
			tclient_msg("tsd has closed the connection while "
				"synchronizing video.\n");
			return TC_ERROR;
		} else if (res == TC_WOULDBLOCK) {
			break;
		}

		res = tc_doVideoSync(fb, stream);
		if (res == TC_OK) {
			tclient_dbg(1, ("%s: finished\n", func));
			break;
		}
		delta = -res;
		tclient_dbg(3, ("%s: polling again\n", func));
	}

	/* set the didSync flag so sound doesn't block forever. */
	fb->snd.didSync = T_TRUE;
	return TC_OK;
}

void tc_resetVideoTimeStamp(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_video *vid = &fb->vid;
	struct tc_video_stream *stream = vid->streams;
	if (stream) {
		tclient_dbg(3, ("%s: Resetting video stamp. Was %d\n", func,
			stream->timestamp));
		stream->timestamp = 0;
		stream->instr_stamps.recvstamp = 0;
		stream->instr_stamps.playstamp = 0;
	}
}

#endif /* TC_USESYNC */
