/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * "Simple" THINC Client
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * Copyright 2002, 2003, 2004, 2007
 */

/* set tabstop=8 */

#include "thinc_client.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <netinet/in.h>

#include <sys/time.h>
#include <sys/types.h>

#include <getopt.h>

#define TIMEOUT		1000

static int tc_init_compress(struct tc_fb *fb);
static void tc_end_compress(struct tc_fb *fb);
static void tc_stats_init(struct tc_fb *fb);
static void tc_stats_reset(struct tc_fb *fb);
static void tc_stats_print(struct tc_fb *fb);
static int tc_cache_init(struct tc_fb *fb);
static void tc_cache_destroy(struct tc_fb *fb);
static int tc_parse_window_size(const char *optarg);

static void tc_instrument_init(struct tc_fb *fb);
static void tc_instrument_reset(struct tc_fb *fb);
static void tc_instrument_print(struct tc_fb *fb);

#ifdef EXIT_FAILURE
#define TC_EXIT_FAIL	EXIT_FAILURE
#else
#define TC_EXIT_FAIL	1
#endif

#ifdef EXIT_SUCCESS
#define TC_EXIT_OK	EXIT_SUCCESS
#else
#define TC_EXIT_OK	0
#endif

/* the main client object. over here so signal handlers can access it.
 * keeps getting uglier... */
static struct tc_fb fb;

int tc_debug_level = TC_DEBUG_LEVEL;

/* signal handlers */
static unsigned char gotsig=T_FALSE;

static void sighandler(int sig)
{
	tclient_msg("Sad: Got signal, exiting\n");
	gotsig = T_TRUE;
}

static int alarm_got=T_FALSE;
int alarm_pending[ALARM_LAST];

#define alarm_init() do {						\
	int __i;							\
	for(__i=0; __i<ALARM_LAST; ++__i)				\
		alarm_pending[__i] = T_FALSE;				\
	alarm_got = T_FALSE;						\
} while(0)

void sigalarm(int sig)
{
	alarm_got = T_TRUE;
	signal(SIGALRM, sigalarm);
}

static void stats(int sig)
{
	signal(SIGUSR1, stats);
	/* print and reset everything */
	tc_stats_print(&fb);
	tc_stats_reset(&fb);
	tc_instrument_print(&fb);
	tc_instrument_reset(&fb);
	tc_time_print();
	tc_time_reset();
}

int print_msg;

#define error_msg(fb, e) 		\
	if(print_msg) {		\
		tclient_msg e;		\
	}

static void usage(const char *progname);

/**** file descriptor management *****/

/* this dude hier is the controller of all file descriptors. we call it the
 * event-er */
struct {
	struct event_handler events[FD_LAST];
	fd_set read_fds;
	fd_set write_fds;
	int maxfd;
} eventer;

static inline void init_events(void)
{
	int i;
	memset(&eventer, 0, sizeof(eventer));
	for(i=0; i<FD_LAST; ++i)
		eventer.events[i].fd = -1;
}

inline void add_fd(int fd, int which, unsigned int rw, 
	int (*got)(struct tc_fb*), int (*pending)(struct tc_fb*))
{
	struct event_handler *ev = &eventer.events[which];

	ev->fd = fd;
	ev->got = got;
	ev->pending = pending;
	if(rw & FD_READ)
		FD_SET(fd, &eventer.read_fds);
	if(rw & FD_WRITE)
		FD_SET(fd, &eventer.write_fds);
	if (fd > eventer.maxfd)
		eventer.maxfd = fd;
}

inline void remove_fd(int which)
{
	struct event_handler *ev = &eventer.events[which];

	if(ev->fd == -1)
		return;

	FD_CLR(ev->fd, &eventer.read_fds);
	FD_CLR(ev->fd, &eventer.write_fds);
	ev->fd = -1;
	ev->got = NULL;
	ev->pending = NULL;
	ev->call_me = 0;
}

#define set_fd_sets(rf, wf) do {					\
	memcpy(&rf, &eventer.read_fds, sizeof(rf));			\
	memcpy(&wf, &eventer.write_fds, sizeof(wf));			\
} while(0)

int main(int argc, char **argv)
{
	int xsock;

 	instr_vars(stotal=0, etotal, sany, eany, srun, erun);
 	instr_vars(spoll, epoll);

	int error = T_FALSE;
	int res;
	int did_x11, i, to_bed;
	fd_set rf, wf;
	struct event_handler *ev;
	int snd_port=0;
	
	int opt, index=0;
	struct option long_options[] = {
		/* name, has_arg, flag, val */
		{"help", 0, 0, 'h'},
		{"record", 1, 0, 'r'},
		{"playback", 1, 0, 'p'},
		{"no-video", 0, 0, 'v'},
		{"no-ssl", 0, 0, 's'},
		{"no-auth", 0, 0, 'a'},
		{"no-sound", 0, 0, 'S'},
 		{"window-size", 1, 0, 'w'},
		{"debug", 1, 0, 'd'},
		{"fullscreen", 0, 0, 'f'},
		/* tmp option as having two connections is at heart a hack */
		{"sound-port", 1, 0, 'Z'},
		{"resize", 0, 0, 'R'},
		{0, 0, 0, 0}
	};
	char *fname = NULL;

	instr_start(srun);

	if(tc_time_init() == TC_ERROR) {
		tclient_msg("Couldn't initialize timeline\n");
		return TC_EXIT_FAIL;
	}

	time_add("INIT", TI_START);

	fb.do_record = T_FALSE;
	fb.do_playback = T_FALSE;
	fb.use_ssl = T_TRUE;
	fb.use_xvideo = T_TRUE;
	fb.use_auth = T_TRUE;
	fb.use_sound = T_TRUE;
 	fb.tcp_window_size = 0;
	fb.fullscreen = T_FALSE;
	fb.do_resize = T_FALSE;

	while(1) {
		opt = getopt_long(argc, argv, "hr:p:svaw:Sd:Z:fR", long_options, 
			&index);
		if(opt == -1)
			/* finished */
			break;
		switch(opt) {
			case 'h':
				usage(argv[0]);
				return TC_EXIT_OK;
			case 'r':
				fname = optarg;
				fb.do_record = T_TRUE;
				break;
			case 'S':
				tclient_msg("Info: Disabling sound support\n");
				fname = optarg;
				fb.use_sound = T_FALSE;
				break;
			case 'p':
				fname = optarg;
				fb.do_playback = T_TRUE;
				break;
			case 'v':
				tclient_msg("Info: Disabling video support\n");
				fb.use_xvideo = T_FALSE;
				break;
			case 's':
				tclient_msg("Info: Disabling ssl\n");
				fb.use_ssl = T_FALSE;
				break;
			case 'a':
				tclient_msg("Info: Disabling authentication\n");
				fb.use_auth = T_FALSE;
				break;
			case 'w':
				tclient_msg("Info: Setting TCP window "
					"size to %s\n", optarg);
				fb.tcp_window_size = 
					tc_parse_window_size(optarg);
				break;
			case 'd':
#ifdef TC_DEBUG
				tclient_msg("Info: Setting debug level to %s\n",
					optarg);
				tc_debug_level = atoi(optarg);
#else
				tclient_msg("Info: Debug info disabled at "
					"compile time\n");
#endif
				break;
			case 'f':
				tclient_msg("Info: Fullscreen mode\n");
				fb.fullscreen = T_TRUE;
				break;
			case 'Z':
				snd_port = atoi(optarg);
				tclient_msg("Info: Using remote sound port "
					"%d\n", snd_port);
				break;
			case 'R':
				tclient_msg("Info: Enabling resize\n");
				fb.do_resize = T_TRUE;
				break;
			default:
				usage(argv[0]);
				return TC_EXIT_FAIL;
		}
	}

	if(optind > argc-2) {
		tclient_msg("Error: Server and/or port not specified\n");
		/* no host and port */
		usage(argv[0]);
		return TC_EXIT_FAIL;
	}

#ifdef TC_RECORD
	/* check for problems before going on */
	if(fb.do_record && fb.do_playback) {
		tclient_msg("Duuuude, you need to make a choice, either "
			"record or playback, but not both!\n");
		return TC_EXIT_FAIL;
	}

	if(fb.do_record) {
		if(tc_record_init(&fb.record, fname, TC_RECORD_ALL)
			== TC_ERROR) {
			tclient_msg("Error: Disabling event recording\n");
			fb.do_record = T_FALSE;
		}
	} else if(fb.do_playback) {
		if(tc_playback_init(&fb, fname, TC_RECORD_ALL) 
			== TC_ERROR) {
			tclient_msg("Error: Can't playback commands\n");
			fb.do_playback = T_FALSE;
		}
	}
#endif /* TC_RECORD */

	print_msg = T_TRUE;
	res = doConnect(&fb, argv[optind], argv[optind+1]);
	if(res == TC_ERROR) {
		error_msg(fb, ("Error: Couldn't connect to server %s:%s\n",
				argv[optind], argv[optind+1]));
		return TC_EXIT_FAIL;
	}
	initIO(&fb);

	/* do initial handshake and get framebuffer */
	if(tc_initConn(&fb) == TC_ERROR) {
		error_msg(fb, ("Error: Couldn't initialize the connection "
			"to server\n"));
		tc_closeConn(&fb, T_TRUE);
		return TC_EXIT_FAIL;
	}

	if(fb.use_sound) {
		if(tc_initSound(&fb) == TC_ERROR) {
			return TC_EXIT_FAIL;
		}

		/* connect to tsd if the sound card is enabled */
		if (tc_doConnectTsd(&fb, snd_port) == TC_ERROR) {
			error_msg(fb, ("Warning: Couldn't connect to tsd at "
				"destination. %s.\n", strerror(errno)));
			error_msg(fb, ("Warning: tsd may not be running. "
				"Disabling sound.\n"));
			tc_killSound(&fb);
			fb.use_sound = T_FALSE;
		} else {
			tclient_msg("Happy: Connected to tsd.\n");
		}
	}

	/* initialize compression. Put it here in case we end up negotiating
	 * compression at connection time */
	if(tc_init_compress(&fb) == TC_ERROR) { 
		error_msg(fb, ("Error: Couldn't initialize decompression "
			"engine\n"));
		tc_destroyFB(fb);
		return TC_EXIT_FAIL;
	}

	/* init x */
	res = tc_x11Init(&fb, argc, argv);
	if(res == TC_ERROR) {
		error_msg(fb, ("Error: Couldn't initialize X\n"));
		tc_closeConn(&fb, T_TRUE);
		tc_destroyFB(fb);
		return TC_EXIT_FAIL;
	}

	if(fb.use_xvideo) {
		res = tc_video_init(&fb);
		if(res == TC_NOVIDEO) {
			error_msg(fb, ("Info: Couldn't initialize video, "
				"disabling it\n"));
			fb.use_xvideo = T_FALSE;
		} else if(res == TC_ERROR) {
			error_msg(fb, ("Error: Couldn't initialize video\n"));
			tc_x11shutdown(&fb);
			tc_closeConn(&fb, T_TRUE);
			tc_destroyFB(fb);
			return TC_EXIT_FAIL;
		}
	} 
	/* so we catch failure above */
	if(!fb.use_xvideo) {
		res = tc_novideo(&fb);
		if(res == TC_ERROR) {
			error_msg(fb, ("Error: Couldn't tell server we don't "
				"support video\n"));
			tc_x11shutdown(&fb);
			tc_closeConn(&fb, T_TRUE);
			tc_destroyFB(fb);
			return TC_EXIT_FAIL;
		}
	}

	tc_stats_init(&fb);

	tc_instrument_init(&fb);

	res = tc_resize_init(&fb);
	if(res == TC_ERROR) {
		error_msg(fb, ("Error: Couldn't initialize resize\n"));
		if(fb.use_xvideo) 
			tc_video_shutdown(&fb);
		tc_x11shutdown(&fb);
		tc_closeConn(&fb, T_TRUE);
		tc_destroyFB(fb);
		return TC_EXIT_FAIL;
	}
	res = tc_cache_init(&fb);
	if(res == TC_ERROR) {
		error_msg(fb, ("Couldn't initialize cache\n"));
		tc_resize_shutdown(&fb);
		if(fb.use_xvideo) 
			tc_video_shutdown(&fb);
		tc_x11shutdown(&fb);
		tc_closeConn(&fb, T_TRUE);
		tc_destroyFB(fb);
		return TC_EXIT_FAIL;
	}
	tclient_dbg(1, ("width %d, height %d\n", fb.curr_width,
		fb.curr_height));

	signal(SIGINT, sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGUSR1, stats);
	signal(SIGALRM, sigalarm);
	alarm_init();

#ifndef NO_X
	xsock = tc_x11GetFD(&fb);
#endif

	tclient_dbg(1, ("Initialization done\n"));
	res = tc_v2Done(&fb);

	tclient_msg("Happy: All set, entering event loop\n\n");

	did_x11 = T_FALSE;

	init_events();
	add_fd(fb.sock, FD_NET, FD_READ, tc_handleMsg, buffer_notempty);
#ifndef NO_X
	add_fd(xsock, FD_X11, FD_READ, tc_handleEvents, tc_hasevents);
#endif
	if(fb.use_sound) {
		add_fd(fb.snd.sock, FD_TSD, FD_READ, tc_handleSound,
			tc_hasSoundData);
	}

#ifdef TC_RECORD
	if(fb.do_playback)
		tc_playback_start(&fb);
#endif

	/* initialize start total. Only for TC_INSTRUMENT */
	instr_start(stotal);
	time_add("INIT", TI_END);
	time_add_extra("MAIN", TI_START);

	/* event loop */
	while(1) {
		if(gotsig) break;

		if(alarm_got) {
#ifdef TC_RECORD
			if(fb.do_playback && alarm_pending[ALARM_RECORD]) {
				tc_playback_next(&fb);
				alarm_pending[ALARM_RECORD] = T_FALSE;
			}
#endif	/* TC_RECORD */
			if(fb.do_resize && alarm_pending[ALARM_RESIZE]) {
				instr_start(sany);
				res = tc_resize_do_resize(&fb);
				instr_end(&fb, sany, eany, do_resize);
				if(res == TC_ERROR) {
					error_msg(fb, ("Error: Couldn't "
						"resize the window, "
						"exiting\n"));
					break;
				}
				alarm_pending[ALARM_RECORD] = T_FALSE;
			}
			did_x11 = T_TRUE;
			alarm_got = T_FALSE;
		}
		/* chk if we have things pending */
		to_bed = T_TRUE;
		for(i=0; i<FD_LAST; ++i) {
			ev = &eventer.events[i];
			if(ev->pending && ev->pending(&fb)) {
				ev->call_me = T_TRUE;
				to_bed = T_FALSE;
			}
		}
		/* nothing to do, so sleepy */
		if(to_bed) {
 			instr_start(sany);
 			if(did_x11) {
 				tc_x11_going_to_bed(&fb);
 				did_x11 = T_FALSE;
 			}
 			instr_end(&fb, sany, eany, x_up);
 			instr_end(&fb, stotal, etotal, total);
			time_add_extra("MAIN", TI_END);
			set_fd_sets(rf, wf);
 			instr_call(&fb, spoll, epoll, poll, 
				res = select(eventer.maxfd+1, &rf, &wf,
					NULL, NULL)
			);
			time_add_extra("MAIN", TI_START);
 			instr_start(stotal);
			if(res == -1) {
				if(errno == EINTR) {
					continue;
				}
				error_msg(fb, ("Error waiting for data from "
					"server: %s\n",	strerror(errno)));
				error = T_TRUE;
				break;
			} else if(res == 0) 
				continue;

			/* ok, something happened, chk it */
			for(i=0; i<FD_LAST; ++i) {
				ev = &eventer.events[i];
				if(ev->fd == -1)
					continue;

				if( 	FD_ISSET(ev->fd, &rf) ||
					FD_ISSET(ev->fd, &wf)
				) {
					ev->call_me = T_TRUE;
				}
			}
		}
		for(i=0; i<FD_LAST; ++i) {
			ev = &eventer.events[i];
			if(ev->fd != -1 && ev->call_me) {
				res = ev->got(&fb);
				if(res == TC_ERROR || res == TC_CLOSED) {
					error_msg(fb, ("Error handling data "
						"on socket %d\n", i));
					error = T_TRUE;
					break;
				}
				/* some special cases */
				if(i == FD_NET && res == TC_OK) {
					/* for ping */
					did_x11 = T_TRUE;
				} else if(i == FD_X11) {
					did_x11 = T_TRUE;
				}
				ev->call_me = 0;
			}
		}
		/* gotta hate double loops */
		if(error)
			break;
	}

	tclient_msg("\n");

	/* the end */
	time_add_extra("MAIN", TI_END);

	if(fb.sock != -1)
		tc_closeConn(&fb, T_TRUE);

	if(fb.cache_raw_bits)
		tclient_msg("Info: Raw cache has %d active entries. There "
			"were %d "
			"collisions\n", thinc_cache_get_active(fb.raw_cache), 
			thinc_cache_get_collisions(fb.raw_cache));
	if(fb.cache_bit_bits)
		tclient_msg("Info: Bitmap cache has %d active entries. There "
			"were %d "
			"collisions\n", thinc_cache_get_active(fb.bit_cache), 
			thinc_cache_get_collisions(fb.bit_cache));
	if(fb.cache_pix_bits)
		tclient_msg("Info: Pixmap cache has %d active entries. There "
			"were %d "
			"collisions\n", thinc_cache_get_active(fb.pix_cache), 
			thinc_cache_get_collisions(fb.pix_cache));
	tc_cache_destroy(&fb);

#ifdef TC_RECORD
	if(fb.do_record)
		tc_record_end(&fb.record);
	if(fb.do_playback)
		tc_playback_end(&fb);
#endif
	tc_end_compress(&fb);

	tc_resize_shutdown(&fb);

	if(fb.use_xvideo)
		tc_video_shutdown(&fb);

 	instr_end(&fb, srun, erun, run_time);
 
	tc_stats_print(&fb);
	tc_instrument_print(&fb);
	tc_time_print();
	tc_time_end();

	tc_killSound(&fb);

 	tc_x11shutdown(&fb);

	tclient_msg("\nCiao.\n\n");

	if(error)
		return TC_EXIT_FAIL;
	else
		return TC_EXIT_OK;
}

static void usage(const char *progname)
{
        fprintf(stdout, "\nusage: %s [options] host port\n", progname);
        fprintf(stdout, "Options:\n"
		"\t-h/--help            print this message and exit\n"
		"\t-s/--no-ssl          don't use ssl for communication\n"
		"\t-a/--no-auth         don't authenticate to server\n"
		"\t-v/--no-video        don't support video\n"
		"\t-S/--no-sound        don't support sound\n"
		"\t-Z/--sound-port      the remote port where the sound\n"
		"\t                     server is listening. If not specified,\n"
		"\t                     it will be the display port + 10000\n"
#ifdef TC_RECORD
		"\t-r/--record file     record all events to file\n"
		"\t-p/--playback file   playback events from file\n"
#endif
		"\t-f/--fullscreen      be fullscreen\n"
		"\t-w/--window-size N   set TCP window size (you can append\n"
		"\t                     K or M)\n"
		"\t-R/--resize          enable resize support\n"
#ifdef TC_DEBUG
		"\t-d/--debug N		set debug level to N\n"
#endif
		"");
	/* print what kind of capabilities we have. ugly but useful */
	fprintf(stdout, "\nthis client was compiled with:\n");
	fprintf(stdout, "  ssl support\n");
	fprintf(stdout, "  video support\n");
	fprintf(stdout, "  shared memory support\n");
	fprintf(stdout, "  sound support\n");
	fprintf(stdout, "  authentication support\n");
#ifdef TC_RECORD
	fprintf(stdout, "  record and playback support\n");
#endif
	fprintf(stdout, "  resize support\n");
#ifdef TC_STATS
	fprintf(stdout, "  stats collecting support\n");
#endif
	fprintf(stdout, "  png support\n");
#ifdef TC_DEBUG
	fprintf(stdout, "  debug info (default level %d)\n", TC_DEBUG_LEVEL);
#endif
#ifdef TC_DONTHAVEPOLL
	fprintf(stdout, "  internal poll function (hi MAC OS!\n");
#endif
#ifdef TC_TESTING
	fprintf(stdout, "  additional testing code\n");
#endif
#ifdef TC_DONT_UPDATE_X
	fprintf(stdout, "  won't draw updates to the screen\n");
#endif
#ifdef NO_X
	fprintf(stdout, "  headless mode. no window, no nothin'\n");
#endif
#ifdef TC_INSTRUMENT
	fprintf(stdout, "  instrumentation code\n");
#endif
#ifdef TC_TIMELINE
	fprintf(stdout, "  timeline\n");
#endif
	fprintf(stdout, "  protocol version 2\n");
}

static void tc_stats_init(struct tc_fb *fb)
{
	tc_stats_reset(fb);
}

static void tc_stats_reset(struct tc_fb *fb)
{
	tc_stats_zero(fb->stats.rawcount);
	tc_stats_zero(fb->stats.rawbytes);
	tc_stats_zero(fb->stats.copycount);
	tc_stats_zero(fb->stats.copybytes);
	tc_stats_zero(fb->stats.sfillcount);
	tc_stats_zero(fb->stats.sfillbytes);
	tc_stats_zero(fb->stats.pfillcount);
	tc_stats_zero(fb->stats.pfillbytes);
	tc_stats_zero(fb->stats.glyphcount);
	tc_stats_zero(fb->stats.glyphbytes);
	tc_stats_zero(fb->stats.bilevelcount);
	tc_stats_zero(fb->stats.bilevelbytes);

	tc_stats_zero(fb->stats.cursorcount);

	tc_stats_zero(fb->stats.rawtime);
	tc_stats_zero(fb->stats.copytime);
	tc_stats_zero(fb->stats.sfilltime);
	tc_stats_zero(fb->stats.pfilltime);
	tc_stats_zero(fb->stats.glyphtime);
	tc_stats_zero(fb->stats.bileveltime);
}

static void tc_stats_print(struct tc_fb *fb)
{
	tc_stats_msg("\n=== stats for this session ===\n");
	tc_stats_msg("rawcount: %d\n", fb->stats.rawcount);
	tc_stats_msg("rawbytes: %d\n", fb->stats.rawbytes);
	tc_stats_msg("copycount: %d\n", fb->stats.copycount);
	tc_stats_msg("copybytes: %d\n", fb->stats.copybytes);
	tc_stats_msg("sfillcount: %d\n", fb->stats.sfillcount);
	tc_stats_msg("sfillbytes: %d\n", fb->stats.sfillbytes);
	tc_stats_msg("pfillcount: %d\n", fb->stats.pfillcount);
	tc_stats_msg("pfillbytes: %d\n", fb->stats.pfillbytes);
	tc_stats_msg("glyphcount: %d\n", fb->stats.glyphcount);
	tc_stats_msg("glyphbytes: %d\n", fb->stats.glyphbytes);
	tc_stats_msg("bilevelcount: %d\n", fb->stats.bilevelcount);
	tc_stats_msg("bilevelbytes: %d\n", fb->stats.bilevelbytes);

	tc_stats_zero(fb->stats.recvcount);
	tc_stats_add(fb->stats.recvcount, fb->stats.rawcount + 
		fb->stats.copycount + fb->stats.sfillcount + 
		fb->stats.pfillcount + fb->stats.glyphcount + 
		fb->stats.bilevelcount); 
	tc_stats_zero(fb->stats.recvbytes);
	tc_stats_add(fb->stats.recvbytes, fb->stats.rawbytes + 
		fb->stats.copybytes + 
		fb->stats.sfillbytes + fb->stats.pfillbytes + 
		fb->stats.glyphbytes + fb->stats.bilevelbytes);

	tc_stats_msg("\ntotal count: %d\n", fb->stats.recvcount);
	tc_stats_msg("total bytes: %d\n", fb->stats.recvbytes);

	tc_stats_msg("cursor count: %d\n", fb->stats.cursorcount);

	tc_stats_msg("\nsound count: %d\n", fb->stats.soundcount);

	tc_stats_msg("\ncompressed count: %d\n", fb->zfb.z_count);
	tc_stats_msg("compressed bytes: %d\n", fb->zfb.z_bytes);

	tc_stats_msg("\ntime spent in raw updates: %ld ms\n", 
		fb->stats.rawtime);
	tc_stats_msg("time spent in copy updates: %ld ms\n",
		fb->stats.copytime);
	tc_stats_msg("time spent in sfill updates: %ld ms\n", 
		fb->stats.sfilltime);
	tc_stats_msg("time spent in pfill updates: %ld ms\n", 
		fb->stats.pfilltime);
	tc_stats_msg("time spent in glyph updates: %ld ms\n", 
		fb->stats.glyphtime);
	tc_stats_msg("time spent in bilevel updates: %ld ms\n", 
		fb->stats.bileveltime);
	tc_stats_msg("\n=== end stats for this session ===\n");
}

static void tc_instrument_init(struct tc_fb *fb)
{
	tc_instrument_reset(fb);
}

static void tc_instrument_reset(struct tc_fb *fb)
{
	instr_init(fb, run_time);
	instr_init(fb, total);
	instr_init(fb, poll);
	instr_init(fb, handle_msg);
	instr_init(fb, handle_x);
	instr_init(fb, handle_snd);

	instr_init(fb, x_expose);
	instr_init(fb, x_update);

	instr_init(fb, total_raw);
	instr_init(fb, total_copy);
	instr_init(fb, total_sfill);
	instr_init(fb, total_pfill);
	instr_init(fb, total_glyph);
	instr_init(fb, total_bilevel);
	instr_init(fb, total_cursor);
	instr_init(fb, total_ping);

	instr_init(fb, x_evs);
	instr_init(fb, x_up);
	instr_init(fb, basic_net);

	instr_init(fb, draw_raw);
	instr_init(fb, draw_copy);
	instr_init(fb, draw_sfill);
	instr_init(fb, draw_pfill);
	instr_init(fb, draw_glyph);
	instr_init(fb, draw_bilevel);

	instr_init(fb, handle_raw);
	instr_init(fb, handle_copy);
	instr_init(fb, handle_sfill);
	instr_init(fb, handle_pfill);
	instr_init(fb, handle_glyph);
	instr_init(fb, handle_bilevel);

	instr_init(fb, video_start);
	instr_init(fb, video_end);
	instr_init(fb, video_next);
	instr_init(fb, video_mod);

	instr_init(fb, do_resize);

	instr_init(fb, read);
	instr_init(fb, io_read);
	instr_init(fb, write);
	instr_init(fb, io_write);

	instr_init(fb, read_memcpy);
	instr_init(fb, read_poll);

	instr_init(fb, init);
	instr_init(fb, read_fb);
	instr_init(fb, init_x);

	instr_init(fb, cache_add);
	instr_init(fb, cache_get);

	instr_init(fb, raw_comp);
	instr_init(fb, raw_memcpy);
	instr_init(fb, raw_malloc);
	instr_init(fb, raw_read);
	instr_init(fb, raw_read_nonblock);
	instr_init(fb, raw_addcache);
	instr_init(fb, raw_getcache);

	instr_init(fb, comp_init);
	instr_init(fb, comp_read);
	instr_init(fb, comp_comp);
	instr_init(fb, comp_end); 

	instr_init(fb, sfill_read);
	instr_init(fb, sfill_creategc);
	instr_init(fb, sfill_changegc);
	instr_init(fb, sfill_fill);
	instr_init(fb, sfill_addcoords);
	instr_init(fb, sfill_byhand);
	instr_init(fb, num_updates);

	instr_init(fb, win_copy);
	instr_init(fb, win_sync);
	instr_init(fb, raw_sync);
	instr_init(fb, sfill_sync_prepare_create);
	instr_init(fb, sfill_sync_prepare_change);
	instr_init(fb, sfill_sync_draw);
	instr_init(fb, pfill_sync);
	instr_init(fb, pfill_sync_prepare);
	instr_init(fb, glyph_sync);

	instr_init(fb, sfill_create_gc);
	instr_init(fb, sfill_change_gc);
	instr_init(fb, sfill_change_color);

	instr_init(fb, raw_pixels);
	instr_init(fb, sfill_pixels);
	instr_init(fb, pfill_pixels);

}

static void tc_instrument_print(struct tc_fb *fb)
{
	long double speed = get_cpu_speed();

	instr_msg("\n=== instrumentation stats ===\n");

	instr_msg("CPU speed %.3LfMhz\n\n", speed);

	speed *= 1000; /* for ms, one more 1000 for s */
	instr_print(fb, speed, init);
	instr_print(fb, speed, read_fb);
	instr_print(fb, speed, init_x);
	instr_msg("\n");

	instr_print(fb, speed, do_resize);
	instr_msg("\n");

	instr_print(fb, speed, video_start);
	instr_print(fb, speed, video_end);
	instr_print(fb, speed, video_next);
	instr_print(fb, speed, video_mod);
	instr_msg("\n");

	instr_print(fb, speed, sfill_read);
	instr_print(fb, speed, sfill_creategc);
	instr_print(fb, speed, sfill_changegc);
	instr_print(fb, speed, sfill_fill);
	instr_print(fb, speed, sfill_addcoords);
	instr_print(fb, speed, sfill_byhand);
	instr_msg("\n");

	instr_print(fb, speed, win_copy);
	instr_print(fb, speed, win_sync);
	instr_print(fb, speed, raw_sync);
	instr_print(fb, speed, sfill_sync_prepare_create);
	instr_print(fb, speed, sfill_sync_prepare_change);
	instr_print(fb, speed, sfill_sync_draw);
	instr_print(fb, speed, pfill_sync);
	instr_print(fb, speed, pfill_sync_prepare);
	instr_print(fb, speed, glyph_sync);
	instr_msg("\n");

	instr_print_simple(fb, sfill_create_gc);
	instr_print_simple(fb, sfill_change_gc);
	instr_print_simple(fb, sfill_change_color);
	instr_print_simple(fb, raw_pixels);
	instr_print_simple(fb, sfill_pixels);
	instr_print_simple(fb, pfill_pixels);
	instr_msg("\n");

	instr_print(fb, speed, cache_add);
	instr_print(fb, speed, cache_get);
	instr_msg("\n");

	instr_print(fb, speed, raw_comp);
	instr_print(fb, speed, raw_memcpy);
	instr_print(fb, speed, raw_malloc);
	instr_print(fb, speed, raw_read);
	instr_print(fb, speed, raw_read_nonblock);
	instr_print(fb, speed, raw_addcache);
	instr_print(fb, speed, raw_getcache);
	instr_msg("\n");

	instr_print(fb, speed, comp_init);
	instr_print(fb, speed, comp_read);
	instr_print(fb, speed, comp_comp);
	instr_print(fb, speed, comp_end); 
	instr_msg("\n");

	instr_print(fb, speed, draw_raw);
	instr_print(fb, speed, draw_copy);
	instr_print(fb, speed, draw_sfill);
	instr_print(fb, speed, draw_pfill);
	instr_print(fb, speed, draw_glyph);
	instr_print(fb, speed, draw_bilevel);
	instr_msg("\n");

	instr_print(fb, speed, handle_raw);
	instr_print(fb, speed, handle_copy);
	instr_print(fb, speed, handle_sfill);
	instr_print(fb, speed, handle_pfill);
	instr_print(fb, speed, handle_glyph);
	instr_print(fb, speed, handle_bilevel);
	instr_msg("\n");

	instr_print(fb, speed, total_raw);
	instr_print(fb, speed, total_copy);
	instr_print(fb, speed, total_sfill);
	instr_print(fb, speed, total_pfill);
	instr_print(fb, speed, total_glyph);
	instr_print(fb, speed, total_bilevel);
	instr_print(fb, speed, total_cursor);
	instr_print(fb, speed, total_ping);
	instr_msg("\n");

	instr_print(fb, speed, x_expose);
	instr_print(fb, speed, x_update);
	instr_msg("\n");

	instr_print(fb, speed, read_memcpy);
	instr_print(fb, speed, read_poll);
	instr_msg("\n");

	instr_print(fb, speed, read);
	instr_print(fb, speed, io_read);
	instr_print(fb, speed, write);
	instr_print(fb, speed, io_write);
	instr_msg("\n");

	instr_print(fb, speed, basic_net);
	instr_print(fb, speed, x_evs);
	instr_print(fb, speed, x_up);
	instr_msg("\n");

	instr_print(fb, speed, handle_msg);
	instr_print(fb, speed, handle_x);
	instr_print(fb, speed, handle_snd);
	instr_print(fb, speed, poll);
	instr_print(fb, speed, total);

	instr_print(fb, speed, run_time);

	instr_print_simple(fb, num_updates);

	instr_msg("\n=== end instrumentation stats ===\n");
}

static int tc_init_compress(struct tc_fb *fb)
{
	tc_stats_zero(fb->zfb.z_count);
	tc_stats_zero(fb->zfb.z_bytes);
	fb->zfb.fb = fb;

	return TC_OK;
}

static void tc_end_compress(struct tc_fb *fb)
{
	/* noop. zlib used to have some code here */
}

void tc_destroyFB(struct tc_fb fb)
{
	if(fb.fbdata)
		free(fb.fbdata);
	if(fb.cursorInfo.cursor)
		free(fb.cursorInfo.cursor);
}

static int tc_cache_init(struct tc_fb *fb)
{
	fb->raw_cache = NULL;
	fb->bit_cache = NULL;
	fb->pix_cache = NULL;
	if(fb->cache_raw_bits) {
		fb->raw_cache = thinc_cache_init(fb->cache_raw_bits);
		if(!fb->raw_cache) 
			return TC_ERROR;
	}
	if(fb->cache_bit_bits) {
		fb->bit_cache = thinc_cache_init(fb->cache_bit_bits);
		if(!fb->bit_cache) {
			thinc_cache_destroy(fb->raw_cache);
			return TC_ERROR;
		}
	}
	if(fb->cache_pix_bits) {
		fb->pix_cache = thinc_cache_init(fb->cache_pix_bits);
		if(!fb->pix_cache) {
			thinc_cache_destroy(fb->raw_cache);
			thinc_cache_destroy(fb->pix_cache);
			return TC_ERROR;
		}
	}
	return TC_OK;
}

static void tc_cache_destroy(struct tc_fb *fb)
{
	if(fb->cache_raw_bits)
		thinc_cache_destroy(fb->raw_cache);
	if(fb->cache_bit_bits)
		thinc_cache_destroy(fb->pix_cache);
	if(fb->cache_pix_bits)
		thinc_cache_destroy(fb->bit_cache);
}

static int tc_parse_window_size(const char *optarg)
{
	int window_size;
	char end = '\0';

	sscanf(optarg, "%d%c", &window_size, &end);
	switch(end) {
		case 'k':
		case 'K': 
			window_size*= 1024; 
			break;
		case 'm':
		case 'M': 
			window_size *= 1024*1024; 
			break;
		default:
			break;
	}

	return window_size;
}
