/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/*
 * Header file for X11 THINC client
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

/* TODO: Create a Region (Xlib XCreateRegion(3)) with our viewport and
 * use that to verify updates */

#ifndef _THINC_CLIENT_H_
#define _THINC_CLIENT_H_

#include "thincCommon.h"
#include "thincProto2.h"
#include <sys/time.h>
#include <time.h>
#include <png.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xvlib.h>

#include "thinc_ssl.h"

extern const char *cert_path;

#include "thinc_cache.h"

/* for timestamp instrumentation */
struct tc_stamps {
	/* time at which client received chunk of data */
	unsigned int recvstamp;
	/* time at which client is to play a chunk of data */
	unsigned int playstamp;
};

#include "tc_sound.h"

#ifdef TC_RECORD
#include "tc_record.h"
#endif

/* all the instrumentation in one place */
#include "tc_instr.h"

#define TC_VERSION1	0
#define TC_VERSION2	1
#define TC_VERSION3	2

/* x11 data */
struct tc_x11;

/* forward declarations */
struct tc_fb;

/* cursor info */
struct tc_cursorInfo {
	unsigned int fg;
	unsigned int bg;
	unsigned int width;
	unsigned int height;
	unsigned int size;
	unsigned int x;
	unsigned int y;
	unsigned int xhot;
	unsigned int yhot;
	/* the actual cursor */
	char *cursor;	/* the data */
	/* if the cursor is not ARGB, these two point inside cursor above */
	char *source;
	char *mask;
	unsigned int argb:1;	/* is the cursor ARGB? */
};

struct tc_compress {
	struct tc_fb *fb;
	/* png */ 
	png_structp png_ptr;
	png_infop info_ptr;
	char *curr;
#ifdef TC_STATS
	int z_count;	/* stats */
	int z_bytes;
#endif
};

/* structs that store format specific info */
/* T_FOURCC_YV12 */
struct tc_video_yv12 {
	/* YUV_OFFSETS_PITCHES */
	/* stored in the XvImage but may be useful for portability */
	unsigned int yuv_offsets[3];
	unsigned int yuv_pitches[3];
};

/* T_FOURCC_YUY2 */
struct tc_video_yuy2 {
	int pitch;
};

/* T_FOURCC_UYVY */
struct tc_video_uyvy {
	int pitch;
};

struct tc_video_stream {
	unsigned int id;
	unsigned int format;
	unsigned int x;
	unsigned int y;
	unsigned int width;
	unsigned int height;
	unsigned int dst_width;
	unsigned int dst_height;
	/* client resizing */
	unsigned int curr_x; 
	unsigned int curr_y;
	unsigned int curr_width;
	unsigned int curr_height;
#ifndef NO_X
	/* TODO: Use double buffering */
	XvImage *image;
#else
	char *image;
#endif
	/* initial stamp on video open for synchronization */
	unsigned int initstamp;
	/* timestamp of current frame for synchronization */
	unsigned int timestamp;
	struct tc_stamps instr_stamps;
	/* for determining whether to waitsync the video */
	int doWaitSync;
	XShmSegmentInfo shminfo;
	/* format specific info */
	union {
		struct tc_video_yv12 yv12;
		struct tc_video_yuy2 yuy2;
		struct tc_video_uyvy uyvy;
	} fmt;
};

#define TC_MAX_STREAMS	1

struct tc_video {
	int have_vid;
	int port_id;
	unsigned long colorkey;
	int num_fmts;
	thinc_vidFormat *fmts;	/* use same struct as proto */
	int num_streams;
	struct tc_video_stream* streams;
	Bool ck_paint;
	GC ck_gc;
};

struct tc_resize;

union tc_io {
	SSL *ssl;
	int sock;
};

/* all the info needed by the client to run */
struct tc_fb {
	/* the connection to the server */
	int sock;
	/* fb info */
	int width;
	int height;
	int curr_width;
	int curr_height;
	int depth;
	int bpp;
	int Bpp;	/* bytes per pixel */
	int Bpp_shift;	/* make multiplication with Bpp faster */
	int byte_order;
	int compress;	/* true if server is using compression */
	int total_size;
	/* the actual fb contents */
	char *fbdata;
	/* the cursor */
	struct tc_cursorInfo cursorInfo;
	struct tc_x11 *x11;

	int (*read)(union tc_io io, void *data, int count);
	int (*write)(union tc_io io, const void *data, int count);
#ifdef TC_STATS
	/* stats */
	struct tc_stats stats;
#endif
#ifdef TC_INSTRUMENT
	struct tc_instrument instr;
#endif
	struct tc_compress zfb;
	struct tc_video vid;
	SSL *ssl;
	struct tc_resize *resize;
	thinc_cache *raw_cache;
	thinc_cache *bit_cache;
	thinc_cache *pix_cache;
	unsigned int cache_raw_bits;
	unsigned int cache_bit_bits;
	unsigned int cache_pix_bits;
	struct tc_snd snd;
	/* for tracking mouse events issued by the server */
	int move_count;

#ifdef TC_RECORD
	struct tc_record record;
#endif

	int tcp_window_size;
	
	union tc_io io;

	struct sockaddr_in remote_addr;	/* remote address */

	/* options are always present to minimize the spaghetti-ness */
	unsigned int use_xvideo:1;
	unsigned int use_ssl:1;
	unsigned int do_record:1;	/* record all events */
	unsigned int do_playback:1;	/* playback events */
	unsigned int use_auth:1;
	unsigned int use_sound:1;
	unsigned int app_share:1;	/* do application sharing */
	unsigned int fullscreen:1;
	unsigned int do_resize:1;
};

/* needs fb */
#include "tc_region.h"

/* Application sharing */
#ifdef TC_APPSHARE
#include "thinc_list.h"

struct tc_app {
	thinc_list	list;
	thinc_appDesc 	app;
	char *name;
};

#endif /* TC_APPSHARE */

/****************** communication functions **************************/
/* Executes handshake with server. Socket should be already connected. */
int tc_initConn(struct tc_fb *fb);
/* processes a server message read from sock */
int tc_handleMsg(struct tc_fb *fb);
/* closes the connection to the thinc server */
void tc_closeConn(struct tc_fb *fb, int sendmsg);
/* write the specified message type onto the socket */
int tc_writeMsg(struct tc_fb *fb, int type, const void *msg, int msg_size);
/* read a message of type type from socket sock */
int tc_readMsg(struct tc_fb *fb, int type, void *msg, int msg_size);
/* reads a message with its header. No type checking is performed */
int tc_readMsgAndHeader(struct tc_fb *fb, thinc_header *h, void *msg, 
	int msg_size);
/* reads in a message header from socket sock. returns 0 on success, -1
 * on error */
int tc_readHeader(struct tc_fb *fb, thinc_header *h);
/* reads a message of type type from socket sock, but without the header */
int tc_readMsgNoHdr(struct tc_fb *fb, int type, void *msg, int msg_size);
/* free memory used by the fb structure */
void tc_destroyFB(struct tc_fb fb);

int tc_writeSoundMsg(struct tc_fb *fb, int type, const void *msg, int msg_size);
int tc_readSoundHeader(struct tc_fb *fb, thinc_header *h);
int tc_readSoundMsgAndHeader(struct tc_fb *fb, thinc_header *h, void *msg, 
	int msg_size);
int tc_readSoundMsgNoHdr(struct tc_fb *fb, void *msg, int msg_size);

/******************** x11 functions ***********************************/
/* initializes x11 component of the client. This consist of opening the
 * connection to the X server and creating the display window and the
 * image. */
int tc_x11Init(struct tc_fb *fb, int argc, char **argv);
inline int tc_x11GetFD(struct tc_fb *fb);
inline void tc_x11_going_to_bed(struct tc_fb *fb);
int tc_x11imgUpdate(struct tc_fb *fb, int x, int y, int width, int height);
int tc_x11update_resized(struct tc_fb *fb, int rsz_x, int rsz_y, int rsz_w,
	int rsz_h);
int tc_x11copyData(struct tc_fb *fb, int sx, int sy, int dx, int dy, 
	int width, int height);
int tc_x11solidFill(struct tc_fb *fb, int x, int y, int width, int height, 
	unsigned long pixel);
void* tc_x11pfillNew(struct tc_fb *fb, char *data, int x, int y,
	int width, int height);
int tc_x11pfillFill(struct tc_fb *fb, void *handle, thinc_rectRegion *r,
	unsigned char resized);
void tc_x11fillDestroyWrap(void *f, void *p);
int tc_x11bitmapFill(struct tc_fb *fb, void *p, thinc_rectRegion *r);
void* tc_x11bilevelNew(struct tc_fb *fb, char *bitmap, int x, int y,
	int width, int height, unsigned long fg, unsigned long bg);
void* tc_x11glyphNew(struct tc_fb *fb, char *bitmap, int x, int y,
	int width, int height, unsigned long color);

/* cursor funcs */
int tc_x11cursorChange(struct tc_fb *fb);
int tc_x11cursorMove(struct tc_fb *fb, int x, int y);
void tc_x11cursorRecolor(struct tc_fb *fb, int fgv, int bgv);

/* shutdowns the x11 component of the client */
void tc_x11shutdown(struct tc_fb *fb);
/* handles x11 events */
int tc_handleEvents(struct tc_fb *fb);

/* Returns void so header files don't need to include X headers. Cast to
 * XImage* */
void* thinc_createImage(struct tc_fb *fb, unsigned int width, 
	unsigned int height, char *data);

/* event functions */
int tc_sendKeyEv(struct tc_fb *fb, int key, int updown);
int tc_sendButtonEv(struct tc_fb *fb, int button, int updown);
int tc_sendMotionEv(struct tc_fb *fb, int x, int y);

/*** XVideo ***/
int tc_video_init(struct tc_fb *fb);
void tc_video_shutdown(struct tc_fb *fb);
int tc_novideo(struct tc_fb *fb);
int tc_handleVideoStart(struct tc_fb *fb, thinc_header *h);
int tc_handleVideoEnd(struct tc_fb *fb, thinc_header *h);
int tc_handleVideoNext(struct tc_fb *fb, thinc_header *h); 
int tc_handleVideoScale(struct tc_fb *fb, thinc_header *h);
int tc_handleVideoResize(struct tc_fb *fb, thinc_header *h);
int tc_handleVideoMove(struct tc_fb *fb, thinc_header *h);
inline int tc_video_client_scale_all(struct tc_fb *fb, double factor_x, 
	double factor_y);
inline int tc_video_client_scale(struct tc_fb *fb, int id, double factor_x, 
	double factor_y);
inline int tc_video_client_move_all(struct tc_fb *fb, double factor_x, 
	double factor_y);
int tc_video_client_move(struct tc_fb *fb, int id, double factor_x,
	double factor_y);

/* png */
int tc_pngInit(struct tc_compress *zfb);
void tc_pngEnd(struct tc_compress *zfb);

/*** resize ***/
int tc_resize_init(struct tc_fb *fb);
void tc_resize_shutdown(struct tc_fb *fb);
int tc_resize_resize(struct tc_fb *fb, int width, int height);
int tc_resize_do_resize(struct tc_fb *fb);
inline int tc_resize_video_all(struct tc_fb *fb);
inline int tc_resize_video(struct tc_fb *fb, int id);
inline int tc_resize_move_video_all(struct tc_fb *fb);
inline int tc_resize_move_video(struct tc_fb *fb, int id);
int tc_resize_handle_update(struct tc_fb *fb, thinc_rawUpdate *r,
	thinc_header *h);
int tc_resize_handle_bitmap(struct tc_fb *fb, thinc_header *h, 
	unsigned int *bitmap_size);
inline int tc_resize_scale_x(struct tc_fb *fb, int num);
inline int tc_resize_scale_y(struct tc_fb *fb, int num);
inline int tc_resize_scale_width(struct tc_fb *fb, int num);
inline int tc_resize_scale_height(struct tc_fb *fb, int num);

/** A/V synchronization **/
#ifdef TC_USESYNC
int tc_doSoundSync(struct tc_fb *fb, struct tc_snd *snd);
int tc_doVideoSync(struct tc_fb *fb, struct tc_video_stream *stream);
void tc_resetVideoTimeStamp(struct tc_fb *fb);
void tc_syncVideoStampWithSoundStamp(struct tc_fb *fb);
int tc_waitSyncVideo(struct tc_fb *fb, struct tc_video_stream *stream, int delta);
int tc_waitSyncSound(struct tc_fb *fb, struct tc_snd *snd, int delta);
int tc_syncSound(struct tc_fb *fb, int delta);
#endif

/******************** Utility functions *******************************/

/* Connects to server host:port. Returns -1 on error */
int doConnect(struct tc_fb *fb, const char *host, const char *port);
/* chk if there's data buffered for reading */
inline int buffer_empty();
inline int buffer_size();
inline int buffer_notempty(struct tc_fb *fb);
int tc_hasevents(struct tc_fb *fb);
void initIO(struct tc_fb *fb);
int tc_v2ReadNotOk(struct tc_fb *fb);
int tc_v2SendNotOk(struct tc_fb *fb, const char *reason);
int tc_v2SendReq(struct tc_fb *fb, void *req, int req_size);
int tc_v2Done(struct tc_fb *fb);
void setSecureIO(struct tc_fb *fb);
/* read/write data from/into passed socket */
int readData(struct tc_fb *fb, void *buf, int count, int block);
int writeData(struct tc_fb *fb, const void *buf, int count);

/* for reading/writing sound data from tsd. */
/* :NOTE: eventually these functions may be thrown away. */
int readSoundData(struct tc_fb *fb, void *data, int count, int block);
int writeSoundData(struct tc_fb *fb, const void *buf, int count);

/* UDP */
int tc_initUDP(struct tc_fb *fb, struct tc_snd *snd, int port);

/* poll wrapper */
int doPoll(int sock, short events, short *revents);
/* prints a message to stdout. Not for debugging (see below) */
void tclient_msg(const char *fmt, ...)
	__attribute__ ((format (__printf__, 1, 2)));
/* prints a message to stderr. Not for debugging (see below) */
void tclient_verb(const char *fmt, ...)
	__attribute__ ((format (__printf__, 1, 2)));

extern int print_msg;

#define tc_error_msg(...) do {		\
	print_msg = T_FALSE;		\
	tclient_msg("Error: " __VA_ARGS__) ;		\
} while(0)

/* cache functions */
int tc_handleUpdateCached(struct tc_fb*, unsigned long int *);
int tc_handleUpdateAddCache(struct tc_fb*, unsigned long int *);

/* For testing purposes */
#ifdef TC_TESTING
/* if condition is false, we die. the ... are for the msg you want to 
 * print before dying */ 
# define tc_pass_or_die(cond, ...) do {				\
	if(!(cond)) {						\
		tclient_msg( __VA_ARGS__);			\
		tclient_msg("\nOh well, don't wait for me tonight...\n"); \
		exit(EXIT_FAILURE);				\
	}							\
} while(0)
#else
# define tc_pass_or_die(cond, msg, ...)
#endif

/* For debugging. To disable altogether, comment TC_DEBUG in Makefile. 3 is
 * current max level */
extern int tc_debug_level;
#ifndef TC_DEBUG_LEVEL
# define TC_DEBUG_LEVEL 0
#endif
void _tclient_dbg(const char *fmt, ...)
	__attribute__ ((format (__printf__, 1, 2)));

#ifdef TC_DEBUG

#define tc_dbg_name(func)	const char *func = __func__;

/* p is the actual parameters to print function. Pass enclosed in () */
#define tclient_dbg(level, p)  do { \
	if((level) <= tc_debug_level) \
		_tclient_dbg p; \
} while(0)
	
#else /* TC_DEBUG */

#define tclient_dbg(level, p) do { /* void */ } while(0)
#define tc_dbg_name(func)

#endif /* TC_DEBUG */
	
/* useful macros */
#ifndef MIN
# define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif

#ifndef MAX
# define MAX(x,y) (((x) > (y)) ? (x) : (y))
#endif

/* these are only for positive numbers
 * (undef first to make sure we get ours) */
#undef FLOOR
#define FLOOR(x) (int)(x)
#undef CEILING
#define CEILING(x) ((x) == FLOOR(x) ? (x) : FLOOR(x)+1)

/* Maximum sizes used for some of the messages */
#define T_MAXW	128
#define T_MAXH	128

/* Max size of the bitmap sent in a Glyph or Bilevel command */
#define T_PIXELS_PER_BYTE	8U
#define T_MAX_BITMAP_SIZE	(T_MAXW*T_MAXH)/T_PIXELS_PER_BYTE

/* size of a bitmap of dimensions (w*h) */
#define T_BITMAP_SIZE(w,h)	(CEILING((float)(w)/T_PIXELS_PER_BYTE)*(h))

/* old gcc's don't know about __func__ but do know about __FUNCTION__ :*( */
#ifndef __func__
#define __func__ __FUNCTION__
#endif

/* return values. TODO: use them */
#define TC_OK		1
#define TC_ERROR	-1
#define TC_CLOSED	0
#define TC_WOULDBLOCK	-2
#define TC_PING		2
#define TC_NOVIDEO	-2
					    /* useful macros */
/* orig -> pointer to start of data.
 * orig_width -> width of data (in pixels, not bytes, e.g. fb->width)
 * x, y -> position you're trying to get to.
 * ptr, stride -> output
 * ptr -> will point to start of data you want
 * stride -> distance between the start of two consecutive lines
 */
#define tc_getDataPtr(fb, orig, orig_width, x, y, ptr, stride) do { \
	int x_off = (x) * (fb)->Bpp; \
	stride = (orig_width) * (fb)->Bpp; \
	ptr = (orig) + x_off + (y) * stride; \
} while(0)

#define tc_widthInByte(fb, width)	((fb)->Bpp*width)

#define fix_tc_rect(rect) \
	(rect).x = ntohs((rect).x); \
	(rect).y = ntohs((rect).y); \
	(rect).width = ntohs((rect).width); \
	(rect).height = ntohs((rect).height);

#ifdef TC_PRINT_TIMESTAMPS
#define tc_timestamp(msg) \
	tc_verb msg
#else
#define tc_timestamp(msg)
#endif	/* TC_PRINT_TIMESTAMPS */

#define pix_to_bytes(fb, width) ( (width) << (fb)->Bpp_shift )

/* protocol 2 stuff */
#define v2_readMsg(err, fb, msg, size) do {	\
	err = tc_readMsgNoHdr(fb, 0, msg, size); \
	if(err == TC_ERROR || err == TC_CLOSED) {	\
		tc_error_msg("Couldn't read server message\n");  \
		return TC_ERROR;			\
	}						\
} while(0)

/* type is the message you're expecting */
#define v2_readHeader(err, fb, h, msg_type) do {	\
	err = tc_readHeader(fb, &h);			\
	if(err == TC_ERROR || err == TC_CLOSED) {	\
		tclient_dbg(1, ("Error: %s: Couldn't read message header\n",  \
			func));				\
		return TC_ERROR;			\
	} else if(h.type == T_SERV_NOTOK) {		\
		tclient_dbg(2, ("%s: received reject from server :(\n", func));\
		tc_v2ReadNotOk(fb);			\
		return TC_ERROR;			\
	} else if(h.type != msg_type) { 		\
		tclient_dbg(1, ("Error: %s: Unknown message %d received, " \
			"expected %d\n", func, h.type, msg_type));	\
		return TC_ERROR;			\
	}						\
	h.length = ntohs(h.length);			\
} while(0)

#define v2_sendReq(err, fb, req, size) do {	\
	err = tc_v2SendReq(fb, req, size);	\
	if(err == TC_ERROR) {			\
		tclient_dbg(1, ("Error: %s: Couldn't send request\n", func)); \
		return TC_ERROR;		\
	}					\
} while(0)

#define v2_checkReply(r, reply, type) do {	\
	r = t_reply_read_type(reply);		\
	if(r == T_REPLY_UNKNOWN) {		\
		tclient_dbg(1, ("Error: %s: Server does not understand "  \
			"request\n", func));	\
		return TC_ERROR;		\
	} else if(r != type) {		\
		tclient_dbg(1, ("Error: %s: Unknown reply %d received from " \
			"server\n", func, r));	\
	}					\
} while(0)

#define v2_getReply(generic, which)	((generic).which)

enum {
	ALARM_RESIZE,
	ALARM_RECORD,
	ALARM_LAST
} alarms;

extern int alarm_pending[ALARM_LAST];

#define set_alarm(which, sec) do {	\
	alarm_pending[which] = T_TRUE;	\
	alarm(sec);			\
} while(0)

#define set_timer(which, sec, usec, err) do {	\
	struct itimerval __timer;		\
	alarm_pending[which] = T_TRUE;		\
	__timer.it_interval.tv_sec = 0;		\
	__timer.it_interval.tv_usec = 0;	\
	__timer.it_value.tv_sec = sec;		\
	__timer.it_value.tv_usec = usec;	\
	err = setitimer(ITIMER_REAL, &__timer, NULL);	\
} while(0)

/* so ppl can add file descriptors to chk on the main loop */
enum {
	FD_NET,
	FD_X11,
	FD_TSD,		/* sound from/to network */
	FD_SND_PLAY,	/* sound playback */
	FD_SND_CAPTURE,	/* sound capture */
	FD_SND_UDP,	/* sound on the UDP socket */
	FD_LAST
} fd_pretty;

struct event_handler {
	int fd;
	/* called when something happened on this fd */
	int (*got)(struct tc_fb *fb);
	/* called to check if there's something pending for this fd */
	int (*pending)(struct tc_fb *fb);

	unsigned int call_me:1;
};

inline void add_fd(int fd, int which, unsigned int rw,
	int (*got)(struct tc_fb*), int (*pending)(struct tc_fb*));
inline void remove_fd(int which);

/* values for rw in add_fd */
#define FD_READ		0x01
#define FD_WRITE	0x02

#endif	/* _THINC_CLIENT_H_ */
