/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions to control connection to thinc server
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * Copyright 2004
 */

/* set tabstop=8 */

#include "thinc_client.h"
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <netinet/in.h>

#include "thinc_auth.h"

#if 0
#define TC_DUMP_BITMAP
#endif

/* compute the size of a framebuffer described by fb */
#define TC_FBSIZE(fb) (((fb)->width << (fb)->Bpp_shift) * (fb)->height)

#ifdef TC_DUMP_BITMAP
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

static void tc_dumpBitmap(char *bitmap, int width, int height);
#endif

#if 0
static void tc_fillImg(struct tc_fb *fb, unsigned long pixel, int x, int y,
	int width, int height);
#endif

/* local functions */
static int tc_doCreateFB(struct tc_fb *fb);
static int tc_doCreateCursor(struct tc_fb *fb);
static int tc_readZFB(struct tc_fb *fb);

static int tc_handlePing(struct tc_fb *fb, thinc_header *h);
static int tc_handleRUpdate(struct tc_fb *, thinc_header *h);
static int tc_handleZUpdate(struct tc_fb *fb, thinc_rawUpdate *r, 
	thinc_header *h);
static int tc_handleZUpdateSimple(struct tc_fb *fb, thinc_rectRegion *rect);
static int tc_handleRszUpdate(struct tc_fb *fb, thinc_rawUpdate *r, 
	thinc_header *h);
static int tc_handleCopy(struct tc_fb *fb, thinc_header *h);
static int tc_handleSFill(struct tc_fb *fb, thinc_header *h);
static int tc_handlePFill(struct tc_fb *fb, thinc_header *h);
static int tc_handleBilevel(struct tc_fb *fb, thinc_header *h);
static int tc_handleGlyph(struct tc_fb *fb, thinc_header *h);

/* cache */
static int tc_addRUpdateCache(struct tc_fb *fb, int x, int y, int width, 
	int height, unsigned long int cache_id);
static int tc_handleRawCachedUpdate(struct tc_fb *fb, thinc_rawUpdate* r);
static int tc_addPixmapCache(struct tc_fb *fb, unsigned long int cache_id,
	char *pixmap, unsigned int pixmap_size);
static int tc_getPixmapCached(struct tc_fb *fb, unsigned long int cache_id,
	char **pixmap, unsigned int pixmap_size);
static int tc_addBitmapCache(struct tc_fb *fb, unsigned long int cache_id,
	char *bitmap, unsigned int bitmap_size);
static int tc_getBitmapCached(struct tc_fb *fb, unsigned long int cache_id,
	char **bitmap, unsigned int bitmap_size);

static char * tc_getData(struct tc_fb *fb, int x, int y, int width,
	int height, unsigned long int *size);
static int tc_putData(struct tc_fb *fb, char *data, int x, int y, 
	int width, int height);

static void tc_pngReadData(png_structp png_ptr, png_bytep data,
	png_size_t length);

static int tc_rawHandleUnresized(struct tc_fb *fb, thinc_rawUpdate *r,
	thinc_header *h);

/* cursor messages */
static int tc_handleCursorChange(struct tc_fb *fb, thinc_header *h);
static int tc_handleCursorShowHide(struct tc_fb *fb, thinc_header *h);
static int tc_handleCursorMove(struct tc_fb *fb, thinc_header *h);
static int tc_handleCursorColor(struct tc_fb *fb, thinc_header *h);

/*****************************
 * functions for protocol v2 *
 *****************************/

/* XXX: temp hack */
static const char *thinc_version_str = "THINC_0.2";
static const unsigned int thinc_version_len = 10; /* with \0 */

/* Local funcs */
static int tc_v2Version(struct tc_fb *fb);
static int tc_v2SecCaps(struct tc_fb *fb);
static int tc_v2SSL(struct tc_fb *fb);
static int tc_v2Auth(struct tc_fb *fb);
static int tc_v2Requests(struct tc_fb *fb);

/* request funcs */
static int tc_v2ReqFBInfo(struct tc_fb *fb);
static int tc_v2ReqCursor(struct tc_fb *fb);
static int tc_v2ReqEncoder(struct tc_fb *fb);
static int tc_v2ReqFBData(struct tc_fb *fb);
static int tc_v2ReqCache(struct tc_fb *fb);
static int tc_v2ReqAppshare(struct tc_fb *fb);

/* helper funcs */
static int tc_readVersionString(struct tc_fb *fb);

#define call_v2(res, func, fb) do {	\
	res = func(fb);			\
	if(res == TC_ERROR) {		\
		return TC_ERROR;	\
	}				\
} while(0)

int tc_initConn(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int res;

	tclient_dbg(1, ("%s: initializing connection to server\n", func));


	/* TODO: Use message stack */
	/* 1. version exchange */
	tclient_dbg(1, ("%s: 1. version exchange\n", func));
	call_v2(res, tc_v2Version, fb);

	/* 2. security caps exchange */
	tclient_dbg(1, ("%s: 2. security caps exchange\n", func));
	call_v2(res, tc_v2SecCaps, fb);

	/* 3. ssl handshake (opt) */
	tclient_dbg(1, ("%s: 3. ssl handshake\n", func));
	call_v2(res, tc_v2SSL, fb);

	/* 4. auth (opt) */
	tclient_dbg(1, ("%s: 4. authentication\n", func));
	call_v2(res, tc_v2Auth, fb);

	/* 5. features */
	tclient_dbg(1, ("%s: 6. features\n", func));
	call_v2(res, tc_v2Requests, fb);

	/* not done yet. we need to do video, but for that we need to
	 * initialize X first. so wait we must */
	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2Version(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;

	tclient_dbg(2, ("%s: exchanging versions with the server\n", func));

	err = tc_readVersionString(fb);
	if(err == TC_ERROR) {
		tc_error_msg("Couldn't exchange versions with the "
			"server\n");
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: sending our version: %s\n", func, 
		thinc_version_str));
	err = writeData(fb, thinc_version_str, thinc_version_len);
	if(err == TC_ERROR) {
		tc_error_msg("Couldn't send our version\n");
		return TC_ERROR;
	}
	
	return TC_OK;
}

static int tc_v2SecCaps(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;
	thinc_header h;
	thinc_serverSecCaps s;
	thinc_clientSecCaps c;
	thinc_sessionSecCaps ss;

	tclient_dbg(2, ("%s: exchanging sec caps\n", func));

	v2_readHeader(err, fb, h, T_SERV_SEC_CAPS);

	v2_readMsg(err, fb, &s, T_SERV_SEC_CAPS_SIZE);

	s.caps = ntohl(s.caps);
	tclient_dbg(2, ("%s: server sec caps: %#x\n", func, s.caps));

	/* our caps must be subset of server caps */
	c.caps = 0;
	if(fb->use_ssl) {
		if(!(s.caps & T_CAPS_SEC_ENC)) {
			tclient_msg("Warning: Server does not support "
				"encryption\n");
		} else {
			tclient_dbg(2, ("%s: setting up encrypted session\n",
				func));
			c.caps |= T_CAPS_SEC_ENC;
		}
	}
	if(fb->use_auth) {
		if(!(s.caps & T_CAPS_SEC_AUTH)) {
			tclient_msg("Info: Server does not support "
				"authentication\n");
		} else {
			tclient_dbg(2, ("%s: doing authentication\n", func));
			c.caps |= T_CAPS_SEC_AUTH;
		}
	}
	tclient_dbg(2, ("%s: our caps %#x\n", func, c.caps));
	c.caps = htonl(c.caps);
	err = tc_writeMsg(fb, T_CLIENT_SEC_CAPS, &c, T_CLIENT_SEC_CAPS_SIZE);
	if(err == TC_ERROR) {
		tc_error_msg("Couldn't send security caps to server\n");
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: reading session sec caps\n", func));

	v2_readHeader(err, fb, h, T_SESS_SEC_CAPS);
	v2_readMsg(err, fb, &ss, T_SESS_SEC_CAPS_SIZE);

	ss.caps = ntohl(ss.caps);
	tclient_dbg(2, ("%s: session sec caps %#x\n", func, ss.caps));
	if(!(ss.caps & T_CAPS_SEC_ENC))
		fb->use_ssl = T_FALSE;
	if(!(ss.caps & T_CAPS_SEC_AUTH))
		fb->use_auth = T_FALSE;

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2SSL(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;

	if(!fb->use_ssl) {
		return TC_OK;
	}
	
	tclient_dbg(2, ("%s: doing SSL handshake\n", func));
	fb->ssl = NULL;
again:
	err = thinc_ssl_connect(fb->sock, cert_path, &fb->ssl);
	if(err == T_SSL_WOULDBLOCK) {
		goto again;
	} else if(err == T_SSL_ERROR || err == T_SSL_CLOSED) {
		tc_error_msg("Couldn't initialize secure connection\n");
		return TC_ERROR;
	}

	setSecureIO(fb);

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2Auth(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct thinc_user info;
	thinc_cAuth auth;
	thinc_header h;
	int res;

	if(!fb->use_auth)
		return TC_OK;

	tclient_dbg(2, ("%s: doing authentication\n", func));

	if(thinc_get_user_info(&info) == T_AUTH_ERROR) {
		tc_error_msg("Couldn't get user authentication info\n");
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: got user information, username %s (%d chars). "
		"sending to server\n", func, info.user, info.user_len));

	auth.user_len = htons(info.user_len);
	auth.pass_len = htons(info.pass_len);

	res = tc_writeMsg(fb, T_CAUTH, &auth, T_CAUTH_SIZE);
	if(res == TC_ERROR) {
		tc_error_msg("Couldn't send authentication info to "
			"server\n");
		thinc_free_user_info(&info);
		return TC_ERROR;
	}

	res = writeData(fb, info.user, info.user_len);
	if(res == TC_ERROR) {
		tc_error_msg("Couldn't send authentication info to "
			"server\n");
		thinc_free_user_info(&info);
		return TC_ERROR;
	}
	res = writeData(fb, info.passwd, info.pass_len);
	if(res == TC_ERROR) {
		tc_error_msg("Couldn't send authentication info to "
			"server\n");
		thinc_free_user_info(&info);
		return TC_ERROR;
	}
	thinc_free_user_info(&info);

	/* now wait for ok or not ok from server */
	v2_readHeader(res, fb, h, T_SERV_OK);

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

#define v2_callReq(err, req, fb, name) do {	\
	err = req(fb);			\
	if(err == TC_ERROR) {		\
		tc_error_msg("Couldn't execute request %s\n", name); \
		return TC_ERROR;	\
	}				\
} while(0)

static int tc_v2Requests(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;

	tclient_dbg(2, ("%s: sending requests to server to finish the "
		"connection\n", func));

	v2_callReq(err, tc_v2ReqFBInfo, fb, "FBINFO");
	v2_callReq(err, tc_v2ReqCursor, fb, "CURSOR");
	v2_callReq(err, tc_v2ReqEncoder, fb, "ENCODER");
	v2_callReq(err, tc_v2ReqAppshare, fb, "APPSHARING");
	/* only do this if server doesn't support application sharing */
	if(!fb->app_share) {
		v2_callReq(err, tc_v2ReqFBData, fb, "FBDATA");
	}
	v2_callReq(err, tc_v2ReqCache, fb, "CACHE");
	
	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

/* tell server we're done and ready to go */
int tc_v2Done(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;

	tclient_dbg(2, ("%s: done with handshake, ready to go\n", func));

	err = tc_writeMsg(fb, T_CLIENT_DONE, NULL, T_CLIENT_DONE_SIZE);
	if(err == TC_ERROR) {
		tc_error_msg("Couldn't finish handshake\n");
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

/******** request functions ******/
static int tc_v2ReqFBInfo(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqFBInfo info;
	thinc_replyFBInfo fbi;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for framebuffer info\n", func));

	t_req_set_type(&info, T_REQ_FBINFO);
	v2_sendReq(err, fb, &info, t_req_size(T_REQ_FBINFO));

	tclient_dbg(2, ("%s: reading fb info\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_FBINFO);
	fbi = v2_getReply(g, fbinfo);
	
	fb->width = fb->curr_width = ntohs(fbi.width);
	fb->height = fb->curr_height = ntohs(fbi.height);
	fb->depth = fbi.depth;
	fb->bpp = fbi.bpp;

	if(tc_doCreateFB(fb) == TC_ERROR) {
		tc_v2SendNotOk(fb, "Couldn't create framebuffer");
		tc_error_msg("Couldn't create framebuffer\n");
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2ReqCursor(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqCursor cr;
	thinc_replyCursor ci;
	thinc_replyGen g;
	struct tc_cursorInfo *cursor = &fb->cursorInfo;

	tclient_dbg(2, ("%s: sending request for cursor\n", func));

	t_req_set_type(&cr, T_REQ_CURSOR);

	v2_sendReq(err, fb, &cr, t_req_size(T_REQ_CURSOR));

	tclient_dbg(2, ("%s: reading cursor info\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_CURSOR);
	if(reply == T_REPLY_NOTOK) {
		tclient_dbg(2, ("%s: server does not support client cursor\n",
			func));
		fb->cursorInfo.cursor = NULL;
		return TC_OK;
	}
	ci = v2_getReply(g, cursor);

	ci.flags = ntohs(ci.flags);
	if(ci.flags & T_CURSOR_ARGB)
		cursor->argb = T_TRUE;
	else
		cursor->argb = T_FALSE;

	cursor->size = ntohs(ci.size);
	cursor->fg = ntohl(ci.fg);
	cursor->bg = ntohl(ci.bg);
	cursor->width = ci.width;
	cursor->height = ci.height;
	cursor->xhot = ci.xhot;
	cursor->yhot = ci.yhot;
	cursor->x = ntohs(ci.x);
	cursor->y = ntohs(ci.y);

	if(tc_doCreateCursor(fb) == TC_ERROR) {
		tc_v2SendNotOk(fb, "Couldn't create cursor");
		tc_error_msg("Couldn't create cursor\n");
		return TC_ERROR;
	}

	err = readData(fb, cursor->cursor, cursor->size, T_TRUE);
	if(err == T_ERROR || err == T_CLOSED) {
		tc_error_msg("Couldn't read cursor data\n");
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2ReqEncoder(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqEncoder encr;
	thinc_replyEncoder ency;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for encoder info\n", func));

	t_req_set_type(&encr, T_REQ_ENCODER);
	v2_sendReq(err, fb, &encr, t_req_size(T_REQ_ENCODER));

	tclient_dbg(2, ("%s: reading encoder info\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_ENCODER);
	ency = v2_getReply(g, encoder);
	/* TODO: Should we do anything else here? (iow do we want to make
	 * zlib/png a runtime option? */
	if(ency.which == T_ENCODER_NONE) {
		tclient_dbg(2, ("%s: server is not encoding data\n", func));
	} else if(ency.which == T_ENCODER_PNG) {
		tclient_dbg(2, ("%s: using PNG encoder\n", func));
	/* FIXME: compiler complains on this comparison */
	} else if((unsigned int)ency.which == T_ENCODER_ZLIB) {
		tc_v2SendNotOk(fb, "Only support PNG encoder");
		tc_error_msg("Server uses ZLIB, but we only have support "
			"for PNG\n");
		return TC_ERROR;
	} else {
		/* eh? */
		tc_v2SendNotOk(fb, "Unknown encoder");
		tc_error_msg("Server uses an unknown encoder %d\n", 
			ency.which);
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2ReqFBData(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqFBData fbr;
	thinc_replyFBData fby;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for fb data\n", func));

	t_req_set_type(&fbr, T_REQ_FBDATA);

	v2_sendReq(err, fb, &fbr, t_req_size(T_REQ_FBDATA));

	tclient_dbg(2, ("%s: reading fb data\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_FBDATA);
	if(reply == T_REPLY_NOTOK) {
		tclient_dbg(2, ("%s: server will send the fb data later "
			"(hopefully!)\n", func));
		return TC_OK;
	}
	fby = v2_getReply(g, fbdata);

	fby.flags = ntohl(fby.flags);

	if(fby.flags & T_FB_COMPRESSED) {
		err = tc_readZFB(fb);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't read framebuffer data\n");
			return TC_ERROR;
		}
	} else {
		err = readData(fb, fb->fbdata, fb->total_size, T_TRUE);
		if(err == TC_ERROR || err == TC_CLOSED) {
			tc_error_msg("Couldn't read framebuffer data\n");
			return TC_ERROR;
		}
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2ReqCache(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqCacheSz cr;
	thinc_replyCacheSz cy;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for cache sizes\n", func));

	t_req_set_type(&cr, T_REQ_CACHESZ);

	v2_sendReq(err, fb, &cr, t_req_size(T_REQ_CACHESZ));

	tclient_dbg(2, ("%s: reading cache sizes\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_CACHESZ);
	if(reply == T_REPLY_NOTOK) {
		tclient_dbg(2, ("%s: server does not support caching updates\n",
			func));
		fb->cache_raw_bits = fb->cache_bit_bits = fb->cache_pix_bits \
			= 0;
		return TC_OK;
	}
	cy = v2_getReply(g, cachesz);

	fb->cache_raw_bits = cy.img;
	fb->cache_bit_bits = cy.bit;
	fb->cache_pix_bits = cy.pix;

	tclient_dbg(2, ("%s: finished. cache sizes %u, %u, %u\n", func,
		fb->cache_raw_bits, fb->cache_bit_bits, fb->cache_pix_bits));
	return TC_OK;
}

#ifdef TC_APPSHARE
static int tc_v2chkAppShare(struct tc_fb *fb);
static int tc_v2getAppList(struct tc_fb *fb, thinc_list *apps);

static int tc_v2ReqAppshare(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;
	thinc_list apps;

	tclient_dbg(2, ("%s: checking if the server supports app sharing\n",
		func));
	err = tc_v2chkAppShare(fb);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't initialize application "
			"sharing\n", func));
		return TC_ERROR;
	}
	if(!fb->app_share) {
		tclient_dbg(2, ("%s: server doesn't do appshare. finished\n",
			func));
		return TC_OK;
	}

	t_list_init(&apps);

	err = tc_v2getAppList(fb, &apps);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't get application list\n",
			func));
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2chkAppShare(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqAppSharing a;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for application sharing\n", func));

	t_req_set_type(&a, T_REQ_APP_SHARING);

	v2_sendReq(err, fb, &a, t_req_size(T_REQ_APP_SHARING));

	tclient_dbg(2, ("%s: reading server response\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	reply = t_reply_read_type(&g);
	if(reply == T_REPLY_UNKNOWN || reply == T_REPLY_NOTOK) {
		tclient_dbg(2, ("%s: server does not support application "
			"sharing\n", func));
		fb->app_share = 0;
	} else if(reply == T_REPLY_OK) {
		tclient_dbg(2, ("%s: server supports application sharing, "
			"initializing\n", func));
		fb->app_share = 1;
	} else {
		tclient_dbg(1, ("Error: %s: Unknown reply %d received from "
			"server\n", func, reply));
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_v2getAppList(struct tc_fb *fb, thinc_list *apps)
{
	tc_dbg_name(func)
	int err, reply, i;
	thinc_header h;
	thinc_reqAppList ar;
	thinc_replyAppList ay;
	thinc_replyGen g;
	struct tc_app *app;
	thinc_appDesc d;

	tclient_dbg(2, ("%s: sending request for application list\n", func));

	t_req_set_type(&ar, T_REQ_APP_LIST);

	v2_sendReq(err, fb, &ar, t_req_size(T_REQ_APP_LIST));

	tclient_dbg(2, ("%s: reading server response\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_APP_LIST);
	ay = v2_getReply(g, app_list);

	tclient_dbg(2, ("%s: reading %d apps\n", func, ay.count));

	for(i=0; i<ay.count; ++i) {
		err = readData(fb, &d, sizeof(d), T_TRUE);
		if(err == TC_ERROR) {
			tclient_dbg(1, ("Error: %s: Couldn't read application "
				"number %d (of %d)\n", func, i, ay.count));
			return TC_ERROR;
		}
		app = malloc(sizeof(*app) + d.name_len + 1);
		if(!app) {
			tclient_dbg(1, ("Error: %s: Couldn't allocate memory "
				"for app\n", func));
			return TC_ERROR;
		}
		app->name = (char*) (app+1);
		app->app.id = d.id;
		app->app.name_len = d.name_len;
		app->app.win_count = ntohs(d.win_count);
		err = readData(fb, app->name, app->app.name_len, T_TRUE);
		if(err == TC_ERROR) {
			tclient_dbg(1, ("Error: %s: Couldn't get application "
				"name\n", func));
			free(app);
			return TC_ERROR;
		}
		app->name[app->app.name_len] = '\0';

		t_list_add_tail(&app->list, apps);

		tclient_dbg(2, ("%s: got application %#x: %s. has %d windows\n",
			func, app->app.id, app->name, app->app.win_count));
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

#else	/* TC_APPSHARE */
static int tc_v2ReqAppshare(struct tc_fb *fb)
{
	fb->app_share = 0;
	return TC_OK;
}
#endif	/* TC_APPSHARE */

/******** helper functions *******/
static int tc_readVersionString(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, len=0, max=32;
	char *c, *ver;

	tclient_dbg(2, ("%s: reading server's version string\n",
		func));

	ver = malloc(max);
	if(!ver)
		return TC_ERROR;
	c = ver;

	while(1) {
		err = readData(fb, c, 1, T_TRUE);
		if(err == TC_ERROR || err == TC_CLOSED) {
			tc_error_msg("Couldn't read version string "
				"from server\n");
			free(ver);
			return TC_ERROR;
		} 
		if(!*c)
			break;
		++c;
		if(++len == max) {
			max <<= 2;
			ver = realloc(ver, max);
			if(!ver)
				return TC_ERROR;
			c = &ver[len];
		}
	}

	tclient_dbg(2, ("%s: server version (%d len): '%s'\n", func,
		len, ver));
	/* TODO: check the version */
	free(ver);

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

int tc_v2ReadNotOk(struct tc_fb *fb)
{
        tc_dbg_name(func)
        int err;
        thinc_serverNotOk n;
        char *reason;

        tclient_dbg(1, ("%s: reading server NOTOK\n", func));

        v2_readMsg(err, fb, &n, T_SERV_NOTOK_SIZE);

        n.reason_len = ntohl(n.reason_len);
        if(!n.reason_len)
                return TC_OK;

        reason = malloc(n.reason_len+1);
        if(!reason)
                return TC_ERROR;

        err = readData(fb, reason, n.reason_len, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		/* not an error worth reporting too loudly */
		tclient_dbg(1, ("Error: %s: Couldn't read server NOTOK "
			"message\n", func));
		return TC_ERROR;
	}
	reason[n.reason_len] = '\0';

        tc_error_msg("Server rejected the connection: %s\n",
                reason);

        free(reason);
        tclient_dbg(1, ("%s: finished\n", func));
        return TC_OK;
}

int tc_v2SendNotOk(struct tc_fb *fb, const char *reason)
{
        tc_dbg_name(func)
        int err, len;
        thinc_clientNotOk n;

        tclient_dbg(1, ("%s: sending NOTOK to server: '%s'\n", func,
                reason));

        len = strlen(reason);

        n.reason_len = htonl(len);
        err = tc_writeMsg(fb, T_CLIENT_NOTOK, &n, T_CLIENT_NOTOK_SIZE);
        if(err == TC_ERROR) {
                tclient_dbg(1, ("Error: %s: Couldn't send NOTOK to server\n",
			func));
                return T_ERROR;
        }
        err = writeData(fb, reason, len);
        if(err == TC_ERROR) {
                tclient_dbg(1, ("Error: %s: Couldn't send NOTOK to server\n",
			func));
                return T_ERROR;
        }

        tclient_dbg(1, ("%s: finished\n", func));
        return TC_OK;
}

int tc_v2SendReq(struct tc_fb *fb, void *req, int req_size) 
{
	return tc_writeMsg(fb, T_CLIENT_REQ, req, req_size);
}

static int tc_doCreateFB(struct tc_fb *fb) 
{
	tc_dbg_name(func)

	fb->Bpp = (fb->bpp >> 3);
	switch(fb->Bpp) {
		case 1: fb->Bpp_shift = 0;
			break;
		case 2: fb->Bpp_shift = 1;
			break;
		case 4: fb->Bpp_shift = 2;
			break;
		default:
			tc_error_msg("BUG: %s:%d:%s(): Don't know how to "
				"deal with bpp %d (%d)\n", __FILE__, __LINE__,
				__func__, fb->bpp, fb->Bpp);
			return TC_ERROR;
	}
	fb->total_size = TC_FBSIZE(fb);

	tclient_dbg(1, ("%s: creating fb, size is %dx%dx%d\n", func,
		fb->width, fb->height, fb->bpp));

	/* ok, create the fb */
	fb->fbdata = malloc(fb->total_size);
	if(!fb->fbdata) {
		/* must send negative ack back to server */
		tclient_dbg(1, ("Error: %s: Couldn't allocate memory "
			"for framebuffer\n", func));
		return TC_ERROR;
	}

	return TC_OK;
}

static int tc_doCreateCursor(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_cursorInfo *cursor = &fb->cursorInfo;

	tclient_dbg(1, ("%s: got cursor info. size %d, w %d, h %d "
		"x %d y %d, fg %#x, bg %#x\n", func, 
		cursor->size, cursor->width, cursor->height, cursor->x,
		cursor->y, cursor->fg, cursor->bg));

	/* allocate memory for the cursor */
	cursor->cursor = calloc(1, cursor->size);
	if(!cursor->cursor) {
		tclient_dbg(1, ("Error: %s: Couldn't allocate memory for "
			"cursor data\n", func));
		return TC_ERROR;
	}

	if(cursor->argb) {
		/* paranoia */
		cursor->source = cursor->mask = NULL;
	} else {
		cursor->source = cursor->cursor;
		cursor->mask = cursor->cursor + (cursor->size>>1);
	}
	/* This counter variable is used to track cursor movement.
	 * See tc_x11cursorMove and tc_MotionEv in x11.c file for 
	 * further details. */
	fb->move_count = 0;

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_readZFB(struct tc_fb *fb)
{
	tc_dbg_name(func)
	thinc_fbZData zline;
	char *zbuff = NULL, *curr;
	int zsize, i, res, zbuffsz = 0, count = 0;
	unsigned long usize;

	tclient_dbg(1, ("%s: reading compressed framebuffer\n", func));

	curr = fb->fbdata;

	for(i=0; i<fb->height; i++) {
		res = readData(fb, &zline, T_FB_ZDATA_SIZE, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read compressed "
				"scanline number %d\n", i));
			return TC_ERROR;
		}

		zsize = ntohl(zline.zsize);
		usize = ntohl(zline.usize);

		if(!zbuff || (zbuffsz < zsize)) {
			zbuffsz = MAX(2*zbuffsz, zsize+10);
			zbuff = realloc(zbuff, zbuffsz);
			if(!zbuff) {
				tclient_dbg(1, ("Error: Couldn't allocate "
					"memory for compressed scanline\n"));
				return TC_ERROR;
			}
		}

		count += zsize;

		res = readData(fb, zbuff, zsize, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read compressed "
				"scanline number %d\n", i));
			free(zbuff);
			return TC_ERROR;
		}

		/* uncompress it */
		res = uncompress((Bytef*)curr, &usize, (Bytef*)zbuff, zsize);
		if(res != Z_OK) {
			tclient_dbg(1, ("Error: Couldn't uncompress scanline "
				"%d: %d\n", i, res));
			free(zbuff);
			return TC_ERROR;
		}

		curr += usize;
	}

	free(zbuff);

	tclient_dbg(1, ("%s: read %d bytes of compressed data\n", func,
		count));

	return TC_OK;
}

/* closes the connection to the thinc server. if sendmsg is true, then
 * we send a shutdown message to the server before closing the socket */
void tc_closeConn(struct tc_fb *fb, int sendmsg)
{
	/* TODO: Add SSL shutdown function */
	close(fb->sock);
	tc_closeConnTsd(fb);
}

/* processes a server message read from sock. If the connection has been
 * closed, we close the function and return TC_CLOSED. On error, TC_ERROR is 
 * returned. On success, TC_OK is returned */
int tc_handleMsg(struct tc_fb *fb)
{
	tc_dbg_name(func)
	thinc_header h;
	int res;
	tc_stats_vars(start, end);

	instr_vars(smsg, emsg);

	tclient_dbg(1, ("%s: reading message from server\n", func));

	/* read the header to figure out what kind of message it is */
	res = tc_readHeader(fb, &h);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tc_error_msg("Couldn't read message header\n");
		return res;
	}

	tclient_dbg(1, ("%s: got header for message of type %d\n", func, 
		h.type));

	/* handle the message */
	switch(h.type) {
		case T_SPING:
			instr_start(smsg);
			res = tc_handlePing(fb, &h);
			instr_end(fb, smsg, emsg, total_ping);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle ping\n");
				return TC_ERROR;
			}
			break;
		/* Framebuffer messages */
		case T_FB_RUPDATE:
			time_add("RAW", TI_START);
			tc_stats_start(start);
			instr_start(smsg);
			res = tc_handleRUpdate(fb, &h);
			instr_end(fb, smsg, emsg, total_raw);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle raw update\n");
				return TC_ERROR;
			}
			tc_stats_end(start, end, fb->stats.rawtime);
			time_add("RAW", TI_END);
			break;
		case T_FB_COPY:
			time_add("COPY", TI_START);
			tc_stats_start(start);
			instr_start(smsg);
			res = tc_handleCopy(fb, &h);
			instr_end(fb, smsg, emsg, total_copy);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle copy\n");
				return TC_ERROR;
			}
			tc_stats_end(start, end, fb->stats.copytime);
			time_add("COPY", TI_END);
			break;
		case T_FB_SFILL:
			time_add("SFILL", TI_START);
			tc_stats_start(start);
			instr_start(smsg);
			res = tc_handleSFill(fb, &h);
			instr_end(fb, smsg, emsg, total_sfill);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle solid fill\n");
				return TC_ERROR;
			}
			tc_stats_end(start, end, fb->stats.sfilltime);
			time_add("SFILL", TI_END);
			break;
		case T_FB_PFILL:
			time_add("PFILL", TI_START);
			tc_stats_start(start);
			instr_start(smsg);
			res = tc_handlePFill(fb, &h);
			instr_end(fb, smsg, emsg, total_pfill);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle pixmap "
					"fill\n");
				return TC_ERROR;
			}
			tc_stats_end(start, end, fb->stats.pfilltime);
			time_add("PFILL", TI_END);
			break;
		case T_FB_GLYPH:
			time_add("GLYPH", TI_START);
			tc_stats_start(start);
			instr_start(smsg);
			res = tc_handleGlyph(fb, &h);
			instr_end(fb, smsg, emsg, total_glyph);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle bitmap "
					"fill\n");
				return TC_ERROR;
			}
			tc_stats_end(start, end, fb->stats.glyphtime);
			time_add("GLYPH", TI_END);
			break;
		case T_FB_BILEVEL:
			time_add("BILEVEL", TI_START);
			tc_stats_start(start);
			instr_start(smsg);
			res = tc_handleBilevel(fb, &h);
			instr_end(fb, smsg, emsg, total_bilevel);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle bilevel "
					"fill\n");
				return TC_ERROR;
			}
			tc_stats_end(start, end, fb->stats.bileveltime);
			time_add("BILEVEL", TI_END);
			break;
		/* Cursor Messages */
		case T_CURSOR_CHANGE:
			time_add("CURSOR", TI_START);
			instr_start(smsg);
			res = tc_handleCursorChange(fb, &h);
			instr_end(fb, smsg, emsg, total_cursor);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle cursor "
					"change msg\n");
				return TC_ERROR;
			}
			time_add("CURSOR", TI_END);
			break;
		case T_CURSOR_SHOWHIDE:
			time_add("CURSOR", TI_START);
			instr_start(smsg);
			res = tc_handleCursorShowHide(fb, &h);
			instr_end(fb, smsg, emsg, total_cursor);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle cursor "
					"show/hide msg\n");
				return TC_ERROR;
			}
			time_add("CURSOR", TI_END);
			break;
		case T_CURSOR_MOVE:
			time_add("CURSOR", TI_START);
			instr_start(smsg);
			res = tc_handleCursorMove(fb, &h);
			instr_end(fb, smsg, emsg, total_cursor);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle cursor move "
					"msg\n");
				return TC_ERROR;
			}
			time_add("CURSOR", TI_END);
			break;
		case T_CURSOR_COLOR:
			time_add("CURSOR", TI_START);
			instr_start(smsg);
			res = tc_handleCursorColor(fb, &h);
			instr_end(fb, smsg, emsg, total_cursor);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle cursor color "
					"msg\n");
				return TC_ERROR;
			}
			time_add("CURSOR", TI_END);
			break;
		/* video messages */
		case T_VIDEO_START:
			instr_start(smsg);
			res = tc_handleVideoStart(fb, &h);
			instr_end(fb, smsg, emsg, video_start);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle video start "
					"msg\n");
				return TC_ERROR;
			}
			break;
		case T_VIDEO_NEXT:
			instr_start(smsg);
			res = tc_handleVideoNext(fb, &h);
			instr_end(fb, smsg, emsg, video_next);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle video next "
					"msg\n");
				return TC_ERROR;
			}
			break;
		case T_VIDEO_MOVE:
			instr_start(smsg);
			res = tc_handleVideoMove(fb, &h);
			instr_end(fb, smsg, emsg, video_mod);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle video "
					"move msg\n");
				return TC_ERROR;
			}
			break;
		case T_VIDEO_SCALE:
			instr_start(smsg);
			res = tc_handleVideoScale(fb, &h);
			instr_end(fb, smsg, emsg, video_mod);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle video "
					"scale msg\n");
				return TC_ERROR;
			}
			break;
		case T_VIDEO_RESIZE:
			instr_start(smsg);
			res = tc_handleVideoResize(fb, &h);
			instr_end(fb, smsg, emsg, video_mod);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle video "
					"resize msg\n");
				return TC_ERROR;
			}
			break;
		case T_VIDEO_END:
			instr_start(smsg);
			res = tc_handleVideoEnd(fb, &h);
			instr_end(fb, smsg, emsg, video_end);
			if(res == TC_ERROR) {
				tc_error_msg("Couldn't handle video end "
					"msg\n");
				return TC_ERROR;
			}
			break;
		default:
			tc_error_msg("Got invalid message %d "
				"from server.\n", h.type);
			return TC_ERROR;
	}

	/* voila! */
	return res;
}

static int tc_handlePing(struct tc_fb *fb, thinc_header *h)
{
	/* nothing to do */
	return TC_PING;
}

/* processes a raw update mesage. Assumes header was already read by the
 * caller. Returns TC_ERROR on error, TC_OK on success */
static int tc_handleRUpdate(struct tc_fb *fb, thinc_header *h)
{
	tc_dbg_name(func)
	thinc_rawUpdate r;
	int res;

	instr_vars(shandle, ehandle);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading raw update message\n", func));

	/* read the msg */
	res = tc_readMsgNoHdr(fb, T_FB_RUPDATE, &r, T_FB_RUPDATE_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read RAW update message\n"));
		return TC_ERROR;
	}

	/* convert to proper byte order */
	r.x = ntohs(r.x);
	r.y = ntohs(r.y);
	r.width = ntohs(r.width);
	r.height = ntohs(r.height);

	/* chk that update is valid */
	if(	r.x < 0 || r.width < 0 || 
		r.x > fb->width || r.x+r.width > fb->width || 
		r.y < 0 || r.height < 0 || 
		r.y > fb->height || r.y+r.height > fb->height) {
		tclient_dbg(1, ("Error: Invalid update parameters "
			"%d,%d,%d,%d\n", r.x, r.y, r.width, r.height));
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: got update message, parameters are %d,%d,%d,%d\n",
		func, r.x, r.y, r.width, r.height));

	/* first distinction is whether update is resized or not */
	if(h->flags & T_FB_RUPDATE_RESIZED) {
		instr_end(fb, shandle, ehandle, handle_raw);
		res = tc_handleRszUpdate(fb, &r, h);
	} else {
		instr_end(fb, shandle, ehandle, handle_raw);
		res = tc_rawHandleUnresized(fb, &r, h);
	}
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't handle RAW update\n"));
		return TC_ERROR;
	}

	/* HACK: size is updated on underlying functions :( */
	tc_stats_inc(fb->stats.rawcount);

	tc_stats_dbg(1, ("%s: Have received %d RUPDATE, with "
		"%d bytes\n", func, fb->stats.rawcount, 
		fb->stats.rawbytes));
	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_rawHandleUnresized(struct tc_fb *fb, thinc_rawUpdate *r,
	thinc_header *h)
{
	tc_dbg_name(func)

	register char *to;
	register unsigned int real_width, line_size, i;
	unsigned int update_size, x_off; 
	int res;

	unsigned long int cache_id = 0;
	unsigned char cache_update = T_NO;

	instr_vars(shandle, ehandle, sdraw, edraw, sany, eany);

	instr_start(shandle);

	tclient_dbg(1, ("%s: handling unresized RAW update\n", func));

	if(h->flags & T_FB_RUPDATE_CACHED) {
		instr_end(fb, shandle, ehandle, handle_raw);
		instr_start(sany);
		res = tc_handleRawCachedUpdate(fb, r);
		instr_end(fb, sany, eany, raw_getcache);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cached "
				"update\n"));
			return TC_ERROR;
		}
		update_size = 0; /* so we don't count this update's size */
		goto end;
	}
	
	if(h->flags & T_FB_RUPDATE_ADDCACHE) {
		instr_start(sany);
		res = tc_handleUpdateAddCache(fb, &cache_id);
		instr_end(fb, sany, eany, raw_addcache);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cache add "
				"message\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got udpate to cache. id: %lu\n", func,
			cache_id));
		cache_update = T_YES;
	}

	/* need update_size for stats */
	line_size = (r->width ? pix_to_bytes(fb, r->width) : fb->Bpp);
	update_size = (r->height ? line_size*r->height : line_size);

	if(h->flags & T_FB_RUPDATE_COMPRESSED) {
		tclient_dbg(1, ("%s: got compressed raw update\n", func));
		instr_end(fb, shandle, ehandle, handle_raw);
		instr_start(sany);
		res = tc_handleZUpdate(fb, r, h);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle compressed "
				"RAW update\n"));
			return TC_ERROR;
		}
		instr_end(fb, sany, eany, raw_comp);
		goto end;
	}

	real_width = pix_to_bytes(fb, fb->width);
	x_off = pix_to_bytes(fb, r->x);

	to = fb->fbdata + x_off + r->y*real_width;

	instr_start(sany);
	for(i=0; i<r->height; ++i) {
		res = readData(fb, to, line_size, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read RAW update "
				"data\n"));
			return TC_ERROR;
		}
		to += real_width;
	}
	instr_end(fb, sany, eany, raw_read);
	instr_end(fb, shandle, ehandle, handle_raw);

#ifndef TC_DONT_UPDATE_X
	/* update the image */
	instr_start(sdraw);
	tclient_dbg(1, ("%s: update the image\n", func));
	res = tc_x11imgUpdate(fb, r->x, r->y, r->width, r->height);
	instr_end(fb, sdraw, edraw, draw_raw);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error udpating image\n"));
		return TC_ERROR;
	}
#endif

end:
	instr_start(shandle);
	/* XXX: HACK HACK HACK XXX */
	/* we need to fix this while update process, with just one place
	 * where updates to the fb happen, etc. For example, if update is
	 * compressed, all that the z-handling functions should do is
	 * get the data and copy it to a buffer, where this function
	 * picks it up and puts it in the fb */
	if(cache_update) {
		instr_start(sany);
		res = tc_addRUpdateCache(fb,r->x,r->y,r->width,r->height,
			cache_id);
		instr_end(fb, sany, eany, raw_addcache);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error adding update to cache\n"));
			return TC_ERROR;
		}
	}

	tc_stats_add(fb->stats.rawbytes, update_size + T_FB_RUPDATE_SIZE + 
		THINC_HEADERSIZE);
	tclient_dbg(1, ("%s: finished\n", func));

	instr_end(fb, shandle, ehandle, handle_raw);
	
	return TC_OK;
}

/* we may have multiple compressed updates per raw update, therefore:
 * -> compute totalsize of uncompressed raw update.
 * -> read compressed updates until all uncompressed data has been
 * read.
 * -> update appropiate region in the framebuffer using coordinates in 
 * raw update
 */
static int tc_handleZUpdate(struct tc_fb *fb, thinc_rawUpdate *r, 
	thinc_header *h)
{
	tc_dbg_name(func)
	int total_size, rect_width, got, err;
	int height;
	thinc_rectRegion rect;
	instr_vars(shandle, ehandle, sdraw, edraw);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading compressed RAW update\n", func));

	rect_width = (r->width ? pix_to_bytes(fb, r->width) : fb->Bpp);
	total_size = (r->height ? rect_width * r->height : rect_width);

	tclient_dbg(1, ("%s: expecting %d bytes of uncompressed data\n",
		func, total_size));

	rect.x = r->x;
	rect.width = r->width;
	height = rect.height = r->height;
	rect.y = r->y;

#if 0
	while(total_size) {
		/* we pass it a rectangle. We assume that the uncompressed
		 * data will be sent in whole lines, if not, this whole
		 * thing breaks. anyway, we pass a rectangle, with the
		 * starting point of where we want it to put the data.
		 * it should read, uncompress and update as much of that
		 * rectangle as it can. Upon return, it should update the
		 * height parameter to tell us how many lines it read, and
		 * return how much uncompressed data it got (or -1 on
		 * error).
		 * Upon return, we update the rectangle accordingly and if
		 * necessary call again for the next batch of compressed
		 * data.
		 */
		got = tc_handleZUpdateSimple(fb, &rect);
		if(got == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle compressed "
				"raw update\n"));
			return TC_ERROR;
		}

		/* update things */
		total_size -= got;
		height -= rect.height;
		if(total_size < 0 || height < 0) {
			tc_error_msg("BUG: %s:%d: %s(): Read more "
				"uncompressed "
				"data than expected. Have %d surplus bytes "
				"and %d lines\n", __FILE__, __LINE__, func, 
				total_size, height);
			return TC_ERROR;
		}

		rect.y = rect.y + rect.height;
		rect.height = height;

		tclient_dbg(1, ("%s: read %d bytes of uncompressed data. "
			"Need to read %d bytes/%d lines more\n", 
			func, got, total_size, height));
	}
#else
	got = tc_handleZUpdateSimple(fb, &rect);
	if(got == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't handle compressed "
			"raw update\n"));
		return TC_ERROR;
	}
#endif


	/* update the framebuffer */
	tclient_dbg(1, ("%s: got all the data, updating display: "
		"x %d, y %d, width %d, height %d\n", func, r->x, r->y,
		r->width, r->height));

	instr_end(fb, shandle, ehandle, handle_raw);
#ifndef TC_DONT_UPDATE_X
	instr_start(sdraw);
	err = tc_x11imgUpdate(fb, r->x, r->y, r->width, r->height);
	instr_end(fb, sdraw, edraw, draw_raw);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't update framebuffer image\n"));
		return TC_ERROR;
	}
#endif

	tc_stats_dbg(1, ("%s: Have received %d ZUPDATE, with "
		"%d bytes\n", func, fb->zfb.z_count, fb->zfb.z_bytes));
	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

/* thin wrapper around actual tc_resize call to avoid cluttering main
 * handle function with #ifdef's */
static int tc_handleRszUpdate(struct tc_fb *fb, thinc_rawUpdate *r, 
	thinc_header *h)
{
	return tc_resize_handle_update(fb, r, h);
}

/* how to do it:
 * use zlib's streaming model:
 * input -> fb->zfb.ubuff
 * output -> directly to the framebuffer.
 *
 * total = zdata.zsize;
 * while(total > 0) {
 * 	toread = MIN(total, fb->zfb.ubuff_size);
 * 	total -= toread;
 * 	read(fb->zfb.ubuff, toread);
 *
 * 	zstream->next_in = fb->zfb.ubuff;
 * 	zstream->avail_in = toread;
 * 	zstream->next_out = fbcurr;
 * 	zstream->avail_out = rect_width;
 *
 * 	=> play it safe. when we get out of this loop we know we're in
 * 	=> a line border, which makes things easier to manage
 * 	while(zstream->avail_in > rect_width) {
 * 	 	inflate(zstream, Z_SYNC_FLUSH);
 *	=> avail_out should be zero here
 *		fbcurr += rect_width;
 * 	 	zstream->next_out = fbcurr;
 * 	 	zstream->avail_out = rect_width;
 * 	 }
 *
 * 	 => ok, need to get more input data
 * }
 */
static int tc_handleZUpdateSimple(struct tc_fb *fb, thinc_rectRegion *rect)
{
	tc_dbg_name(func)
	thinc_fbZData zdata;

	int x_off, rect_width, line_width, zsize, usize, err, height, i;
	char *fbcurr;
	
	struct tc_compress *zfb = &fb->zfb;
	png_structp png_ptr;
	png_infop info_ptr;
	instr_vars(start, end);

	tclient_dbg(1, ("%s: reading compressed update starting at (%d,%d)\n",
		func, rect->x, rect->y));

	/* FIXME: No way we have to do this every single time :( */
	instr_start(start);
	err = tc_pngInit(zfb);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't initialize png\n"));
		return TC_ERROR;
	}
	instr_end(fb, start, end, comp_init);

	png_ptr = zfb->png_ptr;
	info_ptr = zfb->info_ptr;

	instr_start(start);
	err = readData(fb, &zdata, T_FB_ZDATA_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read compressed message "
			"header\n"));
		tc_pngEnd(zfb);
		return TC_ERROR;
	}
	instr_end(fb, start, end, comp_read);

	zsize = ntohl(zdata.zsize);
	usize = ntohl(zdata.usize);

	tclient_dbg(1, ("%s: got zdata update: usize %d, zsize %d, err %d\n", 
		func, usize, zsize, err));
	
	/* we read from the network on demand. since we're doing
	 * buffering, then all is cool */

	/* read png info */
	tclient_dbg(1, ("%s: reading png info\n", func));
	instr_start(start);
	png_read_info(png_ptr, info_ptr);
	instr_end(fb, start, end, comp_init);

	/* decompress it, line by line */
	x_off = pix_to_bytes(fb, rect->x);
	rect_width = (rect->width ? pix_to_bytes(fb, rect->width) : fb->Bpp);
	line_width = pix_to_bytes(fb, fb->width);
	height = usize / rect_width;

	/* safety check */
	if(usize % rect_width != 0) {
		tclient_dbg(1, ("Error: Server didn't send whole line "
			"update!!\n"));
		return TC_ERROR;
	}

	fbcurr = fb->fbdata + x_off + rect->y * line_width;

	tclient_dbg(1, ("%s: decompressing %d lines (%d bytes, %d bytes per "
		"line)\n", func, height, usize, rect_width));

	instr_start(start);
	for(i=0; i<height; ++i) {
		png_read_row(png_ptr, (png_bytep)fbcurr, (png_bytep)NULL);
		fbcurr += line_width;
	}
	instr_end(fb, start, end, comp_comp);
	tclient_dbg(1, ("%s: read %d lines, shutting down png\n", func, 
		height));
	instr_start(start);
	tc_pngEnd(zfb);
	instr_end(fb, start, end, comp_end);

	/* update the rectangle */
	rect->height = height;

	tc_stats_inc(zfb->z_count);
	tc_stats_add(zfb->z_bytes, zsize + T_FB_ZDATA_SIZE);
	tclient_dbg(1, ("%s: finished\n", func));
	return usize;
}

int tc_pngInit(struct tc_compress *zfb)
{
	png_structp png_ptr;
	png_infop info_ptr;

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL,
		NULL);
	if(!png_ptr) {
		tclient_dbg(1, ("Error: Couldn't allocate png structure\n"));
		return TC_ERROR;
	}
	info_ptr = png_create_info_struct(png_ptr);
	if(!info_ptr) {
		tclient_dbg(1, ("Error: Couldn't allocate png info "
			"structure\n"));
		png_destroy_read_struct(&png_ptr, NULL, NULL);
		return TC_ERROR;
	}

	png_set_read_fn(png_ptr, (void*)zfb, tc_pngReadData);

	zfb->png_ptr = png_ptr;
	zfb->info_ptr = info_ptr;

	/* FIXME: BEFORE - AFTER should probably be determined by byte
	 * order of X server??
	 * Note that this is only valid while we don't use the alpha
	 * channel */
	if(zfb->fb->bpp == 32)
		png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);

	return TC_OK;
}

void tc_pngEnd(struct tc_compress *zfb)
{
	png_read_end(zfb->png_ptr, zfb->info_ptr);
	png_destroy_read_struct(&zfb->png_ptr, &zfb->info_ptr, NULL);
}

static void tc_pngReadData(png_structp png_ptr, png_bytep data,
	png_size_t length)
{
	tc_dbg_name(func)
	struct tc_compress *zfb = png_get_io_ptr(png_ptr);
	int err;
	instr_vars(start, end);

	tclient_dbg(1, ("%s: reading %d bytes\n", func, length));

	instr_start(start);
	err = readData(zfb->fb, data, length, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read compressed data. I'll "
			"probably die now\n"));
		return;
	}
	instr_end(zfb->fb, start, end, comp_read);
	tclient_dbg(1, ("%s: finished\n", func));
}

static int tc_handleCopy(struct tc_fb *fb, thinc_header *h)
{
	tc_dbg_name(func)
	thinc_fbCopy c;
	int res;

	instr_vars(shandle, ehandle, sdraw, edraw);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading copy message\n", func));

	/* read the msg */
	res = tc_readMsgNoHdr(fb, T_FB_COPY, &c, T_FB_COPY_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read COPY message\n"));
		return TC_ERROR;
	} 
	
	/* convert to proper byte order */
	c.sx = ntohs(c.sx);
	c.sy = ntohs(c.sy);
	c.dx = ntohs(c.dx);
	c.dy = ntohs(c.dy);
	c.width = ntohs(c.width);
	c.height = ntohs(c.height);

	/* chk the copy is valid */
	if(	c.sx < 0 || c.sx+c.width > fb->width || 
		c.sy < 0 || c.sy+c.height > fb->height ||
		c.dx < 0 || c.dx+c.width > fb->width || 
		c.dy < 0 || c.dy+c.height > fb->height) {
		tclient_dbg(1, ("Error: Invalid copy parameters %d, %d, "
			"%d, %d, "
			"%d, %d\n", c.sx, c.sy, c.dx, c.dy, c.width, 
			c.height));
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: parameters src: %d,%d, dst: %d,%d, size: %d,%d\n", 
		func, c.sx, c.sy, c.dx, c.dy, c.width, c.height));

	/* ok, pass it to the X layer */
	instr_end(fb, shandle, ehandle, handle_copy);
#ifndef TC_DONT_UPDATE_X
	instr_start(sdraw);
	res = tc_x11copyData(fb, c.sx, c.sy, c.dx, c.dy, c.width, c.height);
	instr_end(fb, sdraw, edraw, draw_copy);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't copy data\n"));
		return TC_ERROR;
	}
#endif

	tc_stats_inc(fb->stats.copycount);
	tc_stats_add(fb->stats.copybytes, T_FB_COPY_SIZE + THINC_HEADERSIZE);
	
	tc_stats_dbg(1, ("%s: Have received %d COPY, for %d "
		"bytes\n",func, fb->stats.copycount, fb->stats.copybytes));
	tclient_dbg(1, ("%s: finished\n", func));

	return TC_OK;
}

static int tc_handleSFill(struct tc_fb *fb, thinc_header *h)
{
	tc_dbg_name(func)
	thinc_fbSFill s;
	thinc_rectRegion rect;
	int res, i;
	int size = 0;

	instr_vars(shandle, ehandle, sdraw, edraw, sany, eany);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading solid fill message\n", func));

	/* read the msg */
	instr_start(sany);
	res = tc_readMsgNoHdr(fb, T_FB_SFILL, &s, T_FB_SFILL_SIZE);
	instr_end(fb, sany, eany, sfill_read);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read SFILL message\n"));
		return TC_ERROR;
	}

	/* fix byte order */
	s.pixel = ntohl(s.pixel);
	s.numrects = ntohs(s.numrects);

	tclient_dbg(1, ("%s: got fill for %d rectangles. color %u\n",
		func, s.numrects, s.pixel));
	size+= (s.numrects * T_RECTREGION_SIZE);

	for(i=0; i<s.numrects; i++) {
		instr_start(sany);
		res = readData(fb, &rect, T_RECTREGION_SIZE, T_TRUE);
		instr_end(fb, sany, eany, sfill_read);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read rectangle %d for "
				"PFILL msg\n", i));
			return TC_ERROR;
		} 
		fix_tc_rect(rect);

		tclient_dbg(1, ("%s: rectangle %d: x %d, y %d, w %d, h %d\n",
			func, i, rect.x, rect.y, rect.width, rect.height));
		instr_end(fb, shandle, ehandle, handle_sfill);
#ifndef TC_DONT_UPDATE_X
		instr_start(sdraw);
		res = tc_x11solidFill(fb, rect.x, rect.y, rect.width, 
			rect.height, s.pixel);
		instr_end(fb, sdraw, edraw, draw_sfill);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't exec SFILL\n"));
			return TC_ERROR;
		}
#endif
		/* for next loop */
		instr_start(shandle);
	}
	size += THINC_HEADERSIZE + T_FB_SFILL_SIZE;

	tc_stats_add(fb->stats.sfillcount, s.numrects);
	tc_stats_add(fb->stats.sfillbytes, size);

	tc_stats_dbg(1, ("%s: finished. Have received %d SFILL. this command's "
		"size %d and total bytes until now %d\n", func, 
		fb->stats.sfillcount, size, fb->stats.sfillbytes));

	tclient_dbg(1, ("%s: finished\n", func));

	return TC_OK;
}

static int tc_handlePFill(struct tc_fb *fb, thinc_header *h)
{
	tc_dbg_name(func)
	thinc_fbPFill p;
	char *pixmap = NULL;	/* the pixmap */
	int psize;		/* size of pixmap */
	thinc_rectRegion rect;
	int res, i, line_size;

	unsigned char resized;
	unsigned long int cache_id = 0;
	int cache_update = T_NO, cached = T_NO;
	void *handle;

	instr_vars(shandle, ehandle, sdraw, edraw);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading pixmap fill message\n", func));

	/* read the msg */
	res = tc_readMsgNoHdr(fb, T_FB_PFILL, &p, T_FB_PFILL_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read PFILL message\n"));
		return TC_ERROR;
	}

	p.x = ntohs(p.x);
	p.y = ntohs(p.y);
	p.width = ntohs(p.width);
	p.height = ntohs(p.height);
	p.numrects = ntohs(p.numrects);

	/* chk the fill is valid */
	if( (p.width < 0) || (p.height < 0) || (p.numrects < 0) ) {
		tclient_dbg(1, ("Error: Invalid fill parameters %d, %d, %d\n", 
			p.width, p.height, p.numrects));
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: got fill message: %d, %d, %d. flags %d\n", 
		func, p.width, p.height, p.numrects, h->flags));

	/* check for interesting flags */
	resized = T_FALSE;
	if(h->flags & T_FB_PFILL_RESIZED)
		resized = T_TRUE;
	if(h->flags & T_FB_PIXMAP_CACHED) {
		res = tc_handleUpdateCached(fb, &cache_id);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cached "
				"PFILL update\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got cached pixmap update %lu\n", func,
			cache_id));
		cached = T_YES;
	} else if(h->flags & T_FB_PIXMAP_ADDCACHE) {
		res = tc_handleUpdateAddCache(fb, &cache_id);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cache add "
				"message\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got update to cache, id: %lu\n",
			func, cache_id));
		cache_update = T_YES;
	}

	line_size = pix_to_bytes(fb, p.width);
	psize = line_size * p.height;

	if(cached) {
		res = tc_getPixmapCached(fb, cache_id, &pixmap, psize);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Couldn't get pixmap from cache\n"));
			return TC_ERROR;
		}
	} else {
		tclient_dbg(1, ("%s: allocating pixmap, size %d\n", func,
			psize));
		pixmap = malloc(psize*sizeof(unsigned char));
		if(!pixmap) {
			tclient_dbg(1, ("Couldn't allocate memory "
				"for pixmap: %s\n", strerror(errno)));
			return TC_ERROR;
		}
		
		tclient_dbg(1, ("%s: reading pixmap, size %d\n", func,
			psize));
		/* Read the pixmap */
		res = readData(fb, pixmap, psize, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Couldn't read pixmap for "
				"PFILL message %d\n", res));
			free(pixmap);
  			return TC_ERROR;
  		}
	}
	handle = tc_x11pfillNew(fb, pixmap, p.x, p.y, p.width,
		p.height);
	if(!handle) {
		tclient_dbg(1, ("Error: Couldn't handle PFILL udpate\n"));
		free(pixmap);
		return TC_ERROR;
	}
	if(!cached && !cache_update)
		/* don't need this anymore */
		free(pixmap);

	/* Now read the rectangles and update each one in turn */
	for(i=0; i<p.numrects; i++) {
		res = readData(fb, &rect, T_RECTREGION_SIZE, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read rectangle %d for "
				"PFILL msg\n", i));
			goto error;
		}
		fix_tc_rect(rect);

		tclient_dbg(1, ("%s: filling rectangle %d,%d,%d,%d\n",
			func, rect.x, rect.y, rect.width, rect.height));

		instr_end(fb, shandle, ehandle, handle_pfill);
#ifndef TC_DONT_UPDATE_X
		instr_start(sdraw);
		res = tc_x11pfillFill(fb, handle, &rect, resized);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't fill rectangle %d\n", 
				i));
			goto error;
		}
		instr_end(fb, sdraw, edraw, draw_pfill);
#endif	/* NOT TC_DONT_UPDATE_X */
		/* for next loop */
		instr_start(shandle);
	}

	if(cache_update) {
		res = tc_addPixmapCache(fb, cache_id, pixmap, psize);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error adding pfill update to "
				"cache\n"));
			goto error;
		}
	} else if(cached)
		psize = 0;	/* for stats */

	tc_x11fillDestroyWrap(fb, handle);

	tc_stats_inc(fb->stats.pfillcount);
	tc_stats_add(fb->stats.pfillbytes, psize);
	tc_stats_add(fb->stats.pfillbytes, p.numrects * T_RECTREGION_SIZE);
	tc_stats_add(fb->stats.pfillbytes, T_FB_PFILL_SIZE + THINC_HEADERSIZE);

	tc_stats_dbg(1, ("%s: Finished. Have received %d PFILL. current "
		"command's size %d, total size until now %d\n", func,
		fb->stats.pfillcount, size, fb->stats.pfillbytes));

	tclient_dbg(1, ("%s: Finished\n", func));

	return TC_OK;
error:
	tc_x11fillDestroyWrap(fb, handle);
	return TC_ERROR;
}

static int tc_handleGlyph(struct tc_fb *fb, thinc_header *h)
{
	tc_dbg_name(func)
	thinc_fbGlyph b;
	char *bitmap = NULL;
	int res;
	unsigned int bitmap_size, i;
	thinc_rectRegion rect;
	unsigned long int cache_id = 0;
	int cache_update = T_NO;
	int cached = T_NO;
	void *handle = NULL;

	instr_vars(shandle, ehandle, sdraw, edraw);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading bitmap message\n", func));

	/* read the message */
	res = tc_readMsgNoHdr(fb, T_FB_GLYPH, &b, T_FB_GLYPH_SIZE);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't read GLYPH message\n"));
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Connection closed while waiting "
			"for msg\n"));
		return TC_ERROR;
	}

	b.x = ntohs(b.x);
	b.y = ntohs(b.y);
	b.width = ntohs(b.width);
	b.height = ntohs(b.height);
	b.color = ntohl(b.color);
	b.numrects = ntohs(b.numrects);

	tclient_dbg(1, ("%s: got bitmap message: x %d y %d width %d "
		"height %d color %d\n", func, b.x, b.y, b.width, b.height,
		b.color));

	bitmap_size = T_BITMAP_SIZE(b.width, b.height);
	if(bitmap_size > T_MAX_BITMAP_SIZE) {
		tclient_dbg(1, ("Error: Received bitmap is too big %d\n", 
			bitmap_size));
		return TC_ERROR;
	}

	/* check for interesting flags */
	if(h->flags & T_FB_BITMAP_CACHED) {
		res = tc_handleUpdateCached(fb, &cache_id);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cached "
				"update\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got cached update, id: %lu\n",func,
			cache_id));
		cached = T_YES;
	} else if(h->flags & T_FB_BITMAP_ADDCACHE) {
		res = tc_handleUpdateAddCache(fb, &cache_id);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cache add "
				"message\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got update to cache, id: %lu\n",func,
			cache_id));
		cache_update = T_YES;
	}

	tclient_dbg(1, ("%s: updating bitmap of size %d\n",func,bitmap_size));

	if(cached) {
		res = tc_getBitmapCached(fb, cache_id, &bitmap, bitmap_size);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read bitmap for GLYPH "
				"message\n"));
			goto error;
		}
	} else {
		bitmap = malloc(bitmap_size *sizeof(unsigned char));
		if(!bitmap ) {
			tclient_dbg(1, ("Error: Couldn't allocate memory for "
				"bitmap\n"));
			return TC_ERROR;
		}
		res = readData(fb, bitmap, bitmap_size, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read bitmap for GLYPH "
				"message\n"));
			free(bitmap);
			goto error;
		}
	}
#ifdef TC_DUMP_BITMAP
	tc_dumpBitmap(bitmap, b.width, b.height);
#endif

	handle = tc_x11glyphNew(fb, bitmap, b.x, b.y, b.width,
		b.height, b.color);
	if(!handle) {
		tclient_dbg(1, ("Error: Couldn't handle GLYPH update\n"));
		free(bitmap);
		goto error;
	}
	if(!cached && !cache_update) {
		free(bitmap);
		bitmap = NULL;
	}

	/* have glyph, read rectangles */
	if(b.numrects == 0) {
		/* no rectangles. make up one out of the bitmap's dimensions
		 * if the bitmap is valid */
		if(	b.x < 0 || b.x+b.width > fb->width || 
			b.y < 0 || b.y+b.height > fb->height) {
			tclient_dbg(1, ("Error: invalid parameters in GLYPH "
				"message\n"));
			return TC_ERROR;
		}

		rect.x = b.x;
		rect.y = b.y;
		rect.width = b.width;
		rect.height = b.height;

		/* so that client-graph works as expected */
		if(!cached) {
			tclient_dbg(1, ("%s: filling rectangle %d,%d,%d,%d\n", 
				func, rect.x, rect.y, rect.width, rect.height));
		}
		instr_end(fb, shandle, ehandle, handle_glyph);
#ifndef TC_DONT_UPDATE_X
		instr_start(sdraw);
		res = tc_x11bitmapFill(fb, handle, &rect);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't execute GLYPH message\n"));
			goto error;
		}
		instr_end(fb, sdraw, edraw, draw_glyph);
#endif /* TC_DONT_UPDATE_X */
		goto end;
	}

	/* more than 1 rectangle. Read each one in turn and fill them */
	tclient_dbg(1, ("%s: got %d rectangles\n", func, b.numrects));
	for(i=0; i<b.numrects; i++) {
		res = readData(fb, &rect, T_RECTREGION_SIZE, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read rectangle %d for "
				"GLYPH msg\n", i));
			goto error;
		}

		fix_tc_rect(rect);

		if(	rect.x < 0 || rect.x+rect.width > fb->width || 
			rect.y < 0 || rect.y+rect.height > fb->height) {
			tclient_dbg(1, ("Error: invalid parameters in GLYPH "
				"message\n"));
			return TC_ERROR;
		}

		/* so that client-graph works as expected */
		if(!cached) {
			tclient_dbg(1, ("%s: filling rectangle %d,%d,%d,%d\n", 
				func, rect.x, rect.y, rect.width, 
				rect.height));
		}
		instr_end(fb, shandle, ehandle, handle_glyph);
#ifndef TC_DONT_UPDATE_X
		instr_start(sdraw);
		res = tc_x11bitmapFill(fb, handle, &rect);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't execute GLYPH "
				"message\n"));
			goto error;
		}
		instr_end(fb, sdraw, edraw, draw_glyph);
#endif
		/* for next loop */
		instr_start(shandle);
	}

end:
	instr_start(shandle);

	if(cache_update) {
		res = tc_addBitmapCache(fb, cache_id, bitmap, bitmap_size);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error adding update to cache\n"));
			goto error;
		}
	} else if(cached)
		/* we don't want to count this size in the stats */
		bitmap_size = 0;

	tc_x11fillDestroyWrap(fb, handle);

	tc_stats_inc(fb->stats.glyphcount);
	tc_stats_add(fb->stats.glyphbytes, bitmap_size);
	tc_stats_add(fb->stats.glyphbytes, b.numrects * T_RECTREGION_SIZE);
	tc_stats_add(fb->stats.glyphbytes, T_FB_GLYPH_SIZE + THINC_HEADERSIZE);

	tc_stats_dbg(1, ("%s: finished. Have received %d GLYPH. current "
		"command's size %d. Total size %d\n", func, 
		fb->stats.glyphcount, size, fb->stats.glyphbytes));
	tclient_dbg(1, ("%s: finished\n", func));
	instr_end(fb, shandle, ehandle, handle_glyph);

	return TC_OK;

error:
	tc_x11fillDestroyWrap(fb, handle);
	return TC_ERROR;
}

static int tc_handleBilevel(struct tc_fb *fb, thinc_header *h)
{
	tc_dbg_name(func)
	thinc_fbBilevel b;
	char *bitmap = NULL;
	int res;
	unsigned int bitmap_size, i;
	thinc_rectRegion rect;
	unsigned long int cache_id = 0;
	int cache_update = T_NO;
	int cached = T_NO;
	void *handle = NULL;
	instr_vars(shandle, ehandle, sdraw, edraw);

	instr_start(shandle);

	tclient_dbg(1, ("%s: reading bilevel message\n", func));

	/* read the message */
	res = tc_readMsgNoHdr(fb, T_FB_BILEVEL, &b, T_FB_BILEVEL_SIZE);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't read BILEVEL message\n"));
		return TC_ERROR;
	} else if(res == 0) {
		tclient_dbg(1, ("Error: Connection closed while waiting for msg\n"));
		return TC_ERROR;
	}

	b.x = ntohs(b.x);
	b.y = ntohs(b.y);
	b.width = ntohs(b.width);
	b.height = ntohs(b.height);
	b.fg = ntohl(b.fg);
	b.bg = ntohl(b.bg);
	b.numrects = ntohs(b.numrects);

	tclient_dbg(1, ("%s: got bilevel message: x %d y %d width %d "
		"height %d fg %d bg %d\n", func, b.x, b.y, b.width, b.height,
		b.fg, b.bg));

	bitmap_size = T_BITMAP_SIZE(b.width, b.height);
	if(bitmap_size > T_MAX_BITMAP_SIZE) {
		tclient_dbg(1, ("Error: Received bitmap is too big %d\n", 
			bitmap_size));
		return TC_ERROR;
	}

	/* check for interesting flags */
	if(h->flags & T_FB_BITMAP_CACHED) {
		res = tc_handleUpdateCached(fb, &cache_id);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cached "
				"update\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got cached update, id: %lu\n",func,
			cache_id));
		cached = T_YES;
	} else if(h->flags & T_FB_BITMAP_ADDCACHE) {
		res = tc_handleUpdateAddCache(fb, &cache_id);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't handle cache add "
				"message\n"));
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got update to cache, id: %lu\n",func,
			cache_id));
		cache_update = T_YES;
	}

	if(!cached) {
		bitmap = malloc(bitmap_size *sizeof(unsigned char));
		if(!bitmap ) {
			tclient_dbg(1, ("Error: Couldn't allocate memory for "
				"bitmap\n"));
			return TC_ERROR;
		}
	}

	tclient_dbg(1, ("%s: reading bitmap of size %d\n", func, bitmap_size));

	if(cached) {
		res = tc_getBitmapCached(fb, cache_id, &bitmap, bitmap_size);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read bitmap for "
				"BILEVEL message\n"));
			goto error;
		}
	} else {
		bitmap = malloc(bitmap_size *sizeof(unsigned char));
		if(!bitmap ) {
			tclient_dbg(1, ("Error: Couldn't allocate memory for "
				"bitmap\n"));
			return TC_ERROR;
		}
		res = readData(fb, bitmap, bitmap_size, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read bitmap for "
				"BILEVEL message\n"));
			free(bitmap);
			goto error;
		}
	}
#ifdef TC_DUMP_BITMAP
	tc_dumpBitmap(bitmap, b.width, b.height);
#endif

	handle = tc_x11bilevelNew(fb, bitmap, b.x, b.y, b.width,
		b.height, b.fg, b.bg);
	if(!handle) {
		tclient_dbg(1, ("Error: Couldn't handle BILEVEL update\n"));
		free(bitmap);
		goto error;
	}
	if(!cached && !cache_update) {
		free(bitmap);
		bitmap = NULL;
	}

	/* have bilevel, read rectangles */
	if(b.numrects == 0) {
		/* no rectangles. make up one out of the bitmap's dimensions
		 * if the bitmap is valid */
		if(	b.x < 0 || b.x+b.width > fb->width || 
			b.y < 0 || b.y+b.height > fb->height) {
			tclient_dbg(1, ("Error: invalid parameters in GLYPH "
				"message\n"));
			return TC_ERROR;
		}

		rect.x = b.x;
		rect.y = b.y;
		rect.width = b.width;
		rect.height = b.height;
		if(!cached) {
			tclient_dbg(1, ("%s: filling rectangle %d,%d,%d,%d\n", 
				func, rect.x, rect.y, rect.width, rect.height));
		}
		instr_end(fb, shandle, ehandle, handle_bilevel);
#ifndef TC_DONT_UPDATE_X
		instr_start(sdraw);
		res = tc_x11bitmapFill(fb, handle, &rect);
		instr_end(fb, sdraw, edraw, draw_bilevel);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't execute BILEVEL "
				"message\n"));
			return TC_ERROR;
		}
#endif
		goto end;
	}

	/* more than 1 rectangle. Read each one in turn and fill them */
	tclient_dbg(1, ("%s: have %d rectangles\n", func, b.numrects));
	for(i=0; i<b.numrects; i++) {
		res = readData(fb, &rect, T_RECTREGION_SIZE, T_TRUE);
		if(res == TC_ERROR || res == TC_CLOSED) {
			tclient_dbg(1, ("Error: Couldn't read rectangle %d for "
				"BILEVEL msg\n", i));
			goto error;
		}

		fix_tc_rect(rect);

		if(	rect.x < 0 || rect.x+rect.width > fb->width || 
			rect.y < 0 || rect.y+rect.height > fb->height) {
			tclient_dbg(1, ("Error: invalid parameters in GLYPH "
				"message\n"));
			return TC_ERROR;
		}

		/* so that client-graph works as expected */
		if(!cached) {
			tclient_dbg(1, ("%s: filling rectangle %d,%d,%d,%d\n", 
				func, rect.x, rect.y, rect.width, 
				rect.height));
		}

		instr_end(fb, shandle, ehandle, handle_bilevel);
#ifndef TC_DONT_UPDATE_X
		instr_start(sdraw);
		res = tc_x11bitmapFill(fb, handle, &rect);
		instr_end(fb, sdraw, edraw, draw_bilevel);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error: Couldn't execute BILEVEL "
				"message\n"));
			goto error;
		}
#endif
		/* for next loop */
		instr_start(shandle);
	}

end:
	instr_start(shandle);
	if(cache_update) {
		res = tc_addBitmapCache(fb, cache_id, bitmap, bitmap_size);
		if(res == TC_ERROR) {
			tclient_dbg(1, ("Error adding update to cache\n"));
			goto error;
		}
	} else if(cached)
		/* we don't want to count this size in the stats */
		bitmap_size = 0;

	tc_x11fillDestroyWrap(fb, handle);

	tc_stats_inc(fb->stats.bilevelcount);
	tc_stats_add(fb->stats.bilevelbytes, bitmap_size);
	tc_stats_add(fb->stats.bilevelbytes, b.numrects * T_RECTREGION_SIZE);
	tc_stats_add(fb->stats.bilevelbytes, T_FB_BILEVEL_SIZE + 
		THINC_HEADERSIZE);

	tc_stats_dbg(1, ("%s: finished. Have received %d BILEVEL. current "
		"command's size %d. Total size %d\n", func, 
		fb->stats.bilevelcount, size, fb->stats.bilevelbytes));
	tclient_dbg(1, ("%s: finished\n", func));

	instr_end(fb, shandle, ehandle, handle_bilevel);
	return TC_OK;
error:
	tc_x11fillDestroyWrap(fb, handle);
	return TC_ERROR;
}

static int tc_handleCursorChange(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	int res;
	thinc_cursorChange c;
	struct tc_cursorInfo *cursor = &fb->cursorInfo;

	if(!cursor->cursor) {
		tclient_msg("BUG: %s (%s:%d) We aren't supposed to be doing "
			"cursor handling or something bad happened to our "
			"cursor data\n", func, __FILE__, __LINE__);
		return TC_ERROR;
	}

	res = tc_readMsgNoHdr(fb, T_CURSOR_CHANGE, &c, T_CURSOR_CHANGE_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Couldn't read CURSOR_CHANGE message\n"));
		return TC_ERROR;
	}
	if(h->flags & T_CURSOR_ARGB)
		cursor->argb = T_TRUE;
	else
		cursor->argb = T_FALSE;

	cursor->xhot = c.xhot;
	cursor->yhot = c.yhot;
	c.size = ntohs(c.size);
	/* this should only happen once (at the most) */
	if(c.size > cursor->size) {
		cursor->size = c.size;
		cursor->cursor = realloc(cursor->cursor, cursor->size);
		if(!cursor->cursor) {
			tclient_dbg(1, ("Cannot allocate memory for "
				"new cursor\n"));
			return TC_ERROR;
		}
		if(!cursor->argb) {
			cursor->source = cursor->cursor;
			cursor->mask = cursor->cursor + (cursor->size>>1);
		}
	} else if(c.size != cursor->size) {
		cursor->size = c.size;
		if(!cursor->argb) {
			cursor->source = cursor->cursor;
			cursor->mask = cursor->cursor + (cursor->size>>1);
		}
	}

	tclient_dbg(1, ("%s: reading cursor. x %d y %d. size %d\n", func, 
		c.xhot, c.yhot, cursor->size));

	res = readData(fb, cursor->cursor, cursor->size, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Couldn't read cursor data\n"));
		return TC_ERROR;
	}

#ifndef TC_DONT_UPDATE_X
	res = tc_x11cursorChange(fb);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't change cursor\n"));
		return TC_ERROR;
	}
#endif

	tc_stats_inc(fb->stats.cursorcount);

	tclient_dbg(1, ("%s: finished\n", func));

	return TC_OK;
}

/* X doesn't support cursor hiding, so we just silently ignore this
 * message */
static int tc_handleCursorShowHide(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	int res;
	thinc_cursorShowHide s;

	if(!fb->cursorInfo.cursor) {
		tclient_msg("BUG: %s (%s:%d) We aren't supposed to be doing "
			"cursor handling or something bad happened to our "
			"cursor data\n", func, __FILE__, __LINE__);
		return TC_ERROR;
	}

	res = tc_readMsgNoHdr(fb, T_CURSOR_SHOWHIDE, &s,
		T_CURSOR_SHOWHIDE_SIZE);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't read CURSOR_SHOWHIDE "
			"message\n"));
		return TC_ERROR;
	} else if(res == TC_CLOSED) { /* connection closed */
		tclient_dbg(1, ("Error: Connection closed while reading "
			"CURSOR_SHOWHIDE message\n"));
		return TC_ERROR;
	} 

	tclient_dbg(1, ("%s: We should %s the cursor, but we can't\n", func,
		(s.showhide == T_SHOW ? "show" : "hide")));

	tc_stats_inc(fb->stats.cursorcount);

	return TC_OK;
}

static int tc_handleCursorMove(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	int res;
	thinc_cursorMove m;

	if(!fb->cursorInfo.cursor) {
		tclient_msg("BUG: %s (%s:%d) We aren't supposed to be doing "
			"cursor handling or something bad happened to our "
			"cursor data\n", func, __FILE__, __LINE__);
		return TC_ERROR;
	}

	res = tc_readMsgNoHdr(fb, T_CURSOR_MOVE, &m, T_CURSOR_MOVE_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read CURSOR_MOVE message\n"));
		return TC_ERROR;
	}

	m.x = ntohs(m.x);
	m.y = ntohs(m.y);

	tclient_dbg(1, ("%s: got cursor move message x %d y %d\n", func,
		m.x, m.y));

#ifndef TC_DONT_UPDATE_X
	res = tc_x11cursorMove(fb, m.x, m.y);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't execute motion message\n"));
		return TC_ERROR;
	}
#endif

	tc_stats_inc(fb->stats.cursorcount);

	return TC_OK; 
}

static int tc_handleCursorColor(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	int res;
	thinc_cursorColor c;

	if(!fb->cursorInfo.cursor) {
		tclient_msg("BUG: %s (%s:%d) We aren't supposed to be doing "
			"cursor handling or something bad happened to our "
			"cursor data\n", func, __FILE__, __LINE__);
		return TC_ERROR;
	}

	res = tc_readMsgNoHdr(fb, T_CURSOR_COLOR, &c, T_CURSOR_COLOR_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read CURSOR_COLOR message\n"));
		return TC_ERROR;
	} 
	c.fg = ntohl(c.fg);
	c.bg = ntohl(c.bg);

	tclient_dbg(1, ("%s: got cursor color message fg %d bg %d\n", func,
		c.fg, c.bg));

#ifndef TC_DONT_UPDATE_X
	tc_x11cursorRecolor(fb, c.fg, c.bg);
#endif

	tc_stats_inc(fb->stats.cursorcount);

	return TC_OK;
}

/* read a message of type type from socket sock. If received message has
 * different type, the message is not read (since we don't know the size
 * of the message that's coming in) and the type received is
 * returned (considered an error). On success, the message's type is
 * returned. If type is 0, no checking for correct type is made */
int tc_readMsg(struct tc_fb *fb, int type, void *msg, int msg_size)
{
	tc_dbg_name(func)
	int res;
	static char *buf = NULL, *from;
	static int buf_size;
	thinc_header* h;
	int total;

	total = msg_size + THINC_HEADERSIZE;

	if(!buf) {
		buf = calloc(total+1, sizeof(char));
		buf_size = total+1;
	} else if(total > buf_size) { /* just in case */
		tclient_dbg(3, ("%s: had to realloc. msg_size %d\n", func,
			total));
		buf = realloc(buf, (total+1)*sizeof(char));
		buf_size = total+1;
	}
	if(!buf) { /* something failed */
		tclient_dbg(1, ("Error: Couldn't allocate memory for message "
			"buffer: %s\n", strerror(errno)));
		return TC_ERROR;
	}
		
	/* read everything in one swoop */
	tclient_dbg(3, ("%s: reading %d\n", func, total));
	res = readData(fb, buf, total, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read the message: %s\n",
			strerror(errno)));
		return TC_ERROR;
	}
	/* chk the type, and if good copy it into back to the caller */
	h = (thinc_header*)buf;

	tclient_dbg(3, ("%s: got message %d\n", func, h->type));

	if(type && (h->type != type)) {
		tclient_dbg(1, ("Error: Received different type of message\n"));
		return h->type;
	}

	from = buf + THINC_HEADERSIZE;
	tc_pass_or_die(from < buf+buf_size, 
		"BUG: %s:%d:%s(): pointer went beyond buffer\n",
		__FILE__, __LINE__, func);
	tc_pass_or_die(from + msg_size <= buf+buf_size,
		"BUG: %s:%d:%s(): we just overflowed our read buffer\n",
		__FILE__, __LINE__, func);
	
	memcpy(msg, from, msg_size);

	tclient_dbg(3, ("%s: finished\n", func));

	return h->type;
}

int tc_readMsgAndHeader(struct tc_fb *fb, thinc_header *h, void *msg, 
	int msg_size)
{
	int res;

	/* read the header */
	res = readData(fb, h, THINC_HEADERSIZE, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read message\n"));
		return TC_ERROR;
	}

	/* read the message */
	res = readData(fb, msg, msg_size, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read message\n"));
		return TC_ERROR;
	}

	return h->type;
}

/* reads in a message header from socket sock. returns TC_ERROR on error, 
 * TC_CLOSED if connection closed, or the header size on success */
int tc_readHeader(struct tc_fb *fb, thinc_header *h)
{
	return readData(fb, h, THINC_HEADERSIZE, T_TRUE);
}

/* reads a message of type type from socket sock. Assumes the message
 * header was already read. Returns TC_ERROR on error, TC_CLOSED if 
 * connection closed, the message size on success */
int tc_readMsgNoHdr(struct tc_fb *fb, int type, void *msg, int msg_size)
{
	return readData(fb, msg, msg_size, T_TRUE);
}

/* write the specified message type onto the socket. The message is passed
 * in msg, which is of size msg_size. Returns TC_CLOSED on success, 
 * TC_ERROR on error */
int tc_writeMsg(struct tc_fb *fb, int type, const void *msg, int msg_size)
{
	tc_dbg_name(func)
	int res;
	thinc_header h;

	/* send the header */
	h.type = type;
	h.length = htons(msg_size);

	tclient_dbg(2, ("%s: sending message of type %d\n", func, h.type));

	res = writeData(fb, &h, THINC_HEADERSIZE);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't send message\n"));
		return TC_ERROR;
	} 

	/* send the message */
	if(!msg_size)
		goto end;
	
	res = writeData(fb, msg, msg_size);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't send message\n"));
		return TC_ERROR;
	}

end:
	tclient_dbg(2, ("%s: finished\n", func));

	return TC_OK;
}

/* Ugly hack. This is so we read from the sound socket. Will (hopefully)
 * go away when we have a centralized manager */
int tc_readSoundMsgAndHeader(struct tc_fb *fb, thinc_header *h, void *msg, 
	int msg_size)
{
	int res;

	/* read the header */
	res = readSoundData(fb, h, THINC_HEADERSIZE, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read message\n"));
		return TC_ERROR;
	}

	/* read the message */
	res = readSoundData(fb, msg, msg_size, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tclient_dbg(1, ("Error: Couldn't read message\n"));
		return TC_ERROR;
	}

	return h->type;
}

/* reads a message. Assumes the message
 * header was already read. Returns TC_ERROR on error, TC_CLOSED if 
 * connection closed, the message size on success */
int tc_readSoundMsgNoHdr(struct tc_fb *fb, void *msg, int msg_size)
{
	return readSoundData(fb, msg, msg_size, T_TRUE);
}

/* ibid. */
int tc_writeSoundMsg(struct tc_fb *fb, int type, const void *msg, int msg_size)
{
	tc_dbg_name(func)
	int res;
	thinc_header h;

	/* send the header */
	h.type = type;

	tclient_dbg(2, ("%s: sending message of type %d\n", func, h.type));

	res = writeSoundData(fb, &h, THINC_HEADERSIZE);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't send message\n"));
		return TC_ERROR;
	} 

	/* send the message */
	res = writeSoundData(fb, msg, msg_size);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't send message\n"));
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));

	return TC_OK;
}

/* reads in a message header from socket sock. returns TC_ERROR on error, 
 * TC_CLOSED if connection closed, or the header size on success */
int tc_readSoundHeader(struct tc_fb *fb, thinc_header *h)
{
	return readSoundData(fb, h, THINC_HEADERSIZE, T_TRUE);
}

int tc_handleUpdateAddCache(struct tc_fb *fb, unsigned long int *cache_id)
{
	int res;
	thinc_cacheAdd ca;
	res = readData(fb, &ca, T_FB_CACHE_ADD_SIZE, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED)
		return TC_ERROR;
	*cache_id = ntohl(ca.cache_id);
	return TC_OK;
}

int tc_handleUpdateCached(struct tc_fb *fb, unsigned long int *cache_id)
{
	int res;
	thinc_cacheData cd;
	res = readData(fb, &cd, T_FB_CACHE_DATA_SIZE, T_TRUE);
	if(res == TC_ERROR || res == TC_CLOSED)
		return TC_ERROR;
	*cache_id = ntohl(cd.cache_id);
	return TC_OK;
}

static int tc_addRUpdateCache(struct tc_fb *fb, int x, int y, int width, 
	int height, unsigned long int cache_id)
{
	tc_dbg_name(func)
	unsigned long int length;
	char *data;
	int err;
	instr_vars(start, end);

	tclient_dbg(1, ("%s: adding raw update to cache\n", func));

	if(!fb->raw_cache)
		return TC_ERROR;

	data = tc_getData(fb, x, y, width, height, &length);
	if(!data)
		return TC_ERROR;

	tclient_dbg(1, ("%s: got data, %lu bytes. adding to cache\n", func,
		length));
	instr_start(start);
	err = thinc_cache_add_hash(fb->raw_cache, cache_id, data);
	instr_end(fb, start, end, cache_add);
	if(err == T_ERROR)
		return TC_ERROR;

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_handleRawCachedUpdate(struct tc_fb *fb, thinc_rawUpdate* r)
{
	int res;
	unsigned long int cache_id;
	char *data;
	instr_vars(shandle, ehandle, sdraw, edraw, scache, ecache);

	instr_start(shandle);

	res = tc_handleUpdateCached(fb, &cache_id);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't read cached data message\n"));
		return TC_ERROR;
	}

	instr_start(scache);
	data = thinc_cache_get(fb->raw_cache, cache_id);
	instr_end(fb, scache, ecache, cache_get);
	if(!data) {
		tclient_dbg(1, ("Error: Don't have the data for cached update "
			"%lu\n", cache_id));
		return TC_ERROR;
	}

	/* Only for unresized. resize has its own code */
	res = tc_putData(fb, data, r->x, r->y, r->width, r->height);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't copy cached update to fb\n"));
		return TC_ERROR;
	}

	instr_end(fb, shandle, ehandle, handle_raw);
#ifndef TC_DONT_UPDATE_X
	instr_start(sdraw);
	res = tc_x11imgUpdate(fb, r->x, r->y, r->width, r->height);
	if(res == TC_ERROR) {
		tclient_dbg(1, ("Error: Couldn't udpate fb image\n"));
		return TC_ERROR;
	}
	instr_end(fb, sdraw, edraw, draw_raw);
#endif

	return TC_OK;
}

static char * tc_getData(struct tc_fb *fb, int x, int y, int width,
	int height, unsigned long int *size)
{
	register char *from, *to, *data;
	int x_off, rect_width, line_width, i;
	unsigned long int total;

	instr_vars(sany, eany);

	x_off = pix_to_bytes(fb, x);
	rect_width = (width ? pix_to_bytes(fb, width) : fb->Bpp);
	line_width = pix_to_bytes(fb, fb->width);
	from = fb->fbdata + x_off + y*line_width;

	total = rect_width * height;
	instr_start(sany);
	data = malloc(total*sizeof(unsigned char));
	instr_end(fb, sany, eany, raw_malloc);
	if(!data)
		return NULL;
	to = data;

	for(i=0; i<height; ++i) {
		instr_start(sany);
		memcpy(to, from, rect_width);
		instr_end(fb, sany, eany, raw_memcpy);
		to += rect_width;
		from += line_width;
	}

	*size = total;
	return data;
}

static int tc_putData(struct tc_fb *fb, char *data, int x, int y, 
	int width, int height)
{
	register char *from, *to;
	int x_off, rect_width, line_width, i;
	instr_vars(sany, eany);

	x_off = pix_to_bytes(fb, x);
	rect_width = (width ? pix_to_bytes(fb, width) : fb->Bpp);
	line_width = pix_to_bytes(fb, fb->width);
	to = fb->fbdata + x_off + y*line_width;
	from = data;

	for(i=0; i<height; ++i) {
		instr_start(sany);
		memcpy(to, from, rect_width);
		instr_end(fb, sany, eany, raw_memcpy);
		from += rect_width;
		to += line_width;
	}

	return TC_OK;
}

#if 0
/* our own solid fill function. really slow. don't use */
static void tc_fillImg(struct tc_fb *fb, unsigned long pixel, int x, int y,
	int width, int height)
{
	register unsigned long *curr;
	char *to;
	int x_off, rect_width, line_width;
	register int i, j;
	instr_vars(sany, eany);

	x_off = pix_to_bytes(fb, x);
	rect_width = (width ? pix_to_bytes(fb, width) : fb->Bpp);
	line_width = pix_to_bytes(fb, fb->width);
	to = fb->fbdata + x_off + y*line_width;

	curr = (unsigned long*)to;

	for(i=0; i<height; ++i) {
		for(j=0; j<width; ++j) {
			instr_start(sany);
			*curr++ = pixel;
			instr_end(fb, sany, eany, sfill_byhand);
		}
		to += line_width;
		curr = (unsigned long*)to;
	}
}
#endif

static int tc_getBitmapCached(struct tc_fb *fb, unsigned long int cache_id,
	char **bitmap, unsigned int bitmap_size)
{
	char *data;
	instr_vars(start, end);

	instr_start(start);
	data = thinc_cache_get(fb->bit_cache, cache_id);
	instr_end(fb, start, end, cache_get);
	if(!data) {
		tclient_dbg(1, ("Error: Couldn't get data from cache for %lu\n",
			cache_id));
		return TC_ERROR;
	}

	*bitmap = data;
	return TC_OK;
}

static int tc_addBitmapCache(struct tc_fb *fb, unsigned long int cache_id,
	char *bitmap, unsigned int bitmap_size)
{
	int err;
	instr_vars(start, end);

	instr_start(start);
	err = thinc_cache_add_hash(fb->bit_cache, cache_id, bitmap);
	instr_end(fb, start, end, cache_add);
	if(err == T_ERROR)
		return TC_ERROR;
	return TC_OK;
}

static int tc_getPixmapCached(struct tc_fb *fb, unsigned long int cache_id,
	char **pixmap, unsigned int pixmap_size)
{
	char *data;
	instr_vars(start, end);

	instr_start(start);
	data = thinc_cache_get(fb->pix_cache, cache_id);
	instr_end(fb, start, end, cache_get);
	if(!data) {
		tclient_dbg(1, ("Error: Couldn't get data from cache for %lu\n",
			cache_id));
		return TC_ERROR;
	}

	*pixmap = data;
	return TC_OK;
}

static int tc_addPixmapCache(struct tc_fb *fb, unsigned long int cache_id,
	char *pixmap, unsigned int pixmap_size)
{
	int err;
	instr_vars(start, end);

	instr_start(start);
	err = thinc_cache_add_hash(fb->pix_cache, cache_id, pixmap);
	instr_end(fb, start, end, cache_add);
	if(err == T_ERROR)
		return TC_ERROR;
	return TC_OK;
}

#ifdef TC_DUMP_BITMAP
static void tc_dumpBitmap(char *bitmap, int width, int height)
{
	static int count = 0;
        const char *base = "/tmp/glyph/glyph";
        char fn[255];
        const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
        int dump, write_header = 1;
        unsigned int a, size, w2;

        strncpy(fn, base, 16);
        sprintf(fn+16, "%d", count++);
        tclient_dbg(1, ("writing file %s\n", fn));

        dump = creat(fn,mode);

        /* X bitmaps are padded to 4 bytes */
        w2 = (width+7) >> 3;
        size = w2*height;
	tclient_dbg(1, ("width %d, w2 %d\n", width, w2));

        if(write_header) {
                a = htonl(width);
                write(dump, &a, 4);
                a = htonl(height);
                write(dump, &a, 4);
                a = htonl(size);
                write(dump, &a, 4);
                a = 0;
                write(dump, &a, 4);
        }
        write(dump, bitmap, size);
        close(dump);
}
#endif
